Package: smcfcs
Title: Multiple Imputation of Covariates by Substantive Model
        Compatible Fully Conditional Specification
Version: 2.0.2
Authors@R: c(person("Jonathan", "Bartlett", email = "jonathan.bartlett1@lshtm.ac.uk",role = c("aut", "cre")),
    person("Ruth", "Keogh", email = "ruth.keogh@lshtm.ac.uk",role = c("aut")),
    person("Edouard F.", "Bonneville",role = c("aut")),
    person("Lars", "van der Burg",role = c("aut")),
    person("Claus Thorn", "Ekstrøm",role = c("ctb")))
URL: https://github.com/jwb133/smcfcs
Description: Implements multiple imputation of missing covariates by
    Substantive Model Compatible Fully Conditional Specification.
    This is a modification of the popular FCS/chained equations
    multiple imputation approach, and allows imputation of missing
    covariate values from models which are compatible with the user
    specified substantive model.
Depends: R (>= 3.1.2)
License: GPL-3
LazyData: true
Imports: MASS, survival, VGAM, stats, rlang, checkmate, abind, brglm2
Suggests: knitr, rmarkdown, mitools, ggplot2, kmi, flexsurv, stringr
VignetteBuilder: knitr
RoxygenNote: 7.3.3
Encoding: UTF-8
BugReports: https://github.com/jwb133/smcfcs/issues
NeedsCompilation: no
Packaged: 2026-04-30 11:43:27 UTC; lshjb12
Author: Jonathan Bartlett [aut, cre],
  Ruth Keogh [aut],
  Edouard F. Bonneville [aut],
  Lars van der Burg [aut],
  Claus Thorn Ekstrøm [ctb]
Maintainer: Jonathan Bartlett <jonathan.bartlett1@lshtm.ac.uk>
Repository: CRAN
Date/Publication: 2026-05-01 05:10:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(plot,smcfcs)
export(smcfcs)
export(smcfcs.casecohort)
export(smcfcs.dtsam)
export(smcfcs.finegray)
export(smcfcs.flexsurv)
export(smcfcs.nestedcc)
export(smcfcs.parallel)
import(stats)
importFrom(rlang,.data)
importFrom(survival,Surv)

[![](http://cranlogs.r-pkg.org/badges/grand-total/smcfcs)](https://cran.r-project.org/package=smcfcs)

smcfcs is an R package implementing Substantive Model Compatibly Fully Conditional Specification Multiple Imputation. Examples and further details are given in the package documentation and vignette.

To install the latest GitHub development version, run:

```r
install.packages("devtools")

devtools::install_github("jwb133/smcfcs")
```

## -----------------------------------------------------------------------------
library(smcfcs)
ex_linquad[1:10, ]

## -----------------------------------------------------------------------------
set.seed(123)
# impute missing values in x, compatibly with quadratic substantive model
imps <- smcfcs(originaldata=ex_linquad, smtype = "lm", smformula = "y~z+x+I(x^2)", method = c("", "", "norm", ""))

## -----------------------------------------------------------------------------
# fit substantive model
library(mitools)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + I(x^2)))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
# impute missing values in x, compatibly with model for y which omits the quadratic effect
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x", method = c("", "", "norm", ""))

## -----------------------------------------------------------------------------
# fit substantive model
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + I(x^2)))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
# impute, including v as a covariate in the substantive/outcome model
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+I(x^2)+v", method = c("", "", "norm", ""))
# fit substantive model, which omits v
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + I(x^2)))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
predMatrix <- array(0, dim = c(ncol(ex_linquad), ncol(ex_linquad)))
predMatrix[3, ] <- c(0, 1, 0, 1)
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+I(x^2)", method = c("", "", "norm", ""), predictorMatrix = predMatrix)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + I(x^2)))
summary(MIcombine(models))

## ----fig.width = 6, fig.height = 4--------------------------------------------
# impute once with a larger number of iterations than the default 10
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+I(x^2)", method = c("", "", "norm", ""), predictorMatrix = predMatrix, m = 1, numit = 100)
# plot estimates of the parameters of the substantive model against iteration number
plot(imps)

## -----------------------------------------------------------------------------
library(smcfcs)
ex_linquad[1:10, ]

## -----------------------------------------------------------------------------
set.seed(123)
# impute missing values in x, compatibly with quadratic substantive model
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""))

## -----------------------------------------------------------------------------
# fit substantive model
library(mitools)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
# impute missing values in x, compatibly with model for y which omits the quadratic effect
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x", method = c("", "", "norm", "x^2", ""))

## -----------------------------------------------------------------------------
# fit substantive model
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
# impute, including v as a covariate in the substantive/outcome model
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq+v", method = c("", "", "norm", "x^2", ""))
# fit substantive model, which omits v
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))

## -----------------------------------------------------------------------------
predMatrix <- array(0, dim = c(ncol(ex_linquad), ncol(ex_linquad)))
predMatrix[3, ] <- c(0, 1, 0, 0, 1)
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""), predictorMatrix = predMatrix)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))

## ----fig.width = 6, fig.height = 4--------------------------------------------
# impute once with a larger number of iterations than the default 10
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""), predictorMatrix = predMatrix, m = 1, numit = 100)
# plot estimates of the parameters of the substantive model against iteration number
plot(imps)

#' Substantive model compatible fully conditional specification imputation of covariates.
#'
#' Multiply imputes missing covariate values using substantive model compatible
#' fully conditional specification.
#'
#' smcfcs imputes missing values of covariates using the Substantive Model Compatible
#' Fully Conditional Specification multiple imputation approach proposed by
#' Bartlett \emph{et al} 2015 (see references).
#'
#' Imputation is supported for linear regression (\code{"lm"}),
#' logistic regression (\code{"logistic"}), bias reduced logistic regression (\code{"brlogistic"}),
#' Poisson regression (\code{"poisson"}), Weibull (\code{"weibull"}) and Cox regression
#' for time to event data (\code{"coxph"}),
#' and Cox models for competing risks data (\code{"compet"}). For \code{"coxph"},
#' the event indicator should be integer coded with 0 for censoring and 1 for event.
#' For \code{"compet"}, a Cox model is assumed for each cause specific hazard function,
#' and the event indicator
#' should be integer coded with 0 corresponding to censoring, 1 corresponding to
#' failure from the first cause etc.
#'
#' The function returns a list. The first element \code{impDataset} of the list is a list of the imputed
#' datasets. Models (e.g. the substantive model) can be fitted to each and results
#' combined using Rubin's rules using the mitools package, as illustrated in the examples.
#'
#' The second element \code{smCoefIter} is a three dimensional array containing the values
#' of the substantive model parameters obtained at the end of each iteration of the algorithm.
#' The array is indexed by: imputation number, parameter number, iteration.
#'
#' If the substantive model is linear, logistic or Poisson regression,
#' \code{smcfcs} will automatically impute missing outcomes, if present, using
#' the specified substantive model. However, even in this case, the user should
#' specify "" in the element of method corresponding to the outcome variable.
#'
#' The bias reduced methods make use of the \code{\link[brglm2]{brglm2}} package to fit the corresponding glms
#' using Firth's bias reduced approach. These may be particularly useful to use in case
#' of perfect prediction, since the resulting model estimates are always guaranteed to be
#' finite, even in the case of perfect prediction.
#'
#' The development of this package was supported by the UK Medical Research Council
#' (Fellowship MR/K02180X/1 and grant MR/T023953/1). Part of its development took place while Bartlett was
#' kindly hosted by the University of Michigan's Department of Biostatistics & Institute for
#' Social Research.
#'
#' The structure of many of the arguments to \code{smcfcs} are based on those of
#' the excellent \code{mice} package.
#'
#' @param originaldata The original data frame with missing values.
#' @param smtype A string specifying the type of substantive model. Possible
#' values are \code{"lm"}, \code{"logistic"}, \code{"brlogistic"}, \code{"poisson"}, \code{"weibull"},
#' \code{"coxph"}, \code{"compet"}.
#' @param smformula The formula of the substantive model. For \code{"weibull"} and \code{"coxph"}
#' substantive models the left hand side should be of the form \code{"Surv(t,d)"}. For \code{"compet"}
#' substantive models, a list should be passed consisting of the Cox models
#' for each cause of failure (see example).
#' @param method A required vector of strings specifying for each variable either
#' that it does not need to be imputed (""), the type of regression model to be
#' be used to impute. Possible values are \code{"norm"} (normal linear regression),
#' \code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
#' \code{"poisson"} (Poisson regression),
#' \code{"podds"} (proportional odds regression for ordered categorical variables),
#' \code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
#' or a custom expression which defines a passively imputed variable, e.g.
#' \code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
#' normal variable which is measured with error. If this is specified for a variable,
#' the \code{"errorProneMatrix"} argument should also be used.
#' @param predictorMatrix An optional predictor matrix. If specified, the matrix defines which
#' covariates will be used as predictors in the imputation models
#' (the outcome must not be included). The i'th row of the matrix should consist of
#' 0s and 1s, with a 1 in the j'th column indicating the j'th variable be used
#' as a covariate when imputing the i'th variable. If not specified, when
#' imputing a given variable, the imputation model covariates are the other
#' covariates of the substantive model which are partially observed
#' (but which are not passively imputed) and any fully observed covariates (if present)
#' in the substantive model. Note that the outcome variable is implicitly conditioned
#' on by the rejection sampling scheme used by smcfcs, and should not be specified as a predictor
#' in the predictor matrix.
#' @param m The number of imputed datasets to generate. The default is 5.
#' @param numit The number of iterations to run when generating each imputation.
#' In a (limited) range of simulations good performance was obtained with the
#' default of 10 iterations. However, particularly when the proportion of missingness
#' is large, more iterations may be required for convergence to stationarity.
#' @param rjlimit Specifies the maximum number of attempts which should be made
#' when using rejection sampling to draw from imputation models. If the limit is reached
#' when running a warning will be issued. In this case it is probably advisable to
#' increase the \code{rjlimit} until the warning does not appear.
#' @param noisy logical value (default FALSE) indicating whether output should be noisy, which can
#' be useful for debugging or checking that models being used are as desired.
#' @param errorProneMatrix An optional matrix which if specified indicates that some variables
#' are measured with classical measurement error. If the i'th variable is measured with error
#' by variables j and k, then the (i,j) and (i,k) entries of this matrix should be 1, with the
#' remainder of entries 0. The i'th element of the method argument should then be specified
#' as \code{"latnorm"}. See the measurement error vignette for more details.
#'
#' @return A list containing:
#'
#' \code{impDatasets} a list containing the imputed datasets
#'
#' \code{smCoefIter} a three dimension matrix containing the substantive model parameter
#' values. The matrix is indexed by [imputation,parameter number,iteration]
#'
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @example data-raw/examples.r
#'
#' @references Bartlett JW, Seaman SR, White IR, Carpenter JR. Multiple imputation of covariates
#' by fully conditional specification: accommodating the substantive model. Statistical Methods
#' in Medical Research 2015; 24(4): 462-487. \doi{10.1177/0962280214521348}
#' @import stats
#' @importFrom survival Surv
#' @export
smcfcs <- function(originaldata, smtype, smformula, method, predictorMatrix = NULL, m = 5, numit = 10, rjlimit = 1000, noisy = FALSE, errorProneMatrix = NULL) {
  # call core  smcfcs function, passing through arguments
  smcfcs.core(originaldata, smtype, smformula, method, predictorMatrix, m, numit, rjlimit, noisy, errorProneMatrix = errorProneMatrix)
}

#' Substantive model compatible fully conditional specification imputation of covariates for case cohort studies
#'
#' Multiply imputes missing covariate values using substantive model compatible
#' fully conditional specification for case cohort studies.
#'
#' This version of \code{smcfcs} is designed for use with case cohort studies but where the analyst does not wish to,
#' or cannot (due to not having the necessary data) impute the full cohort. The function's arguments are the same
#' as for the main smcfcs function, except for \code{smformula}, \code{in.subco}, and \code{sampfrac} - see above
#' for details on how these should be specified.
#'
#' @author Ruth Keogh \email{ruth.keogh@@lshtm.ac.uk}
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @param originaldata The case-cohort data set (NOT a full cohort data set with a case-cohort substudy within it)
#' @param smformula A formula of the form "Surv(entertime,t,d)~x", where d is the event (d=1) or censoring (d=0) indicator, t is the event or censoring time and entertime is equal to the time origin (typically 0) for individuals in the subcohort and is equal to (t-0.001) for cases outside the subcohort [this sets cases outside the subcohort to enter follow-up just before their event time. The value 0.001 may need to be modified depending on the time scale.]
#' @param in.subco The name of a column in the dataset with 0/1s that indicates whether the subject is in the subcohort
#' @param sampfrac The proportion of individuals from the underlying full cohort who are in the subcohort
#'
#' @inheritParams smcfcs
#'
#' @example data-raw/cc_example.r
#'
#' @export
smcfcs.casecohort <- function(originaldata, smformula, sampfrac, in.subco, method, predictorMatrix = NULL, m = 5, numit = 10, rjlimit = 1000, noisy = FALSE,
                              errorProneMatrix = NULL) {
  smcfcs.core(originaldata,
    smtype = "casecohort", smformula, method, predictorMatrix, m, numit, rjlimit, noisy, sampfrac = sampfrac, in.subco = in.subco,
    errorProneMatrix = errorProneMatrix
  )
}

#' Substantive model compatible fully conditional specification imputation of covariates for nested case control
#' studies
#'
#' Multiply imputes missing covariate values using substantive model compatible
#' fully conditional specification for nested case control studies.
#'
#' This version of \code{smcfcs} is designed for use with nested case control studies. The function's arguments are the same
#' as for the main smcfcs function, except for \code{smformula}, \code{set}, \code{event} and \code{nrisk} - see above
#' for details on how these should be specified.
#'
#' @author Ruth Keogh \email{ruth.keogh@@lshtm.ac.uk}
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @param originaldata The nested case-control data set (NOT a full cohort data set with a case-cohort substudy within it)
#' @param smformula A formula of the form "Surv(t,case)~x+strata(set)", where case is case-control indicator, t is the event or censoring time. Note that t could be set to the case's event time for the matched controls in a given set. The right hand side should include the case control set as a strata term (see example).
#' @param set variable identifying matched sets in nested case-control study
#' @param event variable which indicates who is a case/control in the nested case-control sample. Note that this is distinct from d.
#' @param nrisk variable which is the number at risk (in the underlying full cohort) at the event time for the case in each matched set (i.e. nrisk is the same for all individuals in a matched set).
#'
#' @inheritParams smcfcs
#' @example data-raw/ncc_example.r
#' @export
smcfcs.nestedcc <- function(originaldata, smformula, set, event, nrisk, method, predictorMatrix = NULL, m = 5, numit = 10, rjlimit = 1000, noisy = FALSE, errorProneMatrix = NULL) {
  smcfcs.core(originaldata,
    smtype = "nestedcc", smformula, method, predictorMatrix, m, numit, rjlimit, noisy, set = set, event = event, nrisk = nrisk,
    errorProneMatrix = errorProneMatrix
  )
}

#' Substantive model compatible fully conditional specification imputation of covariates for
#' discrete time survival analysis
#'
#' Multiply imputes missing covariate values using substantive model compatible
#' fully conditional specification for discrete time survival analysis.
#'
#' For this substantive model type, like for the other substantive model types, \code{smcfcs} expects the \code{originaldata} to have
#' one row per subject. Variables indicating the discrete time of failure/censoring
#' and the event indicator should be passed in \code{smformula}, as described.
#'
#' The default is to model the effect of time as a factor. This will not work in datasets where
#' there is not at least one observed event in each time period. In such cases you must specify
#' a simpler parametric model for the effect of time. At the moment you can specify either a linear or quadratic
#' effect of time (on the log odds scale).
#'
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @param originaldata The data in wide form (i.e. one row per subject)
#' @param smformula A formula of the form "Surv(t,d)~x1+x2+x3", where t is the discrete time variable, d is the binary event
#' indicator, and the covariates should not include time. The time variable should be
#' an integer coded numeric variable taking values from 1 up to the final time period.
#' @param timeEffects Specifies how the effect of time is modelled. \code{timeEffects="factor"} (the default) models time as a
#' factor variable. \code{timeEffects="linear"} and \code{timeEffects="quad"} specify that time be modelled as a continuous
#' linear or quadratic effect on the log odds scale respectively.
#'
#' @inheritParams smcfcs
#' @example data-raw/dtsam_example.r
#' @export
smcfcs.dtsam <- function(originaldata, smformula, timeEffects = "factor", method, predictorMatrix = NULL, m = 5, numit = 10, rjlimit = 1000, noisy = FALSE, errorProneMatrix = NULL) {
  smcfcs.core(originaldata,
    smtype = "dtsam", smformula, method, predictorMatrix, m, numit, rjlimit, noisy, timeEffects = timeEffects,
    errorProneMatrix = errorProneMatrix
  )
}

# this is the core of the smcfcs function, called by wrapper functions for certain different substantive models
smcfcs.core <- function(originaldata, smtype, smformula, method, predictorMatrix = NULL, m = 5, numit = 10, rjlimit = 1000, noisy = FALSE, errorProneMatrix = NULL,
                        ...) {
  # get extra arguments passed in ...
  extraArgs <- list(...)

  stopifnot(is.data.frame(originaldata))
  if (ncol(originaldata) != length(method)) stop("Method argument must have the same length as the number of columns in the data frame.")

  n <- dim(originaldata)[1]

  # create matrix of response indicators
  r <- 1 * (is.na(originaldata) == 0)

  if ((smtype %in% c(
    "lm", "logistic", "brlogistic", "poisson", "coxph", "compet", "casecohort", "nestedcc",
    "weibull", "dtsam", "flexsurv"
  )) == FALSE) {
    stop(paste("Substantive model type ", smtype, " not recognised.", sep = ""))
  }

  # find column numbers of partially observed, fully observed variables, and outcome
  if (smtype == "flexsurv") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    outcomeCol <- c(timeCol, dCol)
    d <- originaldata[, dCol]
    if (!(all(sort(unique(d)) == c(0, 1))) & !(all(unique(d) == 1))) {
      stop("Event indicator for flexsurv must be coded 0/1 for censoring/event.")
    }
    if ((sum(is.na(d))+sum(is.na(originaldata[,timeCol])))>0) {
      stop("Event indicator and time variables should not have NAs.")
    }
  } else if (smtype == "coxph") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    outcomeCol <- c(timeCol, dCol)
    d <- originaldata[, dCol]
    if (!(all(sort(unique(d)) == c(0, 1))) & !(all(unique(d) == 1))) {
      stop("Event indicator for coxph must be coded 0/1 for censoring/event.")
    }
    if ((sum(is.na(d))+sum(is.na(originaldata[,timeCol])))>0) {
      stop("Event indicator and time variables should not have NAs.")
    }

    nullMod <- survival::coxph(survival::Surv(originaldata[, timeCol], originaldata[, dCol]) ~ 1,
      control = survival::coxph.control(timefix = FALSE)
    )
    basehaz <- survival::basehaz(nullMod)
    H0indices <- match(originaldata[, timeCol], basehaz[, 2])
    rm(nullMod)
  } else if (smtype == "weibull") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    outcomeCol <- c(timeCol, dCol)
    d <- originaldata[, dCol]
  } else if (smtype == "dtsam") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    outcomeCol <- c(timeCol, dCol)
    d <- originaldata[, dCol]
    # determine cut points
    cutPoints <- 1:max(originaldata[, timeCol])
    nTimePoints <- length(cutPoints)

    # check all time points are integers
    if (!all(unique(originaldata[, timeCol]) == floor(unique(originaldata[, timeCol])))) {
      stop("Your time variable must only take positive integer values.")
    }
    # check all time points are positive
    if (any(unique(originaldata[, timeCol]) <= 0)) {
      stop("Your time variable must only take positive integer values.")
    }

    # if factor time effects, check there are some events at each integer value
    # so that model can fit
    if (extraArgs$timeEffects == "factor") {
      if (!identical(sort(unique(originaldata[, timeCol][originaldata[, dCol] == 1])), as.numeric(cutPoints))) {
        stop("You cannot fit a dtsam model with factor time effects since there are some periods with no events. See documentation for timeEffects argument for parametric alternatives")
      }
    }
  } else if (smtype == "compet") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula[[1]])[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula[[1]])[[2]][[3]][[2]])]
    outcomeCol <- c(timeCol, dCol)
    d <- originaldata[, dCol]
    numCauses <- length(smformula)
    H0 <- vector("list", numCauses)
    H0indices <- vector("list", numCauses)
    outcomeModBeta <- vector("list", numCauses)
    linpred <- vector("list", numCauses)
    for (cause in 1:numCauses) {
      nullMod <- survival::coxph(as.formula(paste(strsplit(smformula[[cause]], "~")[[1]][1], "~1")), originaldata,
        control = survival::coxph.control(timefix = FALSE)
      )
      basehaz <- survival::basehaz(nullMod)
      H0[[cause]] <- basehaz[, 1]
      H0indices[[cause]] <- match(originaldata[, timeCol], basehaz[, 2])
      linpred[[cause]] <- as.formula(smformula[[cause]])
    }
    rm(nullMod)
  } else if (smtype == "casecohort") {
    subcoCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% extraArgs$in.subco]
    # subcoMembers is a vector of row numbers of those in the subcohort
    subcoMembers <- which(originaldata[, subcoCol] == 1)

    # generate weights for use in later analysis which we use to obtain baseline cumulative hazard
    # assign a weight of /samp.frac to individuals in the subcohort and 0 to those outside the subcohort
    subco.weight <- ifelse(originaldata[, subcoCol] == 1, 1 / extraArgs$sampfrac, 0)


    entertimeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[4]])]
    outcomeCol <- c(entertimeCol, timeCol, dCol)
    d <- originaldata[, dCol]

    # list of unique event times - used in calculation of baseline cumulative hazard
    list.times <- sort(unique(originaldata[, timeCol][originaldata[, dCol] == 1])) # RUTH 21/03/17: ADDED THIS LINE

    smcfcsid <- 1:n
    smformula2 <- paste(smformula, "+cluster(smcfcsid)", sep = "")
  } else if (smtype == "nestedcc") {
    timeCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[2]])]
    dCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(as.formula(smformula)[[2]][[3]])]
    outcomeCol <- c(timeCol, dCol)
    setCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(extraArgs$set)]
    nriskCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(extraArgs$nrisk)]
    eventCol <- (1:dim(originaldata)[2])[colnames(originaldata) %in% toString(extraArgs$event)] # this is distinct from the dCol

    # the below command creates "Surv(t,case)~x" from "Surv(t,case)~x+strata(setno)" (for example) (i.e. it removes the strata part of the formula)
    # this is used when obtaining outmodxb
    # note this is done slightly oddly, but this is because as.formula does not work well for long formulas as it splits across lines
    exp1 <- as.formula(paste(smformula))[[2]]
    exp2 <- as.formula(smformula)[[3]][[2]]
    smformula2 <- paste(deparse(exp1), "~", deparse(exp2, width.cutoff = 500L))

    # This is the indicator of whether an individual ever has the event (regardless of whether they are sometimes used as a control and sometimes (one) as a case)
    d <- originaldata[, eventCol]

    # noncases is a vector of row numbers of those who never have the event (which is a subset of the controls)
    noncases <- which(originaldata[, eventCol] == 0)

    # number of individuals in each sampled risk set (matched set)
    num.sampriskset <- ave(rep(1, dim(originaldata)[1]), originaldata[, setCol], FUN = function(x) sum(x))
  } else {
    outcomeCol <- which(colnames(originaldata) == as.formula(smformula)[[2]])
  }

  if ((smtype == "logistic") | (smtype == "brlogistic")) {
    if (is.numeric(originaldata[, outcomeCol]) == FALSE) {
      stop("For logistic substantive models the outcome variable must be numeric 0/1.")
    } else {
      if (all.equal(sort(unique(originaldata[, outcomeCol])), c(0, 1)) == FALSE) {
        stop("For logistic substantive models the outcome variable must be coded 0/1.")
      }
    }
  }

  if (smtype == "compet") {
    smcovnames <- attr(terms(as.formula(smformula[[1]])), "term.labels")
    for (cause in 2:numCauses) {
      smcovnames <- c(smcovnames, attr(terms(as.formula(smformula[[cause]])), "term.labels"))
    }
    smcovnames <- unique(smcovnames)
  } else if (smtype == "nestedcc") {
    smcovnames <- attr(terms(as.formula(smformula)), "term.labels")[-length(attr(terms(as.formula(smformula)), "term.labels"))]
  } else {
    smcovnames <- attr(terms(as.formula(smformula)), "term.labels")
  }
  smcovcols <- (1:ncol(originaldata))[colnames(originaldata) %in% smcovnames]

  # partial vars are those variables for which an imputation method has been specified among the available regression types
  partialVars <- which((method == "norm") | (method == "latnorm") | (method == "logreg") | (method == "poisson") |
    (method == "podds") | (method == "mlogit") | (method == "brlogreg"))

  if (length(partialVars) == 0) {
    if (((smtype=="flexsurv") & (extraArgs$imputeTimes==TRUE)) == FALSE)
      stop("You have not specified any valid imputation methods in the method argument.")
  }

  # check that methods are given for each partially observed column, and not given for fully observed columns
  for (colnum in 1:ncol(originaldata)) {
    if (method[colnum] != "") {
      # an imputation method has been specified
      if (colnum %in% outcomeCol) {
        stop(paste("An imputation method has been specified for ", colnames(originaldata)[colnum],
          ". Elements of the method argument corresponding to the outcome variable(s) should be empty.",
          sep = ""
        ))
      } else {
        if (sum(r[, colnum]) == n) {
          stop(paste("An imputation method has been specified for ", colnames(originaldata)[colnum],
            ", but it appears to be fully observed.",
            sep = ""
          ))
        }
      }
    } else {
      # no imputation method has been specified
      if (sum(r[, colnum]) < n) {
        # some values are missing
        if (((colnum %in% outcomeCol) == FALSE) & (sum(errorProneMatrix[, colnum]) == 0)) {
          stop(paste("Variable ", colnames(originaldata)[colnum], " does not have an imputation method specified, yet appears to have missing values.", sep = ""))
        }
      }
    }
  }

  # check that if any variables have latnorm specified as method, that at least two error-prone measures
  # are specified for it in errorProneMatrix
  if ("latnorm" %in% method) {
    if (is.null(errorProneMatrix) == TRUE) {
      stop("If you specify method latnorm you must specify the errorProneMatrix argument.")
    } else {
      # check errorProneMatrix is of correct dimensional
      if (identical(dim(errorProneMatrix), c(length(originaldata), length(originaldata))) == FALSE) {
        stop("The errorProneMatrix should be a square matrix with number of rows equal to the number of variables in the dataset.")
      }
      # check entries of errorProneMatrix only consists of 0s and 1s
      if (identical(sort(unique(as.vector(errorProneMatrix))), c(0, 1)) == FALSE) {
        stop("The errorProneMatrix should only consist of 0s and 1s.")
      }
      # check each latnorm variable has at least 2 error-prones
      for (varNum in 1:length(method)) {
        if (method[varNum] == "latnorm") {
          if (sum(errorProneMatrix[varNum, ]) < 2) {
            stop("Each latnorm variable must have two or more error prone measurements specified in the errorProneMatrix argument.")
          }
        }
      }
      # check no error-prone measurement is allocated to more than one latnorm
      if (sum(colSums(errorProneMatrix) > 1) > 0) {
        stop("Each error-prone measurement should be allocated to exactly one latnorm variable.")
      }
    }
  }
  if (is.null(errorProneMatrix) == FALSE) {
    if (("latnorm" %in% method) == FALSE) {
      stop("If you specify errorProneMatrix then at least one variable must be imputed using latnorm.")
    }
  }

  # fully observed vars are those that are fully observed and are covariates in the substantive model
  fullObsVars <- which((colSums(r) == n) & (colnames(originaldata) %in% smcovnames))

  # passive variables
  passiveVars <- which((method != "") & (method != "norm") & (method != "logreg") & (method != "poisson") & (method != "podds") &
    (method != "mlogit") & (method != "latnorm") & (method != "brlogreg"))

  print(paste("Outcome variable(s):", paste(colnames(originaldata)[outcomeCol], collapse = ",")))
  print(paste("Passive variables:", paste(colnames(originaldata)[passiveVars], collapse = ",")))
  print(paste("Partially obs. variables:", paste(colnames(originaldata)[partialVars], collapse = ",")))
  print(paste("Fully obs. substantive model variables:", paste(colnames(originaldata)[fullObsVars], collapse = ",")))

  imputations <- list()
  for (imp in 1:m) {
    imputations[[imp]] <- originaldata
  }

  rjFailCount <- 0
  flexsurvFailCount <- 0

  for (imp in 1:m) {
    print(paste("Imputation ", imp))

    # initial imputation of each partially observed variable based on observed values
    if (length(partialVars)>0) {
      for (var in 1:length(partialVars)) {
        targetCol <- partialVars[var]
        if (method[targetCol] == "latnorm") {
          # first impute any missing replicate error-prone measurements of this variable by a randomly chosen observed value
          errorProneCols <- which(errorProneMatrix[targetCol, ] == 1)
          for (measure in 1:length(errorProneCols)) {
            if (sum(r[, errorProneCols[measure]]) < n) {
              imputations[[imp]][r[, errorProneCols[measure]] == 0, errorProneCols[measure]] <- sample(imputations[[imp]][
                r[, errorProneCols[measure]] == 1,
                errorProneCols[measure]
              ], size = sum(r[, errorProneCols[measure]] == 0), replace = TRUE)
            }
          }

          # initialize latent predictors with mean of their error-prone measurements
          imputations[[imp]][, targetCol] <- apply(imputations[[imp]][, errorProneCols], 1, mean)
        } else {
          imputations[[imp]][r[, targetCol] == 0, targetCol] <- sample(imputations[[imp]][r[, targetCol] == 1, targetCol], size = sum(r[, targetCol] == 0), replace = TRUE)
        }
      }
    }

    # initial imputations of missing outcomes, if present (using improper imputation)
    if ((smtype == "lm") | (smtype == "logistic") | (smtype == "brlogistic") | (smtype == "poisson")) {
      if (sum(r[, outcomeCol]) < n) {
        if (imp == 1) {
          print("Imputing missing outcomes using specified substantive model.")
        }
        # update passive variable(s)
        imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)

        imputationNeeded <- (1:n)[r[, outcomeCol] == 0]
        # estimate parameters of substantive model
        if (smtype == "lm") {
          ymod <- stats::lm(as.formula(smformula), imputations[[imp]])
          beta <- ymod$coef
          sigmasq <- summary(ymod)$sigma^2
          # fill out missing values so that model.matrix works for all rows
          imputations[[imp]][imputationNeeded, outcomeCol] <- 0
          outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% beta
          imputations[[imp]][imputationNeeded, outcomeCol] <- rnorm(length(imputationNeeded), outmodxb[imputationNeeded], sigmasq^0.5)
        } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
          if (smtype == "logistic") {
            ymod <- glm(as.formula(smformula), family = "binomial", imputations[[imp]])
          } else {
            ymod <- glm(as.formula(smformula),
              family = "binomial", imputations[[imp]],
              method = brglm2::brglmFit
            )
          }
          beta <- ymod$coef
          imputations[[imp]][imputationNeeded, outcomeCol] <- 0
          outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% beta
          prob <- expit(outmodxb[imputationNeeded])
          imputations[[imp]][imputationNeeded, outcomeCol] <- rbinom(length(imputationNeeded), 1, prob)
        } else if (smtype == "poisson") {
          ymod <- glm(as.formula(smformula), family = "poisson", imputations[[imp]])
          beta <- ymod$coef
          imputations[[imp]][imputationNeeded, outcomeCol] <- 0
          outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% beta
          imputations[[imp]][imputationNeeded, outcomeCol] <- rpois(length(imputationNeeded), exp(outmodxb[imputationNeeded]))
        }
      }
    }

    for (cyclenum in 1:numit) {
      if (noisy == TRUE) {
        print(paste("Iteration ", cyclenum))
      }
      # update passive variable(s)
      imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)

      if (length(partialVars)>0) {
        for (var in 1:length(partialVars)) {
        targetCol <- partialVars[var]
        if (is.null(predictorMatrix)) {
          predictorCols <- c(partialVars[!partialVars %in% targetCol], fullObsVars)
        } else {
          predictorCols <- which(predictorMatrix[targetCol, ] == 1)
          # ensure that user has not included outcome variable(s) here
          predictorCols <- predictorCols[!predictorCols %in% outcomeCol]
        }
        if ((imp == 1) & (cyclenum == 1)) {
          if (method[targetCol] == "latnorm") {
            print(paste("Imputing: ", colnames(imputations[[imp]])[targetCol], " using ", paste(colnames(imputations[[imp]])[c(predictorCols, which(errorProneMatrix[targetCol, ] == 1))], collapse = ","), " plus outcome", collapse = ","))
          } else {
            print(paste("Imputing: ", colnames(imputations[[imp]])[targetCol], " using ", paste(colnames(imputations[[imp]])[predictorCols], collapse = ","), " plus outcome", collapse = ","))
          }
        }
        if (length(predictorCols) > 0) {
          xmodformula <- as.formula(paste(colnames(imputations[[imp]])[targetCol], "~", paste(colnames(imputations[[imp]])[predictorCols], collapse = "+"), sep = ""))
        } else {
          xmodformula <- as.formula(paste(colnames(imputations[[imp]])[targetCol], "~1", sep = ""))
        }
        if (smtype == "casecohort") {
          xmoddata <- imputations[[imp]][subcoMembers, ]
        } else if (smtype == "nestedcc") {
          xmoddata <- imputations[[imp]][noncases, ]
        } else {
          xmoddata <- imputations[[imp]]
        }
        if (method[targetCol] == "norm") {
          # estimate parameters of covariate model
          xmod <- lm(xmodformula, data = xmoddata)
          # take draw from posterior of covariate model parameters
          beta <- xmod$coef
          sigmasq <- summary(xmod)$sigma^2
          newsigmasq <- (sigmasq * xmod$df) / rchisq(1, xmod$df)
          covariance <- (newsigmasq / sigmasq) * vcov(xmod)
          newbeta <- beta + MASS::mvrnorm(1, mu = rep(0, ncol(covariance)), Sigma = covariance)
          # calculate fitted values
          if ((smtype == "casecohort") | (smtype == "nestedcc")) {
            xfitted <- model.matrix(xmodformula, data = imputations[[imp]]) %*% newbeta
          } else {
            xfitted <- model.matrix(xmod) %*% newbeta
          }
        } else if (method[targetCol] == "latnorm") {
          # estimate parameters of covariate model
          xmod <- lm(xmodformula, data = xmoddata)
          # take draw from posterior of covariate model parameters
          beta <- xmod$coef
          sigmasq <- summary(xmod)$sigma^2
          # draw from sigmasq posterior based on proper inverse gamma prior for sigmasq
          # prior equivalent to 1 observation and guess of sigmasq=1
          newsigmasq <- 1 / rgamma(1, shape = ((n + 1) / 2), rate = ((n * sigmasq + 1) / 2))
          covariance <- (newsigmasq / sigmasq) * vcov(xmod)
          newbeta <- beta + MASS::mvrnorm(1, mu = rep(0, ncol(covariance)), Sigma = covariance)
          # calculate fitted values
          if ((smtype == "casecohort") | (smtype == "nestedcc")) {
            xfitted <- model.matrix(xmodformula, data = imputations[[imp]]) %*% newbeta
          } else {
            xfitted <- model.matrix(xmod) %*% newbeta
          }

          # estimate error variance and draw new value of error variance
          errorProneCols <- which(errorProneMatrix[targetCol, ] == 1)
          xmat <- matrix(imputations[[imp]][, targetCol], nrow = nrow(imputations[[imp]]), ncol = length(errorProneCols))
          uVec <- c(as.matrix(imputations[[imp]][, errorProneCols] - xmat))
          sigmausq <- mean(uVec^2)
          # take draw from posterior of error variance, using proper inverse gamma prior
          sum_ni <- n * length(errorProneCols)
          sigmausq <- 1 / rgamma(1, shape = ((sum_ni + 1) / 2), rate = ((sum_ni * sigmausq + 1) / 2))

          # re-impute any originally missing error-prone measurements, based on classical error model assumption
          for (measure in 1:length(errorProneCols)) {
            nToImpute <- n - sum(r[, errorProneCols[measure]])
            if (nToImpute > 0) {
              # then some values need imputing
              imputations[[imp]][r[, errorProneCols[measure]] == 0, errorProneCols[measure]] <- imputations[[imp]][r[, errorProneCols[measure]] == 0, targetCol] +
                rnorm(nToImpute, 0, sd = sqrt(sigmausq))
            }
          }

          # calculate conditional mean and variance of X|everything else except outcome
          wmean <- rowMeans(imputations[[imp]][, errorProneCols])
          lambda <- newsigmasq / (newsigmasq + sigmausq / length(errorProneCols))
          xfitted <- xfitted + lambda * (wmean - xfitted)
          newsigmasq <- rep(newsigmasq * (1 - lambda), n)
        } else if (method[targetCol] == "logreg") {
          xmod <- glm(xmodformula, family = "binomial", data = xmoddata)
          newbeta <- modPostDraw(xmod)
          if ((smtype == "casecohort") | (smtype == "nestedcc")) {
            xfitted <- expit(model.matrix(xmodformula, data = imputations[[imp]]) %*% newbeta)
          } else {
            xfitted <- expit(model.matrix(xmod) %*% newbeta)
          }
        } else if (method[targetCol] == "brlogreg") {
          xmod <- glm(xmodformula, family = "binomial", data = xmoddata, method = brglm2::brglmFit)
          newbeta <- modPostDraw(xmod)
          if ((smtype == "casecohort") | (smtype == "nestedcc")) {
            xfitted <- expit(model.matrix(xmodformula, data = imputations[[imp]]) %*% newbeta)
          } else {
            xfitted <- expit(model.matrix(xmod) %*% newbeta)
          }
        } else if (method[targetCol] == "poisson") {
          xmod <- glm(xmodformula, family = "poisson", data = xmoddata)
          newbeta <- modPostDraw(xmod)
          if ((smtype == "casecohort") | (smtype == "nestedcc")) {
            xfitted <- exp(model.matrix(xmodformula, data = imputations[[imp]]) %*% newbeta)
          } else {
            xfitted <- exp(model.matrix(xmod) %*% newbeta)
          }
        } else if (method[targetCol] == "podds") {
          if (is.ordered(imputations[[imp]][, targetCol]) == FALSE) stop("Variables to be imputed using method podds must be stored as ordered factors.")
          xmod <- VGAM::vglm(xmodformula, VGAM::propodds, data = xmoddata)
          xmod.dummy <- VGAM::vglm(xmodformula, VGAM::propodds, data = imputations[[imp]])
          newbeta <- VGAM::coef(xmod) + MASS::mvrnorm(1, mu = rep(0, ncol(VGAM::vcov(xmod))), Sigma = VGAM::vcov(xmod))
          linpreds <- matrix((VGAM::model.matrix(xmod.dummy)) %*% newbeta, byrow = TRUE, ncol = (nlevels(imputations[[imp]][, targetCol]) - 1))
          cumprobs <- cbind(1 / (1 + exp(linpreds)), rep(1, nrow(linpreds)))
          xfitted <- cbind(cumprobs[, 1], cumprobs[, 2:ncol(cumprobs)] - cumprobs[, 1:(ncol(cumprobs) - 1)])
        } else if (method[targetCol] == "mlogit") {
          if (is.factor(imputations[[imp]][, targetCol]) == FALSE) stop("Variables to be imputed using method modds must be stored as factors.")
          xmod <- VGAM::vglm(xmodformula, VGAM::multinomial(refLevel = 1), data = xmoddata)
          xmod.dummy <- VGAM::vglm(xmodformula, VGAM::multinomial(refLevel = 1), data = imputations[[imp]])
          newbeta <- VGAM::coef(xmod) + MASS::mvrnorm(1, mu = rep(0, ncol(VGAM::vcov(xmod))), Sigma = VGAM::vcov(xmod))
          linpreds <- matrix((VGAM::model.matrix(xmod.dummy)) %*% newbeta, byrow = TRUE, ncol = (nlevels(imputations[[imp]][, targetCol]) - 1))
          denom <- 1 + rowSums(exp(linpreds))
          xfitted <- cbind(1 / denom, exp(linpreds) / denom)
        }
        if (noisy == TRUE) {
          print(summary(xmod))
        }

        # estimate parameters of substantive model
        if (smtype == "lm") {
          ymod <- lm(as.formula(smformula), imputations[[imp]])
          beta <- ymod$coef
          sigmasq <- summary(ymod)$sigma^2
          varcov <- vcov(ymod)
          outcomeModResVar <- (sigmasq * ymod$df) / rchisq(1, ymod$df)
          covariance <- (outcomeModResVar / sigmasq) * vcov(ymod)
          outcomeModBeta <- beta + MASS::mvrnorm(1, mu = rep(0, ncol(covariance)), Sigma = covariance)
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
          if (smtype == "logistic") {
            ymod <- glm(as.formula(smformula), family = "binomial", imputations[[imp]])
          } else {
            ymod <- glm(as.formula(smformula),
              family = "binomial", imputations[[imp]],
              method = brglm2::brglmFit
            )
          }
          outcomeModBeta <- modPostDraw(ymod)
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "dtsam") {
          # split data to long form
          longData <- survival::survSplit(as.formula(smformula), data = imputations[[imp]], cut = cutPoints)
          # fit logistic model
          if (extraArgs$timeEffects == "factor") {
            dtsamFormula <- paste(colnames(imputations[[imp]])[dCol], "~-1+factor(tstart)+",
              strsplit(smformula, "~")[[1]][2],
              sep = ""
            )
          } else if (extraArgs$timeEffects == "linear") {
            # linear time effect
            dtsamFormula <- paste(colnames(imputations[[imp]])[dCol], "~tstart+",
              strsplit(smformula, "~")[[1]][2],
              sep = ""
            )
          } else {
            # quadratic effect of time
            dtsamFormula <- paste(colnames(imputations[[imp]])[dCol], "~tstart+I(tstart^2)+",
              strsplit(smformula, "~")[[1]][2],
              sep = ""
            )
          }
          ymod <- glm(as.formula(dtsamFormula), family = "binomial", data = longData)
          outcomeModBeta <- modPostDraw(ymod)
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "poisson") {
          ymod <- glm(as.formula(smformula), family = "poisson", imputations[[imp]])
          outcomeModBeta <- modPostDraw(ymod)
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "coxph") {
          ymod <- survival::coxph(as.formula(smformula), imputations[[imp]],
            control = survival::coxph.control(timefix = FALSE), model = TRUE
          )
          outcomeModBeta <- modPostDraw(ymod)
          ymod$coefficients <- outcomeModBeta
          basehaz <- survival::basehaz(ymod, centered = FALSE)[, 1]
          H0 <- basehaz[H0indices]
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "flexsurv") {
          if (cyclenum==1) {
            tryCatch({
              ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                    k=extraArgs$k, scale="hazard")
            },error = function(e) {
              # Silently increment the failure counter
              flexsurvFailCount <<- flexsurvFailCount + 1
            })
          } else {
            # use previous estimates as initial values, to try and prevent non-convergence
            tryCatch(
              {
                # first attempt to fit model
                if (extraArgs$originalKnots==FALSE) {
                  ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                                   k=extraArgs$k, scale="hazard",
                                                   inits=flexsurvEsts)
                } else {
                  ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                                   scale="hazard",
                                                   knots=ymod$knots[2:(length(ymod$knots)-1)],
                                                   inits=flexsurvEsts)
                }
              },
              error = function(e) {
                # Silently increment the failure counter
                flexsurvFailCount <<- flexsurvFailCount + 1
              }
              )
          }
          # save parameter estimates to use as initial values in subsequent iterations
          # to help prevent convergence problems
          flexsurvEsts <- ymod$res.t[,1]
          outcomeModBeta <- as.numeric(flexsurv::normboot.flexsurvreg(ymod, B=1, raw=TRUE))
          # overwrite model estimates in fitted model object
          ymod$res.t[,1] <- outcomeModBeta
          if (noisy == TRUE) {
            print(ymod)
          }
        } else if (smtype == "weibull") {
          ymod <- survival::survreg(as.formula(smformula), data = imputations[[imp]], dist = "weibull")
          outcomeModBeta <- c(coef(ymod), log(ymod$scale)) +
            MASS::mvrnorm(1, mu = rep(0, ncol(vcov(ymod))), Sigma = vcov(ymod))
          weibullScale <- exp(utils::tail(outcomeModBeta, 1))
          outcomeModBeta <- utils::head(outcomeModBeta, -1)
          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "compet") {
          for (cause in 1:numCauses) {
            ymod <- survival::coxph(as.formula(smformula[[cause]]), imputations[[imp]],
              control = survival::coxph.control(timefix = FALSE), model = TRUE
            )
            outcomeModBeta[[cause]] <- modPostDraw(ymod)
            ymod$coefficients <- outcomeModBeta[[cause]]
            basehaz <- survival::basehaz(ymod, centered = FALSE)[, 1]
            H0[[cause]] <- basehaz[H0indices[[cause]]]
            if (noisy == TRUE) {
              print(summary(ymod))
            }
          }
        } else if (smtype == "casecohort") {
          ymod <- survival::coxph(as.formula(smformula2), imputations[[imp]],
            control = survival::coxph.control(timefix = FALSE)
          )
          outcomeModBeta <- modPostDraw(ymod)

          cumhaz.denom.elements <- exp(model.matrix(as.formula(smformula), imputations[[imp]])[, -1] %*% outcomeModBeta)
          cumhaz.denom <- sapply(list.times, function(x) {
            sum(cumhaz.denom.elements[which(originaldata[, timeCol] >= x)] * subco.weight[which(originaldata[, timeCol] >= x)])
          })
          exp.func.denom <- cumsum(1 / cumhaz.denom)
          H0.fun <- stepfun(list.times, c(0, exp.func.denom))
          H0 <- H0.fun(originaldata[, timeCol])

          if (noisy == TRUE) {
            print(summary(ymod))
          }
        } else if (smtype == "nestedcc") {
          ymod <- survival::coxph(as.formula(smformula), imputations[[imp]],
            control = survival::coxph.control(timefix = FALSE)
          )
          outcomeModBeta <- modPostDraw(ymod)

          explan.matrix <- model.matrix(ymod)
          cumbasehaz.denom <- exp(matrix(outcomeModBeta, nrow = 1) %*% t(explan.matrix)) * originaldata[, nriskCol] / num.sampriskset
          cumbasehaz.denom <- ave(cumbasehaz.denom, originaldata[, setCol], FUN = sum)[originaldata[, dCol] == 1] # this is the denominator of the contribution to the cumulative baseline hazard at each event time
          cumbasehaz.t <- originaldata[, timeCol][originaldata[, dCol] == 1] # times to which the baseline cumulative hazards refer
          H0 <- unlist(lapply(originaldata[, timeCol], function(x) {
            sum((1 / cumbasehaz.denom)[cumbasehaz.t <= x])
          }))

          if (noisy == TRUE) {
            print(summary(ymod))
          }
        }

        if ((imp == 1) & (cyclenum == 1) & (var == 1)) {
          if (smtype == "compet") {
            totalCoefVec <- outcomeModBeta[[1]]
            for (cause in 2:numCauses) {
              totalCoefVec <- c(totalCoefVec, outcomeModBeta[[cause]])
            }
            smCoefIter <- array(0, dim = c(m, length(totalCoefVec), numit))
          } else {
            smCoefIter <- array(0, dim = c(m, length(outcomeModBeta), numit))
          }
        }

        if (var == length(partialVars)) {
          # then we have reached end of a cycle
          if (smtype == "compet") {
            totalCoefVec <- outcomeModBeta[[1]]
            for (cause in 2:numCauses) {
              totalCoefVec <- c(totalCoefVec, outcomeModBeta[[cause]])
            }
            smCoefIter[imp, , cyclenum] <- totalCoefVec
          } else {
            smCoefIter[imp, , cyclenum] <- outcomeModBeta
          }
        }

        # impute x, either directly where possibly, or using rejection sampling otherwise
        imputationNeeded <- (1:n)[r[, targetCol] == 0]

        if ((method[targetCol] == "logreg") | (method[targetCol] == "podds") | (method[targetCol] == "mlogit") |
          (method[targetCol] == "brlogreg")) {
          # directly sample
          if ((method[targetCol] == "logreg") | (method[targetCol] == "brlogreg")) {
            numberOutcomes <- 2
            fittedMean <- cbind(1 - xfitted, xfitted)
          } else {
            numberOutcomes <- nlevels(imputations[[imp]][, targetCol])
            fittedMean <- xfitted
          }

          outcomeDensCovDens <- array(dim = c(length(imputationNeeded), numberOutcomes), 0)

          for (xMisVal in 1:numberOutcomes) {
            if ((method[targetCol] == "logreg") | (method[targetCol] == "brlogreg")) {
              if (is.factor(imputations[[imp]][, targetCol]) == TRUE) {
                valToImpute <- levels(imputations[[imp]][, targetCol])[xMisVal]
              } else {
                valToImpute <- xMisVal - 1
              }
            } else {
              valToImpute <- levels(imputations[[imp]][, targetCol])[xMisVal]
            }
            imputations[[imp]][imputationNeeded, targetCol] <- valToImpute

            # update passive variables
            imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)

            if (smtype == "lm") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              deviation <- imputations[[imp]][imputationNeeded, outcomeCol] - outmodxb[imputationNeeded]
              outcomeDens <- dnorm(deviation, mean = 0, sd = outcomeModResVar^0.5)
            } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              prob <- expit(outmodxb[imputationNeeded])
              outcomeDens <- prob * imputations[[imp]][imputationNeeded, outcomeCol] + (1 - prob) * (1 - imputations[[imp]][imputationNeeded, outcomeCol])
            } else if (smtype == "dtsam") {
              outcomeDens <- dtsamOutcomeDens(
                imputations[[imp]], extraArgs$timeEffects, outcomeModBeta, nTimePoints,
                smformula, timeCol, dCol
              )[imputationNeeded]
            } else if (smtype == "poisson") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              outcomeDens <- dpois(imputations[[imp]][imputationNeeded, outcomeCol], exp(outmodxb[imputationNeeded]))
            } else if (smtype == "weibull") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              # weibull survival function
              outcomeDens <- (1 - d[imputationNeeded]) * (1 - survival::psurvreg(imputations[[imp]][imputationNeeded, timeCol], mean = outmodxb[imputationNeeded], scale = weibullScale)) +
                d[imputationNeeded] * (survival::dsurvreg(imputations[[imp]][imputationNeeded, timeCol], mean = outmodxb[imputationNeeded], scale = weibullScale))
            } else if ((smtype == "coxph") | (smtype == "casecohort")) {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]])
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              outcomeDens <- exp(-H0[imputationNeeded] * exp(outmodxb[imputationNeeded])) * (exp(outmodxb[imputationNeeded])^d[imputationNeeded])
            } else if (smtype == "flexsurv") {
              survEst <- summary(ymod, newdata=imputations[[imp]][imputationNeeded,], type="survival",
                                 ci=FALSE, t=imputations[[imp]][imputationNeeded,timeCol],
                                 cross=FALSE, tidy=TRUE)
              survEst <- as.matrix(survEst)[,"est"]
              hazEst <- summary(ymod, newdata=imputations[[imp]][imputationNeeded,], type="hazard",
                                ci=FALSE, t=imputations[[imp]][imputationNeeded,timeCol],
                                cross=FALSE, tidy=TRUE)
              hazEst <- as.matrix(hazEst)[,"est"]
              outcomeDens <- survEst*(hazEst^imputations[[imp]][imputationNeeded,dCol])
            } else if (smtype == "nestedcc") {
              outmodxb <- model.matrix(as.formula(smformula2), imputations[[imp]])
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              outcomeDens <- exp(-H0[imputationNeeded] * exp(outmodxb[imputationNeeded])) * (exp(outmodxb[imputationNeeded])^d[imputationNeeded])
            } else if (smtype == "compet") {
              outcomeDens <- rep(1, length(imputationNeeded))
              for (cause in 1:numCauses) {
                outmodxb <- model.matrix(linpred[[cause]], imputations[[imp]])
                outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta[[cause]])
                outcomeDens <- outcomeDens * exp(-H0[[cause]][imputationNeeded] * exp(outmodxb[imputationNeeded])) * (exp(outmodxb[imputationNeeded])^(d[imputationNeeded] == cause))
              }
            }
            outcomeDensCovDens[, xMisVal] <- outcomeDens * fittedMean[imputationNeeded, xMisVal]
          }
          directImpProbs <- outcomeDensCovDens / rowSums(outcomeDensCovDens)

          if ((method[targetCol] == "logreg") | (method[targetCol] == "brlogreg")) {
            directImpProbs <- directImpProbs[, 2]
            if (is.factor(imputations[[imp]][, targetCol]) == TRUE) {
              imputations[[imp]][imputationNeeded, targetCol] <- levels(imputations[[imp]][, targetCol])[1]
              imputations[[imp]][imputationNeeded, targetCol][rbinom(length(imputationNeeded), 1, directImpProbs) == 1] <- levels(imputations[[imp]][, targetCol])[2]
            } else {
              imputations[[imp]][imputationNeeded, targetCol] <- rbinom(length(imputationNeeded), 1, directImpProbs)
            }
          } else {
            imputations[[imp]][imputationNeeded, targetCol] <- levels(imputations[[imp]][, targetCol])[apply(directImpProbs, 1, catdraw)]
          }

          # update passive variables
          imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)
        } else {
          # use rejection sampling
          # first draw for all subjects who need imputing, using a small number of attempts
          firstTryLimit <- 25
          j <- 1

          while ((length(imputationNeeded) > 0) & (j < firstTryLimit)) {
            # sample from covariate model
            if ((method[targetCol] == "norm") | (method[targetCol] == "latnorm")) {
              imputations[[imp]][imputationNeeded, targetCol] <- rnorm(length(imputationNeeded), xfitted[imputationNeeded], newsigmasq^0.5)
            } else if (method[targetCol] == "poisson") {
              imputations[[imp]][imputationNeeded, targetCol] <- rpois(length(imputationNeeded), xfitted[imputationNeeded])
            }

            # update passive variables
            imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)

            # accept/reject
            uDraw <- runif(length(imputationNeeded))
            if (smtype == "lm") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              deviation <- imputations[[imp]][imputationNeeded, outcomeCol] - outmodxb[imputationNeeded]
              reject <- 1 * (log(uDraw) > -(deviation^2) / (2 * array(outcomeModResVar, dim = c(length(imputationNeeded), 1))))
            } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              prob <- expit(outmodxb[imputationNeeded])
              prob <- prob * imputations[[imp]][imputationNeeded, outcomeCol] + (1 - prob) * (1 - imputations[[imp]][imputationNeeded, outcomeCol])
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "dtsam") {
              prob <- dtsamOutcomeDens(
                imputations[[imp]], extraArgs$timeEffects, outcomeModBeta, nTimePoints,
                smformula, timeCol, dCol
              )[imputationNeeded]
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "poisson") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              prob <- dpois(imputations[[imp]][imputationNeeded, outcomeCol], exp(outmodxb[imputationNeeded]))
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "weibull") {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
              s_t <- 1 - survival::psurvreg(imputations[[imp]][imputationNeeded, timeCol], mean = outmodxb[imputationNeeded], scale = weibullScale)
              prob <- -exp(1) * log(s_t) * s_t
              prob <- d[imputationNeeded] * prob + (1 - d[imputationNeeded]) * s_t
              reject <- 1 * (uDraw > prob)
            } else if ((smtype == "coxph") | (smtype == "casecohort")) {
              outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]])
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              s_t <- exp(-H0[imputationNeeded] * exp(outmodxb[imputationNeeded]))
              prob <- exp(1 + outmodxb[imputationNeeded] - (H0[imputationNeeded] * exp(outmodxb[imputationNeeded]))) * H0[imputationNeeded]
              prob <- d[imputationNeeded] * prob + (1 - d[imputationNeeded]) * s_t
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "flexsurv") {
              survEst <- summary(ymod, newdata=imputations[[imp]][imputationNeeded,], type="survival",
                                 ci=FALSE, t=imputations[[imp]][imputationNeeded,timeCol],
                                 cross=FALSE, tidy=TRUE)
              survEst <- as.matrix(survEst)[,"est"]
              hazEst <- summary(ymod, newdata=imputations[[imp]][imputationNeeded,], type="hazard",
                                ci=FALSE, t=imputations[[imp]][imputationNeeded,timeCol],
                                cross=FALSE, tidy=TRUE)
              hazEst <- as.matrix(hazEst)[,"est"]
              cumhazEst <- summary(ymod, newdata=imputations[[imp]][imputationNeeded,], type="cumhaz",
                                ci=FALSE, t=imputations[[imp]][imputationNeeded,timeCol],
                                cross=FALSE, tidy=TRUE)
              cumhazEst <- as.matrix(cumhazEst)[,"est"]
              prob <- imputations[[imp]][imputationNeeded,dCol] * (survEst*exp(1)*cumhazEst) +
                (1 - imputations[[imp]][imputationNeeded,dCol]) * survEst
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "nestedcc") {
              outmodxb <- model.matrix(as.formula(smformula2), imputations[[imp]])
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              s_t <- exp(-H0[imputationNeeded] * exp(outmodxb[imputationNeeded]))
              prob <- exp(1 + outmodxb[imputationNeeded] - (H0[imputationNeeded] * exp(outmodxb[imputationNeeded]))) * H0[imputationNeeded]
              prob <- d[imputationNeeded] * prob + (1 - d[imputationNeeded]) * s_t
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "compet") {
              prob <- rep(1, length(imputationNeeded))
              for (cause in 1:numCauses) {
                outmodxb <- model.matrix(linpred[[cause]], imputations[[imp]])
                outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta[[cause]])
                prob <- prob * exp(-H0[[cause]][imputationNeeded] * exp(outmodxb[imputationNeeded])) * (H0[[cause]][imputationNeeded] * exp(1 + outmodxb[imputationNeeded]))^(d[imputationNeeded] == cause)
              }
              reject <- 1 * (uDraw > prob)
            }
            imputationNeeded <- imputationNeeded[reject == 1]

            j <- j + 1
          }

          # now, for those remaining, who must have low acceptance probabilities, sample by subject
          for (i in imputationNeeded) {
            tempData <- imputations[[imp]][i, ]
            tempData <- tempData[rep(1, rjlimit), ]
            if (method[targetCol] == "norm") {
              tempData[, targetCol] <- rnorm(rjlimit, xfitted[i], newsigmasq^0.5)
            } else if ((method[targetCol] == "logreg") | (method[targetCol] == "brlogreg")) {
              tempData[, targetCol] <- rbinom(rjlimit, size = 1, xfitted[i])
            } else if (method[targetCol] == "poisson") {
              tempData[, targetCol] <- rpois(rjlimit, xfitted[i])
            } else if (method[targetCol] == "latnorm") {
              tempData[, targetCol] <- rnorm(rjlimit, xfitted[i], newsigmasq[i]^0.5)
            }

            # passively impute
            tempData <- updatePassiveVars(tempData, method, passiveVars)

            # accept reject
            uDraw <- runif(rjlimit)
            if (smtype == "lm") {
              outmodxb <- model.matrix(as.formula(smformula), tempData) %*% outcomeModBeta
              deviation <- tempData[, outcomeCol] - outmodxb
              reject <- 1 * (log(uDraw) > -(deviation^2) / (2 * array(outcomeModResVar, dim = c(rjlimit, 1))))
            } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
              outmodxb <- model.matrix(as.formula(smformula), tempData) %*% outcomeModBeta
              prob <- expit(outmodxb)
              prob <- prob * tempData[, outcomeCol] + (1 - prob) * (1 - tempData[, outcomeCol])
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "dtsam") {
              prob <- dtsamOutcomeDens(
                tempData, extraArgs$timeEffects, outcomeModBeta, nTimePoints,
                smformula, timeCol, dCol
              )
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "poisson") {
              outmodxb <- model.matrix(as.formula(smformula), tempData) %*% outcomeModBeta
              prob <- dpois(tempData[, outcomeCol], exp(outmodxb))
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "weibull") {
              outmodxb <- model.matrix(as.formula(smformula), tempData) %*% outcomeModBeta
              s_t <- 1 - survival::psurvreg(tempData[, timeCol], mean = outmodxb, scale = weibullScale)
              if (d[i] == 1) {
                prob <- -exp(1) * log(s_t) * s_t
                # the following line fixes a numerical error which occurs if s_t=0 for any draws
                prob[is.na(prob)] <- 0
              } else {
                prob <- s_t
              }
              reject <- 1 * (uDraw > prob)
            } else if ((smtype == "coxph") | (smtype == "casecohort")) {
              outmodxb <- model.matrix(as.formula(smformula), tempData)
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              s_t <- exp(-H0[i] * exp(outmodxb))
              prob <- exp(1 + outmodxb - (H0[i] * exp(outmodxb))) * H0[i]
              prob <- d[i] * prob + (1 - d[i]) * s_t
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "flexsurv") {
              survEst <- summary(ymod, newdata=tempData, type="survival",
                                 ci=FALSE, t=tempData[,timeCol],
                                 cross=FALSE, tidy=TRUE)
              survEst <- as.matrix(survEst)[,"est"]
              hazEst <- summary(ymod, newdata=tempData, type="hazard",
                                ci=FALSE, t=tempData[,timeCol],
                                cross=FALSE, tidy=TRUE)
              hazEst <- as.matrix(hazEst)[,"est"]
              cumhazEst <- summary(ymod, newdata=tempData, type="cumhaz",
                                   ci=FALSE, t=tempData[,timeCol],
                                   cross=FALSE, tidy=TRUE)
              cumhazEst <- as.matrix(cumhazEst)[,"est"]
              prob <- imputations[[imp]][i,dCol] * (survEst*exp(1)*cumhazEst) +
                (1 - imputations[[imp]][i,dCol]) * survEst
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "nestedcc") {
              outmodxb <- model.matrix(as.formula(smformula2), tempData)
              outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta)
              s_t <- exp(-H0[i] * exp(outmodxb))
              prob <- exp(1 + outmodxb - (H0[i] * exp(outmodxb))) * H0[i]
              prob <- d[i] * prob + (1 - d[i]) * s_t
              reject <- 1 * (uDraw > prob)
            } else if (smtype == "compet") {
              prob <- rep(1, rjlimit)
              for (cause in 1:numCauses) {
                outmodxb <- model.matrix(linpred[[cause]], tempData)
                outmodxb <- as.matrix(outmodxb[, 2:dim(outmodxb)[2]]) %*% as.matrix(outcomeModBeta[[cause]])
                prob <- prob * exp(-H0[[cause]][i] * exp(outmodxb)) * (H0[[cause]][i] * exp(1 + outmodxb))^(d[i] == cause)
              }
              reject <- 1 * (uDraw > prob)
            }
            if (sum(reject) < rjlimit) {
              imputations[[imp]][i, targetCol] <- tempData[reject == 0, targetCol][1]
            } else {
              rjFailCount <- rjFailCount + 1
            }
          }
          # update passive variables
          imputations[[imp]] <- updatePassiveVars(imputations[[imp]], method, passiveVars)
        }
      }
      }

      # imputations of missing outcomes, if present (using proper imputation), for regression and logistic
      # substantive models
      if ((smtype == "lm") | (smtype == "logistic") | (smtype == "brlogistic")) {
        if (sum(r[, outcomeCol]) < n) {
          imputationNeeded <- (1:n)[r[, outcomeCol] == 0]
          # estimate parameters of substantive model using those with outcomes observed
          if (smtype == "lm") {
            ymod <- lm(as.formula(smformula), imputations[[imp]][r[, outcomeCol] == 1, ])
            beta <- ymod$coef
            sigmasq <- summary(ymod)$sigma^2
            varcov <- vcov(ymod)
            outcomeModResVar <- (sigmasq * ymod$df) / rchisq(1, ymod$df)
            covariance <- (outcomeModResVar / sigmasq) * vcov(ymod)
            outcomeModBeta <- beta + MASS::mvrnorm(1, mu = rep(0, ncol(covariance)), Sigma = covariance)
            outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
            imputations[[imp]][imputationNeeded, outcomeCol] <- rnorm(length(imputationNeeded), outmodxb[imputationNeeded], sigmasq^0.5)
          } else if ((smtype == "logistic") | (smtype == "brlogistic")) {
            if (smtype == "logistic") {
              ymod <- glm(as.formula(smformula), family = "binomial", imputations[[imp]][r[, outcomeCol] == 1, ])
            } else {
              ymod <- glm(as.formula(smformula),
                family = "binomial", imputations[[imp]][r[, outcomeCol] == 1, ],
                method = brglm2::brglmFit
              )
            }
            outcomeModBeta <- modPostDraw(ymod)
            outmodxb <- model.matrix(as.formula(smformula), imputations[[imp]]) %*% outcomeModBeta
            prob <- expit(outmodxb[imputationNeeded])
            imputations[[imp]][imputationNeeded, outcomeCol] <- rbinom(length(imputationNeeded), 1, prob)
          }
        }
      } else if (smtype == "flexsurv") {
        if (extraArgs$imputeTimes==TRUE) {
          # impute censored times
          if (cyclenum==1) {
            tryCatch({
              ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                               k=extraArgs$k, scale="hazard")
            },error = function(e) {
              # Silently increment the failure counter
              flexsurvFailCount <<- flexsurvFailCount + 1
            })
          } else {
            # use previous estimates as initial values, to try and prevent non-convergence
            tryCatch(
              {
                if (extraArgs$originalKnots==FALSE) {
                  ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                                 k=extraArgs$k, scale="hazard",
                                                 inits=flexsurvEsts)
                } else {
                  # use knots based on fit to original dataset where censored times had not been imputed
                  ymod <- flexsurv::flexsurvspline(as.formula(smformula), imputations[[imp]],
                                                   scale="hazard",
                                                   knots=ymod$knots[2:(length(ymod$knots)-1)],
                                                   inits=flexsurvEsts)
                }
              },
              error = function(e) {
                flexsurvFailCount <<- flexsurvFailCount + 1
              }
            )
          }
          flexsurvEsts <- ymod$res.t[,1]
          outcomeModBeta <- as.numeric(flexsurv::normboot.flexsurvreg(ymod, B=1, raw=TRUE))

          if (exists("smCoefIter")==FALSE) {
            smCoefIter <- array(0, dim = c(m, length(outcomeModBeta), numit))
          }
          smCoefIter[imp, , cyclenum] <- outcomeModBeta

          # overwrite model estimates in fitted model object
          ymod$res.t[,1] <- outcomeModBeta
          if (noisy == TRUE) {
            print(ymod)
          }

          # impute censored times
          if (is.null(extraArgs$censtime)) {
            # impute all censored times
            timeImp <- simulate(ymod, nsim=1, newdata=imputations[[imp]][d==0,],
                              start=originaldata[d==0,timeCol])
            imputations[[imp]][d==0,timeCol] <- timeImp$time_1
            imputations[[imp]][,dCol] <- 1
          } else {
            # impute but still impose censoring at specified value(s)
            timeImp <- simulate(ymod, nsim=1, newdata=imputations[[imp]][d==0,],
                                start=originaldata[d==0,timeCol],
                                censtime=extraArgs$censtime)
            imputations[[imp]][d==0,timeCol] <- timeImp$time_1
            imputations[[imp]][d==0,dCol] <- timeImp$event_1
          }
        }
      }
    }
  }

  if (rjFailCount > 0) {
    warning(paste("Rejection sampling failed ", rjFailCount, " times (across all variables, iterations, and imputations). You may want to increase the rejection sampling limit.", sep = ""))
  }
  if (flexsurvFailCount > 0) {
    warning(paste("Flexsurv fit failed ", flexsurvFailCount, " times (across all variables, iterations, and imputations). See documentation for more details.", sep = ""))
  }

  # Added smformula and smtype to metadata, and make "smcfcs class"
  res <- list(
    impDatasets = imputations,
    smCoefIter = smCoefIter,
    smInfo = list("smtype" = smtype, "smformula" = smformula),
    extraArgs = extraArgs # For plot of dtsam
  )
  class(res) <- "smcfcs"

  return(res)
}

updatePassiveVars <- function(data, method, passivecols) {
  for (i in passivecols) {
    data[, i] <- with(data, eval(parse(text = method[i])))
  }
  data
}

expit <- function(x) {
  exp(x) / (1 + exp(x))
}

sumna <- function(x) {
  sum(is.na(x) == FALSE)
}

# returns first non missing entry of x
firstnonna <- function(x) {
  x[is.na(x) == FALSE][1]
}

catdraw <- function(prob) {
  (1:length(prob))[rmultinom(1, size = 1, prob = prob) == 1]
}

modPostDraw <- function(modobj) {
  beta <- modobj$coef
  varcov <- vcov(modobj)
  beta + MASS::mvrnorm(1, mu = rep(0, ncol(varcov)), Sigma = varcov)
}

# a helper function which calculates the substantive model probability value
# for each individual for the discrete time survival (logistic) substantive model
dtsamOutcomeDens <- function(inputData, timeEffects, outcomeModBeta, nTimePoints, smformula, timeCol, dCol) {
  inputDataN <- dim(inputData)[1]

  # calculate covariate effects
  modelMatrix <- model.matrix(
    as.formula(paste("~", strsplit(smformula, "~")[[1]][2], sep = "")),
    inputData
  )
  # remove intercept column
  modelMatrix <- modelMatrix[,2:ncol(modelMatrix)]

  if (timeEffects == "factor") {
    # first add in time effects on log odds scale
    outmodxb <- matrix(outcomeModBeta[1:nTimePoints], nrow = inputDataN, ncol = nTimePoints, byrow = TRUE)
    covXbEffects <- modelMatrix %*% utils::tail(outcomeModBeta, length(outcomeModBeta) - nTimePoints)
  } else if (timeEffects == "linear") {
    # linear time
    # first add in time effects on log odds scale
    outmodxb <- outcomeModBeta[1] +
      matrix(outcomeModBeta[2] * ((1:nTimePoints) - 1), nrow = inputDataN, ncol = nTimePoints, byrow = TRUE)
    # calculate covariate effects
    covXbEffects <- modelMatrix %*% utils::tail(outcomeModBeta, length(outcomeModBeta) - 2)
  } else {
    # quadratic time
    # first add in time effects on log odds scale
    outmodxb <- outcomeModBeta[1] +
      matrix(outcomeModBeta[2] * ((1:nTimePoints) - 1) + outcomeModBeta[3] * (((1:nTimePoints) - 1)^2), nrow = inputDataN, ncol = nTimePoints, byrow = TRUE)
    # calculate covariate effects
    covXbEffects <- modelMatrix %*% utils::tail(outcomeModBeta, length(outcomeModBeta) - 3)
  }

  # add in covariate effects
  outmodxb <- outmodxb + matrix(covXbEffects, nrow = inputDataN, ncol = nTimePoints)
  # prob is matrix of conditional probabilities/hazard of event in each period
  prob <- expit(outmodxb)
  logSurvProb <- log(1 - prob)
  logSurvProbCumSum <- cbind(rep(0, inputDataN), t(apply(logSurvProb, 1, cumsum)))

  # create vector of last time point each person survived to, +1
  lastSurvPlusOne <- 1 + inputData[, timeCol] - inputData[, dCol]
  # create vector of log of probability of survival to time when each person actually survived to
  logSurvProbIndividual <- logSurvProbCumSum[cbind(1:inputDataN, lastSurvPlusOne)]
  # return vector of outcome density values
  exp(logSurvProbIndividual + inputData[, dCol] * log(prob[cbind(1:inputDataN, inputData[, timeCol])]))
}

library(smcfcs)
library(survival)
context("Cox proportional hazards model testing")

test_that("Cox imputation is approximately unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 10000
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs(simData,
        smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", "")
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, coxph(Surv(t, d) ~ x + z))
      abs(summary(MIcombine(models))[1, 1] - 1) < 0.1
    },
    TRUE
  )
})

test_that("Cox imputation works with only one covariate", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 10000
      x <- rnorm(n)
      t <- -log(runif(n)) / (0.01 * exp(x))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x)

      imps <- smcfcs(simData,
        smtype = "coxph", smformula = "Surv(t, d)~x",
        method = c("", "", "norm")
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, coxph(Surv(t, d) ~ x))
      abs(summary(MIcombine(models))[1, 1] - 1) < 0.1
    },
    TRUE
  )
})

library(smcfcs)
context("Testing imputation with podds using polr for ordered factors")

test_that("podds using polr runs, one covariate in polr model", {
  expect_error(
    {
      set.seed(1234)
      n <- 100

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      y <- z+(x=="b")-(x=="c")+2*(x=="d")+rnorm(n)

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.ordered(x),z=z,y=y)

      imps <- smcfcs(originaldata=simData,
        smtype="lm",
        smformula = "y~z+I(x=='b')+I(x=='c')+I(x=='d')",
        method = c("podds","", ""),
      )
    },
    NA
  )

})

test_that("podds using polr runs, two covariates in polr model", {
  expect_error(
    {
      set.seed(1234)
      n <- 100

      z1 <- rnorm(n)
      z2 <- z1+rnorm(n)
      xtilde <- 2*z1-z2 + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      y <- z1+z2+(x=="b")-(x=="c")+2*(x=="d")+rnorm(n)

      x[1*runif(n)<(exp(z1)/(1+exp(z1)))] <- NA

      simData <- data.frame(x=as.ordered(x),z1=z1,z2=z2,y=y)

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z1+z2+I(x=='b')+I(x=='c')+I(x=='d')",
                     method = c("podds","","","")
      )
    },
    NA
  )
})


test_that("podds using polr is unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(2347633)
      n <- 100000

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      y <- z+(x=="b")-(x=="c")+2*(x=="d")+rnorm(n)

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.ordered(x),z=z,y=y)

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+I(x=='b')+I(x=='c')+I(x=='d')",
                     method = c("podds","", ""),
      )

      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, lm(y ~ z+I(x=='b')+I(x=='c')+I(x=='d')))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[5] < 2) & (MIcombineRes$`upper)`[5]>2)
    },
    TRUE
  )

})


test_that("podds using polr with two covariates is unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1762348012)
      n <- 100000

      z1 <- rnorm(n)
      z2 <- z1+rnorm(n)
      xtilde <- 2*z1-z2 + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      y <- z1+z2+(x=="b")-(x=="c")+2*(x=="d")+rnorm(n)

      x[1*runif(n)<(exp(z1)/(1+exp(z1)))] <- NA

      simData <- data.frame(x=as.ordered(x),z1=z1,z2=z2,y=y)

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z1+z2+I(x=='b')+I(x=='c')+I(x=='d')",
                     method = c("podds","","","")
      )

      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, lm(y ~ z1+z2+I(x=='b')+I(x=='c')+I(x=='d')))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[6] < 2) & (MIcombineRes$`upper)`[6]>2)
    },
    TRUE
  )
})


test_that("podds using polr runs with flexsurv, one covariate in polr model", {
  expect_error(
    {
      set.seed(1234)
      n <- 100

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      t <- rexp(n)
      d <- 1*(t<0.5)
      t[d==0] <- 0.5

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.ordered(x),z=z,t=t,d=d)

      imps <- smcfcs.flexsurv(originaldata=simData,
                     smformula = "Surv(t,d)~z+I(x=='b')+I(x=='c')+I(x=='d')",
                     method = c("podds","", "", "")
      )
    },
    NA
  )

})


test_that("podds using polr runs with flexsurv, not included with I()", {
  expect_error(
    {
      set.seed(1234)
      n <- 100

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1.5)) & (xtilde<0.25)] <- "b"
      x[(xtilde>(0.25)) & (xtilde<0.75)] <- "c"
      x[(xtilde>(0.75)) ] <- "d"
      table(x)

      t <- rexp(n)
      d <- 1*(t<0.5)
      t[d==0] <- 0.5

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.ordered(x),z=z,t=t,d=d)

      imps <- smcfcs.flexsurv(originaldata=simData,
                              smformula = "Surv(t,d)~z+x",
                              method = c("podds","", "", ""),
      )
    },
    NA
  )

})

library(smcfcs)
library(survival)
library(mitools)
context("Case cohort testing")

test_that("Case cohort imputation runs and is approximately unbiased, binary covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)

      # run simulation study
      nSims <- 100
      xEsts <- array(0, dim = nSims)
      n <- 10000

      for (sim in 1:nSims) {
        print(sim)
        z <- rnorm(n)
        x <- 1 * (runif(n) < (exp(z) / (1 + exp(z))))
        t <- -log(runif(n)) / (0.01 * exp(x + z))
        d <- 1 * (t < 2)
        t[d == 0] <- 2
        x[(runif(n) < 0.5)] <- NA

        fullcohortdata <- data.frame(t, d, x, z)
        fullcohortdata$in.subco <- 0
        # we sample a 10% subcohort
        fullcohortdata$in.subco[sample(n, size = n * 0.1)] <- 1
        fullcohortdata$id <- 1:n

        ccdata <- fullcohortdata[(fullcohortdata$in.subco == 1) | (fullcohortdata$d == 1), ]
        ccdata$entertime <- 0
        ccdata$entertime[ccdata$in.subco == 0] <- ccdata$t[ccdata$in.subco == 0] - 0.000001
        imps <- smcfcs.casecohort(ccdata,
          smformula = "Surv(entertime, t, d)~x+z", sampfrac = 0.1,
          in.subco = "in.subco", method = c("", "", "logreg", "", "", "", "")
        )
        impobj <- imputationList(imps$impDatasets)
        models <- with(impobj, coxph(Surv(entertime, t, d) ~ x + z + cluster(id)))
        xEsts[sim] <- summary(MIcombine(models))[1, 1]
      }
      print(mean(xEsts))
      abs(mean(xEsts - 1)) < 0.1
    },
    TRUE
  )
})

test_that("Case cohort imputation runs and is approximately unbiased, continuous covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)

      # run simulation study
      nSims <- 100
      xEsts <- array(0, dim = nSims)
      n <- 10000

      for (sim in 1:nSims) {
        print(sim)
        z <- rnorm(n)
        x <- z + rnorm(n)
        t <- -log(runif(n)) / (0.01 * exp(x + z))
        d <- 1 * (t < 1)
        t[d == 0] <- 1
        x[(runif(n) < 0.5)] <- NA

        fullcohortdata <- data.frame(t, d, x, z)
        fullcohortdata$in.subco <- 0
        # we sample a 10% subcohort
        fullcohortdata$in.subco[sample(n, size = n * 0.1)] <- 1
        fullcohortdata$id <- 1:n

        ccdata <- fullcohortdata[(fullcohortdata$in.subco == 1) | (fullcohortdata$d == 1), ]
        ccdata$entertime <- 0
        ccdata$entertime[ccdata$in.subco == 0] <- ccdata$t[ccdata$in.subco == 0] - 0.000001
        imps <- smcfcs.casecohort(ccdata,
          smformula = "Surv(entertime, t, d)~x+z", sampfrac = 0.1,
          in.subco = "in.subco", method = c("", "", "norm", "", "", "", "")
        )
        impobj <- imputationList(imps$impDatasets)
        models <- with(impobj, coxph(Surv(entertime, t, d) ~ x + z + cluster(id)))
        xEsts[sim] <- summary(MIcombine(models))[1, 1]
      }
      print(mean(xEsts))
      abs(mean(xEsts - 1)) < 0.1
    },
    TRUE
  )
})

library(smcfcs)
library(survival)
library(mitools)
context("Covariate measurement error testing")

test_that("Linear regression with cov. measurement error runs", {
  expect_error(
    {
      set.seed(1234)

      n <- 100 # sample size
      x <- rnorm(n)
      w1 <- x + rnorm(n)
      w2 <- x + rnorm(n)
      y <- x + rnorm(n)
      x <- rep(NA, n)

      simData <- data.frame(x, w1, w2, y)

      errMat <- matrix(0, nrow = 4, ncol = 4)
      errMat[1, c(2, 3)] <- 1

      imps <- smcfcs(simData,
        smtype = "lm", smformula = "y~x",
        method = c("latnorm", "", "", ""),
        errorProneMatrix = errMat, numit = 100, m = 1
      )
    },
    NA
  )
})

test_that("Linear regression with cov. measurement error is consistent", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)

      n <- 10000 # sample size
      x <- rnorm(n)
      w1 <- x + rnorm(n)
      w2 <- x + rnorm(n)
      y <- x + rnorm(n)
      x <- rep(NA, n)

      simData <- data.frame(x, w1, w2, y)

      errMat <- matrix(0, nrow = 4, ncol = 4)
      errMat[1, c(2, 3)] <- 1

      imps <- smcfcs(simData,
        smtype = "lm", smformula = "y~x",
        method = c("latnorm", "", "", ""),
        errorProneMatrix = errMat, numit = 100, m = 1
      )
      as.logical(abs(coef(lm(y ~ x, data = imps$impDatasets[[1]]))[2] - 1) < 0.05)
    },
    TRUE
  )
})

test_that("Linear regression with cov. measurement error coverage is ok", {
  skip_on_cran()
  expect_equal(
    {
      library(mitools)
      set.seed(1234)
      nSim <- 1000
      n <- 1000
      ests <- array(0, dim = c(nSim, 2))

      for (i in 1:nSim) {
        x <- rnorm(n)
        w1 <- x + rnorm(n)
        w2 <- x + rnorm(n)
        w2[round(n * 0.1):n] <- NA
        y <- x + rnorm(n)
        x <- rep(NA, n)

        simData <- data.frame(x, w1, w2, y)

        errMat <- matrix(0, nrow = 4, ncol = 4)
        errMat[1, c(2, 3)] <- 1

        imps <- smcfcs(
          originaldata = simData,
          smtype = "lm",
          smformula = "y~x",
          method = c("latnorm", "", "", ""),
          errorProneMatrix = errMat, numit = 100, m = 5
        )
        impobj <- imputationList(imps$impDatasets)
        modelSummary <- summary(MIcombine(with(impobj, lm(y ~ x))))
        ests[i, ] <- c(modelSummary[2, 3], modelSummary[2, 4])
      }
      # check coverage is close to 95%
      abs(mean((ests[, 1] < 1) & (ests[, 2] > 1)) - 0.95) < (qnorm(0.999) * ((0.95 * 0.05) / nSim)^0.5)
    },
    TRUE
  )
})

library(smcfcs)
library(survival)
library(mitools)
context("Nested case control testing")

test_that("Nested case control imputation runs and is approximately unbiased, continuous covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)

      # run simulation study
      nSims <- 100
      xEsts <- array(0, dim = nSims)
      n <- 10000

      for (sim in 1:nSims) {
        print(sim)
        z <- rnorm(n)
        x <- z + rnorm(n)
        t <- -log(runif(n)) / (0.01 * exp(x + z))
        d <- 1 * (t < 1)
        t[d == 0] <- 1
        x[(runif(n) < 0.5)] <- NA

        fullcohortdata <- data.frame(t, d, x, z)
        fullcohortdata$id <- 1:n

        # Compute number at risk at each event time using the full cohort data
        nrisk.fit <- survfit(Surv(t, d) ~ 1, data = fullcohortdata)
        ord.t.d1 <- order(fullcohortdata$t[fullcohortdata$d == 1])

        m <- 1 # 1 control per case
        ncc <- NULL

        no.sample <- 0
        for (i in which(fullcohortdata$d == 1))
        {
          # select control(s) for nested case-control
          possible.controls <- which(fullcohortdata$t >= fullcohortdata$t[i])
          # remove the case from this vector
          possible.controls <- possible.controls[which(possible.controls != i)]

          if (length(possible.controls) >= m) {
            controls <- sample(possible.controls, m)
            numAtRisk <- 1 + length(possible.controls)
            ncc <- rbind(ncc, cbind(fullcohortdata[i, ], numrisk = numAtRisk))
            ncc <- rbind(ncc, cbind(fullcohortdata[controls, ], numrisk = numAtRisk))
            no.sample <- no.sample + 1
          }
        }

        ncc$setno <- rep(1:no.sample, each = m + 1)
        ncc$case <- rep(c(1, rep(0, m)), no.sample)

        predictorMatrix <- matrix(0, nrow = dim(ncc)[2], ncol = dim(ncc)[2])
        predictorMatrix[which(colnames(ncc) == "x"), c(which(colnames(ncc) == "z"))] <- 1

        imps <- smcfcs.nestedcc(
          originaldata = ncc, set = "setno", nrisk = "numrisk", event = "d", smformula = "Surv(t,case)~x+z+strata(setno)",
          method = c("", "", "norm", "", "", "", "", ""), predictorMatrix = predictorMatrix
        )

        impobj <- imputationList(imps$impDatasets)
        models <- with(impobj, clogit(case ~ x + z + strata(setno)))

        xEsts[sim] <- summary(MIcombine(models))[1, 1]
      }
      print(mean(xEsts))
      abs(mean(xEsts) - 1) < 0.1
    },
    TRUE
  )
})

test_that("Nested case control imputation runs and is approximately unbiased, binary covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(5678)

      # perform simulation study
      nSims <- 100
      xEsts <- array(0, dim = nSims)
      n <- 10000

      for (sim in 1:nSims) {
        print(sim)
        z <- rnorm(n)
        x <- rbinom(n, 1, exp(z) / (1 + exp(z)))
        t <- -log(runif(n)) / (0.01 * exp(x + z))
        d <- 1 * (t < 1)
        t[d == 0] <- 1
        x[(runif(n) < 0.5)] <- NA

        fullcohortdata <- data.frame(t, d, x, z)
        fullcohortdata$id <- 1:n

        # Compute number at risk at each event time using the full cohort data
        nrisk.fit <- survfit(Surv(t, d) ~ 1, data = fullcohortdata)
        ord.t.d1 <- order(fullcohortdata$t[fullcohortdata$d == 1])

        m <- 1 # 1 control per case
        ncc <- NULL

        no.sample <- 0
        for (i in which(fullcohortdata$d == 1))
        {
          # select control(s) for nested case-control
          possible.controls <- which(fullcohortdata$t >= fullcohortdata$t[i])
          # remove the case from this vector
          possible.controls <- possible.controls[which(possible.controls != i)]

          if (length(possible.controls) >= m) {
            controls <- sample(possible.controls, m)
            numAtRisk <- 1 + length(possible.controls)
            ncc <- rbind(ncc, cbind(fullcohortdata[i, ], numrisk = numAtRisk))
            ncc <- rbind(ncc, cbind(fullcohortdata[controls, ], numrisk = numAtRisk))
            no.sample <- no.sample + 1
          }
        }

        ncc$setno <- rep(1:no.sample, each = m + 1)
        ncc$case <- rep(c(1, rep(0, m)), no.sample)

        predictorMatrix <- matrix(0, nrow = dim(ncc)[2], ncol = dim(ncc)[2])
        predictorMatrix[which(colnames(ncc) == "x"), c(which(colnames(ncc) == "z"))] <- 1

        imps <- smcfcs.nestedcc(
          originaldata = ncc, set = "setno", nrisk = "numrisk", event = "d", smformula = "Surv(t,case)~x+z+strata(setno)",
          method = c("", "", "logreg", "", "", "", "", ""), predictorMatrix = predictorMatrix
        )

        impobj <- imputationList(imps$impDatasets)
        models <- with(impobj, clogit(case ~ x + z + strata(setno)))

        xEsts[sim] <- summary(MIcombine(models))[1, 1]
      }
      print(mean(xEsts))
      abs(mean(xEsts) - 1) < 0.1
    },
    TRUE
  )
})

library(smcfcs)
context("Testing imputation with restrictions for handling coarsening")

test_that("Restrictions runs and respects restrictions with coarsened factor covariate", {
  expect_equal(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)

      p1 <- 1/(1+exp(0+z)+exp(1+z))
      p2 <- exp(0+z)/(1+exp(0+z)+exp(1+z))
      p3 <- 1-p1-p2
      u <- runif(n)
      x <- rep("a",n)
      x[(u>p1) & (u<(p1+p2))] <- "b"
      x[(u>(p1+p2))] <- "c"

      y <- z+(x=="b")+2*(x=="c")+rnorm(n)

      # generate coarsening variable, where 0 indicates no coarsening
      # 1 indicates missing
      # 2 indicates coarsened variable=a if x=a or b/c if x=b or x=c
      # 3 indicates coarsened variable=b if x=b or a/c if x=a or x=c
      c <- sample(c(0, 1, 2, 3), size = n, replace = TRUE, prob = rep(0.25,4))
      xobs <- x
      xobs[c==1] <- "NA"
      xobs[(c==2) & ((x=="b") | (x=="c"))] <- "b/c"
      xobs[(c==3) & ((x=="a") | (x=="c"))] <- "a/c"
      table(xobs, useNA = "always")

      # set x to NA when not exactly observed
      x[c==1] <- NA
      x[(c==2) & ((x=="b") | (x=="c"))] <- NA
      x[(c==3) & ((x=="a") | (x=="c"))] <- NA

      simData <- data.frame(x=factor(x),xobs=xobs,z=z,y=y)

      restrictionsX = c("xobs = a/c ~ a + c",
                        "xobs = b/c ~ b + c")
      restrictions = append(list(restrictionsX), as.list(c("", "", "")))

      imps <- smcfcs(originaldata=simData,
        smtype="lm",
        smformula = "y~z+x",
        method = c("mlogit","", "", ""),
        restrictions = restrictions
      )

      # check imputations respect restrictions
      fail <- 0
      for (i in 1:length(imps$impDatasets)) {
        rowsToCheck <- subset(imps$impDatasets[[i]], xobs %in% c("a/c", "b/c"))
        rowsToCheck$valid <- with(rowsToCheck,
                               ifelse(xobs == "a/c", x %in% c("a", "c"),
                                      ifelse(xobs == "b/c", x %in% c("b", "c"), NA)))
        fail <- fail + sum(!rowsToCheck$valid)
      }
      fail
    },
    0
  )

})


test_that("Restrictions fails if argument has wrong length", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)

      p1 <- 1/(1+exp(0+z)+exp(1+z))
      p2 <- exp(0+z)/(1+exp(0+z)+exp(1+z))
      p3 <- 1-p1-p2
      u <- runif(n)
      x <- rep("a",n)
      x[(u>p1) & (u<(p1+p2))] <- "b"
      x[(u>(p1+p2))] <- "c"

      y <- z+(x=="b")+2*(x=="c")+rnorm(n)

      # generate coarsening variable, where 0 indicates no coarsening
      # 1 indicates missing
      # 2 indicates coarsened variable=a if x=a or b/c if x=b or x=c
      # 3 indicates coarsened variable=b if x=b or a/c if x=a or x=c
      c <- sample(c(0, 1, 2, 3), size = n, replace = TRUE, prob = rep(0.25,4))
      xobs <- x
      xobs[c==1] <- "NA"
      xobs[(c==2) & ((x=="b") | (x=="c"))] <- "b/c"
      xobs[(c==3) & ((x=="a") | (x=="c"))] <- "a/c"
      table(xobs, useNA = "always")

      # set x to NA when not exactly observed
      x[c==1] <- NA
      x[(c==2) & ((x=="b") | (x=="c"))] <- NA
      x[(c==3) & ((x=="a") | (x=="c"))] <- NA

      simData <- data.frame(x=factor(x),xobs=xobs,z=z,y=y)

      restrictionsX = c("xobs = a/c ~ a + c",
                        "xobs = b/c ~ b + c")
      restrictions = append(list(restrictionsX), as.list(c("", "", "", "")))

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+x",
                     method = c("mlogit","", "", ""),
                     restrictions = restrictions
      )

    }
  )

})



test_that("Restrictions is unbiased with coarsened factor covariate", {
  skip_on_cran()
  expect_equal({
      set.seed(79968)
      n <- 100000
      z <- rnorm(n)

      p1 <- 1/(1+exp(0+z)+exp(1+z))
      p2 <- exp(0+z)/(1+exp(0+z)+exp(1+z))
      p3 <- 1-p1-p2
      u <- runif(n)
      x <- rep("a",n)
      x[(u>p1) & (u<(p1+p2))] <- "b"
      x[(u>(p1+p2))] <- "c"

      y <- z+(x=="b")+2*(x=="c")+rnorm(n)

      # generate coarsening variable, where 0 indicates no coarsening
      # 1 indicates missing
      # 2 indicates coarsened variable=a if x=a or b/c if x=b or x=c
      # 3 indicates coarsened variable=b if x=b or a/c if x=a or x=c
      c <- sample(c(0, 1, 2, 3), size = n, replace = TRUE, prob = rep(0.25,4))
      xobs <- x
      xobs[c==1] <- "NA"
      xobs[(c==2) & ((x=="b") | (x=="c"))] <- "b/c"
      xobs[(c==3) & ((x=="a") | (x=="c"))] <- "a/c"
      table(xobs, useNA = "always")

      # set x to NA when not exactly observed
      x[c==1] <- NA
      x[(c==2) & ((x=="b") | (x=="c"))] <- NA
      x[(c==3) & ((x=="a") | (x=="c"))] <- NA

      simData <- data.frame(x=factor(x),xobs=xobs,z=z,y=y)

      restrictionsX = c("xobs = a/c ~ a + c",
                        "xobs = b/c ~ b + c")
      restrictions = append(list(restrictionsX), as.list(c("", "", "")))

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+x",
                     method = c("mlogit","", "", ""),
                     restrictions = restrictions
      )

      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, lm(y ~ z+x))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[4] < 2) & (MIcombineRes$`upper)`[4]>2)
    },
    TRUE
  )

})


test_that("Restrictions runs respecting restrictions with coarsened ordered factor covariate", {
  expect_equal(
    {
      set.seed(1234)
      n <- 100

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1)) & (xtilde<1)] <- "b"
      x[(xtilde>(1)) ] <- "c"
      table(x)

      y <- z+(x=="b")+2*(x=="c")+rnorm(n)

      # generate coarsening variable, where 0 indicates no coarsening
      # 1 indicates missing
      # 2 indicates coarsened variable=a if x=a or b/c if x=b or x=c
      # 3 indicates coarsened variable=b if x=b or a/c if x=a or x=c
      c <- sample(c(0, 1, 2, 3), size = n, replace = TRUE, prob = rep(0.25,4))
      xobs <- x
      xobs[c==1] <- "NA"
      xobs[(c==2) & ((x=="b") | (x=="c"))] <- "b/c"
      xobs[(c==3) & ((x=="a") | (x=="c"))] <- "a/c"
      table(xobs, useNA = "always")

      # set x to NA when not exactly observed
      x[c==1] <- NA
      x[(c==2) & ((x=="b") | (x=="c"))] <- NA
      x[(c==3) & ((x=="a") | (x=="c"))] <- NA

      simData <- data.frame(x=as.ordered(x),xobs=xobs,z=z,y=y)

      restrictionsX = c("xobs = a/c ~ a + c",
                        "xobs = b/c ~ b + c")
      restrictions = append(list(restrictionsX), as.list(c("", "", "")))

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+x",
                     method = c("podds","", "", ""),
                     restrictions = restrictions
      )

      # check imputations respect restrictions
      fail <- 0
      for (i in 1:length(imps$impDatasets)) {
        rowsToCheck <- subset(imps$impDatasets[[i]], xobs %in% c("a/c", "b/c"))
        rowsToCheck$valid <- with(rowsToCheck,
                                  ifelse(xobs == "a/c", x %in% c("a", "c"),
                                         ifelse(xobs == "b/c", x %in% c("b", "c"), NA)))
        fail <- fail + sum(!rowsToCheck$valid)
      }
      fail
    },
    0
  )

})



test_that("Restrictions is unbiased with coarsened ordered factor covariate", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(79968)
      n <- 100000

      z <- rnorm(n)
      xtilde <- 2*z + rlogis(n)

      x <- rep("a",n)
      x[(xtilde>(-1)) & (xtilde<1)] <- "b"
      x[(xtilde>(1)) ] <- "c"
      table(x)

      y <- z+(x=="b")+2*(x=="c")+rnorm(n)

      # generate coarsening variable, where 0 indicates no coarsening
      # 1 indicates missing
      # 2 indicates coarsened variable=a if x=a or b/c if x=b or x=c
      # 3 indicates coarsened variable=b if x=b or a/c if x=a or x=c
      c <- sample(c(0, 1, 2, 3), size = n, replace = TRUE, prob = rep(0.25,4))
      xobs <- x
      xobs[c==1] <- "NA"
      xobs[(c==2) & ((x=="b") | (x=="c"))] <- "b/c"
      xobs[(c==3) & ((x=="a") | (x=="c"))] <- "a/c"
      table(xobs, useNA = "always")

      # set x to NA when not exactly observed
      x[c==1] <- NA
      x[(c==2) & ((x=="b") | (x=="c"))] <- NA
      x[(c==3) & ((x=="a") | (x=="c"))] <- NA

      simData <- data.frame(x=as.ordered(x),xobs=xobs,z=z,y=y)

      restrictionsX = c("xobs = a/c ~ a + c",
                        "xobs = b/c ~ b + c")
      restrictions = append(list(restrictionsX), as.list(c("", "", "")))

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+x",
                     method = c("podds","", "", ""),
                     restrictions = restrictions
      )

      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, lm(y ~ z+I(x=="b")+I(x=="c")))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[4] < 2) & (MIcombineRes$`upper)`[4]>2)
    },
    TRUE
  )

})

library(smcfcs)
library(survival)
context("Error trap testing")

test_that("Checking outcome model check for logistic models", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    y <- as.factor(1 * (runif(n) < 0.5))
    x[(runif(n) < 0.5)] <- NA

    simData <- data.frame(x, y)

    imps <- smcfcs(simData,
      smtype = "logistic", smformula = "y~x",
      method = c("norm", "")
    )
  })
})

test_that("Checking outcome model check for logistic models", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    y <- 1 + 1 * (runif(n) < 0.5)
    x[(runif(n) < 0.5)] <- NA

    simData <- data.frame(x, y)

    imps <- smcfcs(simData,
      smtype = "logistic", smformula = "y~x",
      method = c("norm", "")
    )
  })
})

test_that("Checking outcome model check for logistic models", {
  expect_output({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    y <- 1 * (runif(n) < 0.5)
    x[(runif(n) < 0.5)] <- NA

    simData <- data.frame(x, y)

    imps <- smcfcs(simData,
      smtype = "logistic", smformula = "y~x",
      method = c("norm", "")
    )
  })
})

test_that("Checking error trap for method statement 1", {
  expect_error({
    set.seed(1234)
    n <- 100
    x1 <- rnorm(n)
    x2 <- rnorm(n)
    y <- y <- x1 + x2 + rnorm(n)
    x1[(runif(n) < 0.5)] <- NA

    simData <- data.frame(x1, x2, y)

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x1+x2",
      method = c("", "", "")
    )
  })
})

test_that("Checking error trap for method statement 2", {
  expect_error({
    set.seed(1234)
    n <- 100
    x1 <- rnorm(n)
    x2 <- rnorm(n)
    y <- y <- x1 + x2 + rnorm(n)
    x1[(runif(n) < 0.5)] <- NA

    simData <- data.frame(x1, x2, y)

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x1+x2",
      method = c("", "norm", "")
    )
  })
})

test_that("Checking measurement error error checks 1", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    w2 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, w2, y)

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", "", "")
    )
  })
})

test_that("Checking measurement error error checks 2", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, y)
    errMat <- array(0, dim = c(3, 3))

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", ""), errorProneMatrix = errMat
    )
  })
})

test_that("Checking measurement error error checks 3", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, y)
    errMat <- array(0, dim = c(3, 3))
    errMat[1, 1] <- 1

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", ""), errorProneMatrix = errMat
    )
  })
})

test_that("Checking measurement error error checks 4", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    w2 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, w2, y)
    errMat <- array(0, dim = c(4, 4))
    errMat[1, 2] <- 1
    errMat[1, 3] <- 1
    errMat[4, 2] <- 1

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", "", ""), errorProneMatrix = errMat
    )
  })
})

test_that("Checking measurement error error checks 5", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    w2 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, w2, y)
    errMat <- array(0, dim = c(4, 4))
    errMat[1, 2] <- 1
    errMat[1, 3] <- 1
    errMat[4, 2] <- 2

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", "", ""), errorProneMatrix = errMat
    )
  })
})


test_that("Checking measurement error error checks 6", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    w2 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)

    simData <- data.frame(x, w1, w2, y)
    errMat <- array(0, dim = c(5, 5))
    errMat[1, 2] <- 1
    errMat[1, 3] <- 1

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("latnorm", "", "", ""), errorProneMatrix = errMat
    )
  })
})

test_that("Checking measurement error error checks 7", {
  expect_error({
    set.seed(1234)
    n <- 100
    x <- rnorm(n)
    w1 <- x + rnorm(n)
    w2 <- x + rnorm(n)
    y <- y <- x + rnorm(n)
    x <- rep(NA, n)
    z <- rnorm(n)
    z[1:50] <- NA

    simData <- data.frame(w1, w2, y, z)
    errMat <- array(0, dim = c(4, 4))
    errMat[3, 1] <- 1
    errMat[3, 2] <- 1

    imps <- smcfcs(simData,
      smtype = "lm", smformula = "y~x",
      method = c("", "", "", "norm"), errorProneMatrix = errMat
    )
  })
})

test_that("Cox imputation fails if event indicator is not coded right", {
  expect_error({
    set.seed(1234)
    n <- 1000
    z <- rnorm(n)
    x <- z + rnorm(n)
    t <- -log(runif(n)) / (1 * exp(x + z))
    d <- 1 * (t < 10) + 1
    t[d == 1] <- 10
    x[(runif(n) < 0.5)] <- NA

    simData <- data.frame(t, d, x, z)

    imps <- smcfcs(simData,
      smtype = "coxph", smformula = "Surv(t, d)~x+z",
      method = c("", "", "norm", "")
    )
  })
})

test_that("Cox error check doesn't fail when events are 1 1 1 0 0 0", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- runif(n)
      d <- c(rep(1, n / 2), rep(0, n / 2))
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs(simData,
        smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", "")
      )
    },
    NA
  )
})

library(smcfcs)
library(survival)
context("Discrete time survival analysis testing")

test_that("Basic setup runs (CRAN check)", {
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- -3 + 0.2 * (1:T)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      table(simData$failtime, simData$d)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        method = c("logreg", "", "", ""), m = M
      )
    },
    NA
  )
})

test_that("Errors if times are not integers", {
  expect_error({
    set.seed(1234)
    n <- 1000
    x1 <- 1 * (runif(n) < 0.5)
    x2 <- x1 + rnorm(n)
    T <- 10
    # define vector of intercepts
    alpha <- -3 + 0.2 * (1:T)
    beta <- c(1, -1)
    yMat <- array(0, dim = c(n, T))
    for (i in 1:T) {
      yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
    }
    failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
    # event indicator
    d <- rep(1, n)
    d[is.na(failtime)] <- 0
    failtime[is.na(failtime)] <- 10
    mean(d)

    simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
    simData$failtime <- simData$failtime / 2

    simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

    # impute using smcfcs.dtsam
    M <- 5
    imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
      method = c("logreg", "", "", ""), m = M
    )
  })
})

test_that("Errors if times not all positive", {
  expect_error({
    set.seed(1234)
    n <- 1000
    x1 <- 1 * (runif(n) < 0.5)
    x2 <- x1 + rnorm(n)
    T <- 10
    # define vector of intercepts
    alpha <- -3 + 0.2 * (1:T)
    beta <- c(1, -1)
    yMat <- array(0, dim = c(n, T))
    for (i in 1:T) {
      yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
    }
    failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
    # event indicator
    d <- rep(1, n)
    d[is.na(failtime)] <- 0
    failtime[is.na(failtime)] <- 10
    mean(d)

    simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
    simData$failtime <- simData$failtime - 1

    simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

    # impute using smcfcs.dtsam
    M <- 5
    imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
      method = c("logreg", "", "", ""), m = M
    )
  })
})

test_that("Errors if factor time model but some times with no events", {
  expect_error({
    set.seed(1234)
    n <- 1000
    x1 <- 1 * (runif(n) < 0.5)
    x2 <- x1 + rnorm(n)
    T <- 10
    # define vector of intercepts
    alpha <- -3 + 0.2 * (1:T)
    beta <- c(1, -1)
    yMat <- array(0, dim = c(n, T))
    for (i in 1:T) {
      yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
    }
    failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
    # event indicator
    d <- rep(1, n)
    d[is.na(failtime)] <- 0
    failtime[is.na(failtime)] <- 10
    mean(d)

    simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
    simData$d[simData$failtime == 5] <- 0

    simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

    # impute using smcfcs.dtsam
    M <- 5
    imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
      method = c("logreg", "", "", ""), m = M
    )
  })
})

test_that("Still runs with some times with no events if linear time effects used", {
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- -3 + 0.2 * (1:T)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      simData$d[simData$failtime == 5] <- 0

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        method = c("logreg", "", "", ""), m = M, timeEffects = "linear"
      )
    },
    NA
  )
})

# the following tests are to check statistical properties are as they should be

test_that("Full data dtsam is unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- seq(-1, -0.1, 0.1)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)

      # fit dtsam model
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ -1 + factor(tstart) + x1 + x2, family = "binomial", data = longData)
      summary(mod)
      ciLower <- coef(mod) - qnorm(0.999) * sqrt(diag(vcov(mod)))
      ciUpper <- coef(mod) + qnorm(0.999) * sqrt(diag(vcov(mod)))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})

test_that("MAR is unbiased, binary covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- -3 + 0.2 * (1:T)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      table(simData$failtime, simData$d)

      # fit dtsam model to full dataset
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ factor(tstart) + x1 + x2, family = "binomial", data = longData)
      summary(mod)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        method = c("logreg", "", "", ""), m = M
      )

      # fit dtsam model to each dataset
      ests <- vector(mode = "list", length = M)
      vars <- vector(mode = "list", length = M)
      for (i in 1:M) {
        longData <- survSplit(Surv(failtime, d) ~ x1 + x2,
          data = imps$impDatasets[[i]],
          cut = 1:T
        )
        mod <- glm(d ~ -1 + factor(tstart) + x1 + x2, family = "binomial", data = longData)
        ests[[i]] <- coef(mod)
        vars[[i]] <- diag(vcov(mod))
      }

      rubin <- mitools::MIcombine(ests, vars)

      ciLower <- rubin$coefficients - qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciUpper <- rubin$coefficients + qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})

test_that("MCAR is unbiased, cts covariate missing", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(7234423)
      n <- 1000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- -3 + 0.2 * (1:T)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      table(simData$failtime, simData$d)

      # fit dtsam model to full dataset
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ factor(tstart) + x1 + x2, family = "binomial", data = longData)
      summary(mod)

      # simData$x2[runif(n)<plogis((x1-mean(x1)/sd(x1)))] <- NA
      simData$x2[runif(n) < 0.5] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        method = c("", "norm", "", ""), m = M, rjlimit = 10000
      )

      # fit dtsam model to each dataset
      ests <- vector(mode = "list", length = M)
      vars <- vector(mode = "list", length = M)
      for (i in 1:M) {
        longData <- survSplit(Surv(failtime, d) ~ x1 + x2,
          data = imps$impDatasets[[i]],
          cut = 1:T
        )
        mod <- glm(d ~ -1 + factor(tstart) + x1 + x2, family = "binomial", data = longData)
        ests[[i]] <- coef(mod)
        vars[[i]] <- diag(vcov(mod))
      }

      rubin <- mitools::MIcombine(ests, vars)

      ciLower <- rubin$coefficients - qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciUpper <- rubin$coefficients + qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})

test_that("MAR is unbiased, binary covariate missing, linear time effect", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- c(-2, 0.1)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[1] + alpha[2] * (i - 1) + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)

      # fit dtsam model to full dataset
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ tstart + x1 + x2, family = "binomial", data = longData)
      summary(mod)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA
      # simData$x2[runif(n)<0.25] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        timeEffects = "linear",
        method = c("logreg", "", "", ""), m = M
      )

      # fit dtsam model to each dataset
      ests <- vector(mode = "list", length = M)
      vars <- vector(mode = "list", length = M)
      for (i in 1:M) {
        longData <- survSplit(Surv(failtime, d) ~ x1 + x2,
          data = imps$impDatasets[[i]],
          cut = 1:T
        )
        mod <- glm(d ~ tstart + x1 + x2, family = "binomial", data = longData)
        ests[[i]] <- coef(mod)
        vars[[i]] <- diag(vcov(mod))
      }

      rubin <- mitools::MIcombine(ests, vars)

      ciLower <- rubin$coefficients - qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciUpper <- rubin$coefficients + qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})

test_that("MAR is unbiased, binary covariate missing, quadratic time effect", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- c(-3, 0.1, 0.02)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[1] + alpha[2] * (i - 1) + alpha[3] * (i - 1)^2 + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)

      # fit dtsam model to full data
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ tstart + I(tstart^2) + x1 + x2, family = "binomial", data = longData)
      summary(mod)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA
      # simData$x2[runif(n)<0.25] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
        timeEffects = "quad",
        method = c("logreg", "", "", ""), m = M
      )

      # fit dtsam model to each dataset
      ests <- vector(mode = "list", length = M)
      vars <- vector(mode = "list", length = M)
      for (i in 1:M) {
        longData <- survSplit(Surv(failtime, d) ~ x1 + x2,
          data = imps$impDatasets[[i]],
          cut = 1:T
        )
        mod <- glm(d ~ tstart + I(tstart^2) + x1 + x2, family = "binomial", data = longData)
        ests[[i]] <- coef(mod)
        vars[[i]] <- diag(vcov(mod))
      }

      rubin <- mitools::MIcombine(ests, vars)

      ciLower <- rubin$coefficients - qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciUpper <- rubin$coefficients + qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})


test_that("Binary covariate as factor runs and matches when numeric", {
  expect_equal(
    {
      set.seed(1234)
      n <- 100
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- c(-3, 0.1, 0.02)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[1] + alpha[2] * (i - 1) + alpha[3] * (i - 1)^2 + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      set.seed(71223)
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
                           timeEffects = "quad",
                           method = c("logreg", "", "", ""), m = M
      )
      # now make x1 a factor
      simData$x1 <- factor(simData$x1)
      set.seed(71223)
      impsx1factor <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
                           timeEffects = "quad",
                           method = c("logreg", "", "", ""), m = M
      )

      imps$smCoefIter
    },
    impsx1factor$smCoefIter
  )
})

test_that("MAR is unbiased, categorical covariate missing, linear time effect", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      u <- runif(n)
      x1 <- rep(0,n)
      x1[u>0.7] <- 2
      x1[(u>0.3) & (u <= 0.7)] <- 1
      x2 <- 0.5*(x1==1)+1*(x1==2) + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- c(-2, 0.1)
      beta <- c(1, -1, 0.5)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[1] + alpha[2] * (i - 1)
                                           + beta[1] * 1*(x1==1)
                                           + beta[2] * 1*(x1==2)
                                           + beta[3] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = factor(x1), x2 = x2, failtime = failtime, d = d)

      # fit dtsam model to full dataset
      longData <- survSplit(Surv(failtime, d) ~ x1 + x2, data = simData, cut = 1:T)
      mod <- glm(d ~ tstart + x1 + x2, family = "binomial", data = longData)
      summary(mod)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.dtsam(simData, "Surv(failtime,d)~x1+x2",
                           timeEffects = "linear",
                           method = c("mlogit", "", "", ""), m = M
      )

      # fit dtsam model to each dataset
      ests <- vector(mode = "list", length = M)
      vars <- vector(mode = "list", length = M)
      for (i in 1:M) {
        longData <- survSplit(Surv(failtime, d) ~ x1 + x2,
                              data = imps$impDatasets[[i]],
                              cut = 1:T
        )
        mod <- glm(d ~ tstart + x1 + x2, family = "binomial", data = longData)
        ests[[i]] <- coef(mod)
        vars[[i]] <- diag(vcov(mod))
      }

      rubin <- mitools::MIcombine(ests, vars)

      ciLower <- rubin$coefficients - qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciUpper <- rubin$coefficients + qt(0.999, df = rubin$df) * sqrt(diag(rubin$variance))
      ciIncluded <- (ciLower < c(alpha, beta)) * (ciUpper > c(alpha, beta))

      mean(ciIncluded) == 1
    },
    TRUE
  )
})

library(smcfcs)
context("Bias reduced model testing")

test_that("smcfcs gives no warnings or errors with perfect prediction and brlogreg method", {
  expect_error(
    {
      n <- 10
      z <- c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1)
      x <- z
      y <- x + rnorm(n)
      x[c(5, 10)] <- NA

      simData <- data.frame(x, z, y)
      rm(x, z, y)
      # impute x using z as an auxiliary variable
      predMat <- array(0, dim = c(3, 3))
      predMat[1, 2] <- 1
      set.seed(62377)
      imps <- smcfcs(simData,
        smtype = "lm", smformula = "y~x",
        method = c("brlogreg", "", ""), predictorMatrix = predMat, m = 1
      )
    },
    NA
  )
})

test_that("brlogreg gives results with minimal bias under perfect prediction between predictors", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      nSim <- 100

      simEsts <- array(0, dim = c(nSim, 2))

      for (i in 1:nSim) {
        z <- 1 * (runif(n) < 0.5)
        # true log odds parameters of x|z are finite but very large
        x <- 1 * (runif(n) < plogis(-10 + 20 * z))
        x[runif(n) < 0.25] <- NA
        y <- x + rnorm(n)

        simData <- data.frame(x, z, y)
        rm(x, z, y)
        # impute x using z as an auxiliary variable
        predMat <- array(0, dim = c(3, 3))
        predMat[1, 2] <- 1

        imps <- smcfcs(simData,
          smtype = "lm", smformula = "y~x",
          method = c("brlogreg", "", ""), predictorMatrix = predMat
        )
        library(mitools)
        impobj <- imputationList(imps$impDatasets)
        smcfcsRes <- MIcombine(with(impobj, lm(y ~ x)))
        simEsts[i, ] <- smcfcsRes$coefficients
      }
      ttestRes <- t.test(simEsts[, 2], mu = 1)
      ttestRes$p.value > 0.05
    },
    TRUE
  )
})


test_that("smcfcs gives no warnings or errors with perfect prediction and brlogistic method
          for substantive model and covariate model", {
  expect_error(
    {
      n <- 10
      x <- c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1)
      y <- c(0, 0, 0, 0, 0, 1, 1, 1, 1, 1)
      x[c(5, 10)] <- NA

      simData <- data.frame(x, y)
      rm(x, y)
      set.seed(62377)
      imps <- smcfcs(simData,
        smtype = "brlogistic", smformula = "y~x",
        method = c("brlogreg", ""), m = 1
      )
    },
    NA
  )
})

library(smcfcs)
library(survival)
context("Weibull model testing")

test_that("Weibull imputation runs, binary covariate", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      x <- 1 * (runif(n) < 0.5)
      z <- x + rnorm(n)
      t <- rsurvreg(n, mean = x + z, scale = 2)
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs(simData,
        smtype = "weibull", smformula = "Surv(t, d)~x+z",
        method = c("", "", "logreg", ""), m = 1
      )
    },
    NA
  )
})

test_that("Weibull imputation is consistent, binary covariate", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x <- 1 * (runif(n) < 0.5)
      z <- x + rnorm(n)
      t <- rsurvreg(n, mean = x + z, scale = 2)
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs(simData,
        smtype = "weibull", smformula = "Surv(t, d)~x+z",
        method = c("", "", "logreg", ""), m = 1
      )
      fitmod <- survreg(Surv(t, d) ~ x + z, data = imps$impDatasets[[1]])
      as.logical((abs(coef(fitmod)[2] - 1) < 0.05) & (abs(fitmod$scale - 2) < 0.05))
    },
    TRUE
  )
})

test_that("Weibull imputation is consistent, cts covariate", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 100000
      x <- rnorm(n)
      z <- x + rnorm(n)
      t <- rsurvreg(n, mean = x + z, scale = 2)
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs(simData,
        smtype = "weibull", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""), m = 1
      )
      fitmod <- survreg(Surv(t, d) ~ x + z, data = imps$impDatasets[[1]])
      as.logical((abs(coef(fitmod)[2] - 1) < 0.05) & (abs(fitmod$scale - 2) < 0.05))
    },
    TRUE
  )
})

library(smcfcs)
context("Testing imputation with mlogit")

test_that("mlogit runs without error", {
  expect_no_error(
    {
      set.seed(1234)
      n <- 1000

      z <- rnorm(n)
      xb2 <- exp(-1+1*z)
      xb3 <- exp(0.5-0.75*z)
      pr2 <- xb2/(1+xb2+xb3)
      pr3 <- xb3/(1+xb2+xb3)
      pr1 <- 1-pr2-pr3
      u <- runif(n)
      x <- rep("a",n)
      x[(u>pr1) & (u>(pr1+pr2))] <- "c"
      x[(u>pr1) & (u<(pr1+pr2))] <- "b"

      y <- z+(x=="b")-(x=="c")+rnorm(n)

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.factor(x),z=z,y=y)

      imps <- smcfcs(originaldata=simData,
        smtype="lm",
        smformula = "y~z+I(x=='b')+I(x=='c')",
        method = c("mlogit","", "")
      )
    },
  )
})

test_that("mlogit covariate imp is unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(2347633)
      n <- 100000

      z <- rnorm(n)
      xb2 <- exp(-1+1*z)
      xb3 <- exp(0.5-0.75*z)
      pr2 <- xb2/(1+xb2+xb3)
      pr3 <- xb3/(1+xb2+xb3)
      pr1 <- 1-pr2-pr3
      u <- runif(n)
      x <- rep("a",n)
      x[(u>pr1) & (u>(pr1+pr2))] <- "c"
      x[(u>pr1) & (u<(pr1+pr2))] <- "b"

      y <- z+(x=="b")-(x=="c")+rnorm(n)

      x[1*runif(n)<(exp(z)/(1+exp(z)))] <- NA

      simData <- data.frame(x=as.factor(x),z=z,y=y)

      imps <- smcfcs(originaldata=simData,
                     smtype="lm",
                     smformula = "y~z+I(x=='b')+I(x=='c')",
                     method = c("mlogit","", "")
      )

      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, lm(y ~ z+I(x=='b')+I(x=='c')))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[4] < (-1)) & (MIcombineRes$`upper)`[4]>(-1))
    },
    TRUE
  )

})

#' Parallel substantive model compatible imputation
#'
#' Runs substantive model compatible imputation using parallel cores
#'
#' This function can be used to call one of the substantive model compatible imputation
#' methods using parallel cores, to reduce computation time. You must specify
#' the arguments required for the standard smcfcs call, and then specify your
#' the arguments for how to use parallel cores.
#'
#' @author Edouard F. Bonneville \email{e.f.bonneville@@lumc.nl}
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @param smcfcs_func Specifies which base smcfcs function to call. Possible values
#' are `smcfcs`, `smcfcs.casecohort`, `smcfcs.dtasam`, `smcfcs.nestedcc`. Defaults
#' to `smcfcs`.
#' @param seed Optional seed, set as `set.seed` when `n_cores = 1`,
#' or as `parallel::clusterSetRNGStream` when `n_cores > 1`.
#' @param m Number of imputed datasets to generate.
#' @param n_cores Number of cores over which to split the `m` imputations. If
#' `n_cores` is not divisible exactly by `m`, one of the cores will perform
#' more/less imputations that the rest such that the final result still contains
#' `m` imputed datasets.
#' @param cl_type Either "PSOCK" or "FORK". If running on a Windows system
#' "PSOCK" is recommended, otherwise for Linux/Mac machines "FORK" tends to
#' offer faster computation - see \link[mice]{parlmice}.
#' @param outfile Optional character path to location for
#' output from the workers. Useful to diagnose rejection sampling warnings.
#' File path must be formulated as "path/to/filename.txt".
#' @param ... Additional arguments to pass on to \link[smcfcs]{smcfcs},
#' \link[smcfcs]{smcfcs.casecohort},
#' \link[smcfcs]{smcfcs.dtsam}, or
#' \link[smcfcs]{smcfcs.nestedcc}.
#'
#' @return An object of type "smcfcs", as would usually be returned from
#' \link[smcfcs]{smcfcs}.
#' @export
#'
#' @examples
#' \dontrun{
#' # Detect number of cores
#' parallel::detectCores()
#'
#' imps <- smcfcs.parallel(
#'   smcfcs_func = "smcfcs",
#'   seed = 2021,
#'   n_cores = 2,
#'   originaldata = smcfcs::ex_compet,
#'   m = 10,
#'   smtype = "compet",
#'   smformula = list(
#'     "Surv(t, d == 1) ~ x1 + x2",
#'     "Surv(t, d == 2) ~ x1 + x2"
#'   ),
#'   method = c("", "", "norm", "norm")
#' )
#' }
smcfcs.parallel <- function(smcfcs_func = "smcfcs",
                            seed = NULL,
                            m = 5,
                            n_cores = parallel::detectCores() - 1,
                            cl_type = "PSOCK",
                            outfile = "",
                            ...) {
  checkmate::matchArg(
    x = smcfcs_func,
    choices = c("smcfcs", "smcfcs.casecohort", "smcfcs.dtsam", "smcfcs.nestedcc")
  )

  # Save chosen function as string and and expression for later use
  smcfcs_func_string <- paste0("smcfcs::", smcfcs_func)
  smcfcs_func_expr <- parse(text = smcfcs_func_string)

  # Check smcfcs arguments
  args <- list(...)
  args_smcfcs <- names(formals(eval(smcfcs_func_expr)))
  # if ... is allowed by smcfcs subfunction, include main smcfcs argument as well
  if ("..." %in% args_smcfcs) {
    args_smcfcs <- unique(c(args_smcfcs, names(formals(smcfcs::smcfcs))))
  }
  check_args <- !(names(args) %in% args_smcfcs)

  if (any(check_args)) {
    wrong_args <- paste(names(args)[check_args], collapse = ", ")
    mssg <- paste0(
      "The following are not valid arguments of",
      smcfcs_func_string, " : ", wrong_args
    )
    stop(mssg)
  }

  # Check parallel arguments
  checkmate::assert_int(x = seed, null.ok = TRUE)
  checkmate::assert_int(x = m, lower = 1)
  # checkmate::assert_int(x = m_per_core, lower = 1, upper = max(1,floor(m / n_cores)), null.ok = TRUE)
  checkmate::matchArg(x = cl_type, choices = c("PSOCK", "FORK"))
  checkmate::assert_int(x = n_cores, lower = 1, upper = min(parallel::detectCores(), m))
  if (outfile != "") checkmate::assert_path_for_output(x = outfile, overwrite = TRUE)

  # Standard smcfcs if n_cores = 1
  if (n_cores == 1) {
    if (!is.null(seed)) set.seed(seed)
    args$m <- m
    res <- do.call(eval(smcfcs_func_expr), args)
  } else {
    # Determine number of imputations per core
    imp_specs <- determine_imp_specs(n_cores, m)

    # Set up the cluster
    cl <- parallel::makeCluster(n_cores, type = cl_type, outfile = outfile)
    parallel::clusterSetRNGStream(cl, seed)

    # Export necessary objects/functions
    parallel::clusterCall(cl, assign, "Surv", survival::Surv, envir = .GlobalEnv)
    parallel::clusterCall(cl, assign, "strata", survival::strata, envir = .GlobalEnv)
    parallel::clusterExport(
      cl = cl,
      varlist = c(
        "args", "imp_specs", "seed", "m", "n_cores", "cl_type", "smcfcs",
        "smcfcs.casecohort", "smcfcs.dtsam", "smcfcs.nestedcc", "smcfcs_func_expr"
      ),
      envir = environment()
    )

    # Run the imputations
    imps <- parallel::parLapply(cl = cl, X = seq_along(imp_specs), function(x) {
      args$m <- imp_specs[x]
      do.call(eval(smcfcs_func_expr), args)
    })

    parallel::stopCluster(cl)

    # Combine imputations
    res <- combine_smcfcs_objects(imps)
  }

  return(res)
}


# Prepare imputations per core
determine_imp_specs <- function(n_cores,
                                m) {
  imp_specs <- rep(floor(m / n_cores), times = n_cores)
  modul <- m %% n_cores

  # Add remaining imps to add to m
  if (modul != 0) imp_specs[length(imp_specs)] <- imp_specs[length(imp_specs)] + modul
  return(imp_specs)
}


# Helper to combine smcfcs objects
combine_smcfcs_objects <- function(smcfcs_list) {
  # Combine imputed datasets
  ls_impdats <- do.call("c", lapply(smcfcs_list, "[[", "impDatasets"))

  # Combine monitoring of imputations
  coef_array <- abind::abind(lapply(smcfcs_list, "[[", "smCoefIter"), along = 1)

  # Polish and return
  res <- list(
    impDatasets = ls_impdats,
    smCoefIter = coef_array,
    smInfo = list(
      smtype = smcfcs_list[[1]]$smInfo$smtype,
      smformula = smcfcs_list[[1]]$smInfo$smformula
    )
  )

  class(res) <- "smcfcs"
  return(res)
}

library(smcfcs)
library(survival)
context("Testing parallel computation")

test_that("Cox imputation runs in parallel", {
  expect_error(
    {
      set.seed(1234)
      n <- 10000
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.parallel(
        originaldata = simData, smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""), m = 5, n_cores = 2, seed = 7243
      )
    },
    NA
  )
})

test_that("smcfcs.parallel works with 1 imputation per core", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.parallel(
        originaldata = simData, smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""), m = 2, n_cores = 2, seed = 7243
      )
    },
    NA
  )
})

test_that("smcfcs.parallel throws error if you specify more cores than necessary for 1 imp per core", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.parallel(
        originaldata = simData, smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""), m = 5, n_cores = 10, seed = 7243
      )
    },
    NULL
  )
})

test_that("smcfcs.parallel throws error if you specify non-integer seed", {
  expect_error(
    {
      set.seed(1234)
      n <- 100
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.parallel(
        originaldata = simData, smtype = "coxph", smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""), m = 5, n_cores = 2, seed = 1.5
      )
    },
    NULL
  )
})

test_that("Test case cohort imputation runs", {
  expect_error(
    {
      set.seed(1234)

      # run simulation study
      n <- 10000

      z <- rnorm(n)
      x <- 1 * (runif(n) < (exp(z) / (1 + exp(z))))
      t <- -log(runif(n)) / (0.01 * exp(x + z))
      d <- 1 * (t < 2)
      t[d == 0] <- 2
      x[(runif(n) < 0.5)] <- NA

      fullcohortdata <- data.frame(t, d, x, z)
      fullcohortdata$in.subco <- 0
      # we sample a 10% subcohort
      fullcohortdata$in.subco[sample(n, size = n * 0.1)] <- 1
      fullcohortdata$id <- 1:n

      ccdata <- fullcohortdata[(fullcohortdata$in.subco == 1) | (fullcohortdata$d == 1), ]
      ccdata$entertime <- 0
      ccdata$entertime[ccdata$in.subco == 0] <- ccdata$t[ccdata$in.subco == 0] - 0.000001
      imps <- smcfcs.parallel(
        originaldata = ccdata, smformula = "Surv(entertime, t, d)~x+z", sampfrac = 0.1,
        in.subco = "in.subco", method = c("", "", "logreg", "", "", "", ""),
        smcfcs_func = "smcfcs.casecohort", m = 5, n_cores = 2, seed = 7243
      )
    },
    NA
  )
})

test_that("Nested case control imputation runs", {
  expect_error(
    {
      set.seed(1234)

      n <- 10000

      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (0.01 * exp(x + z))
      d <- 1 * (t < 1)
      t[d == 0] <- 1
      x[(runif(n) < 0.5)] <- NA

      fullcohortdata <- data.frame(t, d, x, z)
      fullcohortdata$id <- 1:n

      # Compute number at risk at each event time using the full cohort data
      nrisk.fit <- survfit(Surv(t, d) ~ 1, data = fullcohortdata)
      ord.t.d1 <- order(fullcohortdata$t[fullcohortdata$d == 1])

      m <- 1 # 1 control per case
      ncc <- NULL

      no.sample <- 0
      for (i in which(fullcohortdata$d == 1))
      {
        # select control(s) for nested case-control
        possible.controls <- which(fullcohortdata$t >= fullcohortdata$t[i])
        # remove the case from this vector
        possible.controls <- possible.controls[which(possible.controls != i)]

        if (length(possible.controls) >= m) {
          controls <- sample(possible.controls, m)
          numAtRisk <- 1 + length(possible.controls)
          ncc <- rbind(ncc, cbind(fullcohortdata[i, ], numrisk = numAtRisk))
          ncc <- rbind(ncc, cbind(fullcohortdata[controls, ], numrisk = numAtRisk))
          no.sample <- no.sample + 1
        }
      }

      ncc$setno <- rep(1:no.sample, each = m + 1)
      ncc$case <- rep(c(1, rep(0, m)), no.sample)

      predictorMatrix <- matrix(0, nrow = dim(ncc)[2], ncol = dim(ncc)[2])
      predictorMatrix[which(colnames(ncc) == "x"), c(which(colnames(ncc) == "z"))] <- 1

      imps <- smcfcs.parallel(
        originaldata = ncc, set = "setno", nrisk = "numrisk", event = "d", smformula = "Surv(t,case)~x+z+strata(setno)",
        method = c("", "", "norm", "", "", "", "", ""), predictorMatrix = predictorMatrix,
        smcfcs_func = "smcfcs.nestedcc", m = 5, n_cores = 2, seed = 7243
      )
    },
    NA
  )
})

test_that("Fails if invalid arg supplied", {
  expect_error(
    {
      set.seed(1234)

      n <- 10000

      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (0.01 * exp(x + z))
      d <- 1 * (t < 1)
      t[d == 0] <- 1
      x[(runif(n) < 0.5)] <- NA

      fullcohortdata <- data.frame(t, d, x, z)
      fullcohortdata$id <- 1:n

      # Compute number at risk at each event time using the full cohort data
      nrisk.fit <- survfit(Surv(t, d) ~ 1, data = fullcohortdata)
      ord.t.d1 <- order(fullcohortdata$t[fullcohortdata$d == 1])

      m <- 1 # 1 control per case
      ncc <- NULL

      no.sample <- 0
      for (i in which(fullcohortdata$d == 1))
      {
        # select control(s) for nested case-control
        possible.controls <- which(fullcohortdata$t >= fullcohortdata$t[i])
        # remove the case from this vector
        possible.controls <- possible.controls[which(possible.controls != i)]

        if (length(possible.controls) >= m) {
          controls <- sample(possible.controls, m)
          numAtRisk <- 1 + length(possible.controls)
          ncc <- rbind(ncc, cbind(fullcohortdata[i, ], numrisk = numAtRisk))
          ncc <- rbind(ncc, cbind(fullcohortdata[controls, ], numrisk = numAtRisk))
          no.sample <- no.sample + 1
        }
      }

      ncc$setno <- rep(1:no.sample, each = m + 1)
      ncc$case <- rep(c(1, rep(0, m)), no.sample)

      predictorMatrix <- matrix(0, nrow = dim(ncc)[2], ncol = dim(ncc)[2])
      predictorMatrix[which(colnames(ncc) == "x"), c(which(colnames(ncc) == "z"))] <- 1

      imps <- smcfcs.parallel(
        originaldata = ncc, set = "setno", nrisk = "numrisk", event = "d", smformula = "Surv(t,case)~x+z+strata(setno)",
        method = c("", "", "norm", "", "", "", "", ""), predictorMatrix = predictorMatrix,
        smcfcs_func = "smcfcs.nestedcc", m = 5, n_cores = 2, seed = 7243, sampfrac = 0.1,
      )
    }
  )
})

test_that("DTSAM imputation runs", {
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      x1 <- 1 * (runif(n) < 0.5)
      x2 <- x1 + rnorm(n)
      T <- 10
      # define vector of intercepts
      alpha <- -3 + 0.2 * (1:T)
      beta <- c(1, -1)
      yMat <- array(0, dim = c(n, T))
      for (i in 1:T) {
        yMat[, i] <- 1 * (runif(n) < plogis(alpha[i] + beta[1] * x1 + beta[2] * x2))
      }
      failtime <- apply(yMat, 1, function(x) which(x == 1)[1])
      # event indicator
      d <- rep(1, n)
      d[is.na(failtime)] <- 0
      failtime[is.na(failtime)] <- 10
      mean(d)

      simData <- data.frame(x1 = x1, x2 = x2, failtime = failtime, d = d)
      table(simData$failtime, simData$d)

      simData$x1[runif(n) < plogis((x2 - mean(x2) / sd(x2)))] <- NA

      # impute using smcfcs.dtsam
      M <- 5
      imps <- smcfcs.parallel(
        originaldata = simData, smformula = "Surv(failtime,d)~x1+x2",
        method = c("logreg", "", "", ""),
        smcfcs_func = "smcfcs.dtsam", m = M, n_cores = 2, seed = 7243
      )
    },
    NA
  )
})

#' Substantive model compatible fully conditional specification imputation of
#' covariates for a Fine-Gray model
#'
#' Multiply imputes missing covariate values using substantive model compatible
#' fully conditional specification for competing risks outcomes, when the
#' substantive model is a Fine-Gray model for the subdistribution hazard of one event.
#'
#' In the presence of random right censoring, the function first multiply imputes
#' the potential censoring times for those failing from competing events using
#' \link[kmi]{kmi}, and thereafter uses \link[smcfcs]{smcfcs} to impute the missing
#' covariates. See Bonneville \emph{et al.} 2024 for further details on the methodology.
#'
#' The function does not (yet) support parallel computation.
#'
#' @author Edouard F. Bonneville \email{e.f.bonneville@@lumc.nl}
#'
#' @param smformula The formula of the substantive model, given as a string. Needs to
#' be of the form "Surv(t, d) ~ x1 + x2", where t is a vector of competing event
#' times, and d is a (numeric) competing event indicator, where 0 must designate
#' a censored observation.
#' @param cause Numeric, designating the competing event of interest (default is
#' `cause = 1`).
#' @param kmi_args List, containing arguments to be passed on to \link[kmi]{kmi}.
#' The "formula" element is a formula where the right-hand side specifies the
#' covariates used for multiply imputing the potential censoring times for
#' individual's failing from competing events. The default is `formula = ~ 1`,
#' which uses marginal Kaplan-Meier estimator of the censoring distribution.
#' @param ... Additional arguments to pass on to \link[smcfcs]{smcfcs}
#' @inheritParams smcfcs
#'
#' @return An object of type "smcfcs", as would usually be returned from
#' \link[smcfcs]{smcfcs}.
#' @export
#'
#' @references Bonneville EF, Beyersmann J, Keogh RH, Bartlett JW, Morris TP,
#' Polverelli N, de Wreede LC, Putter H. Multiple imputation of missing covariates
#' when using the Fine--Gray model. 2024. Submitted.
#'
#' @examples
#' \dontrun{
#' library(survival)
#' library(kmi)
#'
#' imps <- smcfcs.finegray(
#'   originaldata = ex_finegray,
#'   smformula = "Surv(times, d) ~ x1 + x2",
#'   method = c("", "", "logreg", "norm"),
#'   cause = 1,
#'   kmi_args = list("formula" = ~ 1)
#' )
#'
#' if (requireNamespace("mitools", quietly = TRUE)) {
#'   library(mitools)
#'   impobj <- imputationList(imps$impDatasets)
#'   # Important: use Surv(newtimes, newevent) ~ ... when pooling
#'   # (respectively: subdistribution time and indicator for cause of interest)
#'   models <- with(impobj, coxph(Surv(newtimes, newevent) ~ x1 + x2))
#'   summary(MIcombine(models))
#' }
#' }
#'
smcfcs.finegray <- function(originaldata,
                            smformula,
                            method,
                            cause = 1,
                            m = 5,
                            numit = 10,
                            rjlimit = 5000,
                            kmi_args = list(
                              "formula" = ~ 1,
                              "bootstrap" = FALSE,
                              "nboot" = 10
                            ),
                            ...) {

  if (!requireNamespace("kmi", quietly = TRUE)) {
    stop("Package {kmi} needed for this function to work. Please install it.", call. = FALSE)
  }

  # Locate/sort out outcome variables
  outcome_vars <- all.vars(update(as.formula(smformula), . ~ 1))
  time_var_name <- outcome_vars[[1]] # will need to add warning/error if (tstart, tstop) format
  status_var_name <- outcome_vars[[2]]
  time_var <- originaldata[[time_var_name]]
  status_var <- originaldata[[status_var_name]]

  # Some checks
  checkmate::assert_numeric(x = time_var, lower = 0, finite = TRUE, any.missing = FALSE)
  checkmate::assert_numeric(x = status_var, lower = 0, finite = TRUE, any.missing = FALSE)

  # Get functional form RHS of formula
  smformula_rhs <- unlist(strsplit(smformula, split = "~"))[2]

  # Use names consistent with kmi's names
  smformula_processed <- paste0("Surv(newtimes, newevent) ~", smformula_rhs)
  meths_smcfcs <- c(method, "newtimes" = "", newevent = "")

  # Check number censored, since based on this we use {kmi} or not
  num_censored <- sum(status_var == 0)

  # If no censoring: just pre-process data
  if (num_censored == 0) {

    # Set competing events to event time larger that largest event 1 time
    # (in this case: largest event 1 time + 0.1)
    eps <- 0.1
    max_ev1_time <- max(time_var[status_var == cause])
    newtimes <- time_var
    newtimes[status_var != cause] <- max_ev1_time + eps
    newevent <- as.numeric(status_var == cause)

    # Bind to original data
    originaldata_processed <- cbind.data.frame(originaldata, newtimes, newevent)

    # Run imputations
    smcfcs_obj <- smcfcs(
      originaldata = originaldata_processed,
      smtype = "coxph",
      smformula = smformula_processed,
      method = meths_smcfcs,
      rjlimit = rjlimit,
      numit = numit,
      m = m,
      ...
    )

  } else { # Need to multiply impute censoring times!

    # Prepare kmi() formula (for now default Kaplan-Meier imputation)
    lhs_kmi <- paste0("Surv(", paste(outcome_vars, collapse = ", "), " != 0)")
    form_kmi <- as.formula(paste0(lhs_kmi, deparse1(kmi_args$formula)))
    rhs_cens <- all.vars(kmi_args$formula)

    # Some checks for the censoring
    if (length(rhs_cens) > 0) {
      if (any(!(rhs_cens %in% colnames(originaldata))))
        stop("Variables used to model the censoring distribution need to be in originaldata.")
      if (anyNA(originaldata[, rhs_cens]))
        warning(
          "          One or more of the variables used to model the censoring
          distribution has missing data. This MI method works under the assumption
          that only complete covariates can be associated with the censoring. If
          nothing is changed by the user, {kmi} deals with these missing values
          internally by using single imputation."
        )
    }

    args_cens_imps <- c(
      list(
        "formula" = form_kmi,
        "data" = originaldata,
        "etype" = as.symbol(status_var_name),
        "failcode" = cause,
        "nimp" = m
      ),
      kmi_args[-1] # remove formula
    )

    # Impute missing censoring times in first loop
    kmi_imps <- do.call(kmi::kmi, args = args_cens_imps)

    # And now impute the covariates
    imps_loop <- lapply(kmi_imps$imputed.data, function(new_outcomes) {

      df_imp <- cbind(kmi_imps$original.data, new_outcomes)
      df_imp$newevent <- as.numeric(df_imp$newevent) - 1L # Make numeric for smcfcs

      smcfcs_modif <- smcfcs(
        originaldata = df_imp,
        smtype = "coxph",
        smformula = smformula_processed,
        method = meths_smcfcs,
        rjlimit = rjlimit,
        numit = numit,
        m = 1L, # one imputation per kmi dataset
        ...
      )

      return(smcfcs_modif)
    })

    smcfcs_obj <- combine_smcfcs_objects(imps_loop)
  }

  # Do invisible and capture output bits here?
  return(smcfcs_obj)
}

library(smcfcs)
library(survival)
library(kmi)
context("Fine-Gray model testing")

# In style of test_coxph.r, probably need to re-write with mini simulations
test_that("Fine-Gray imputation is approximately unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(4321)
      n <- 10000
      z <- rnorm(n = n)
      x <- rnorm(n = n, mean = z)
      p <- 0.15
      betas <- c(0.75, 0.5)
      xb1 <- betas[1] * x - betas[2] * z # Fine-Gray linear predictor
      xb2 <- -x + 0.5 * z # Linear predictor for hazard conditional on cause 2
      d_tilde <- 1 + rbinom(n = n, size = 1, prob = (1 - p)^exp(xb1))
      cause1_ind <- d_tilde == 1
      U <- runif(n = n)
      t_tilde <- numeric(length = n)
      num <- 1 - (1 - U[cause1_ind] * (1 - (1 - p)^xb1[cause1_ind]))^(1 / xb1[cause1_ind])
      t_tilde[cause1_ind] <- -log(1 - num / p)
      t_tilde[!cause1_ind] <- -log(U[!cause1_ind]) / exp(xb2[!cause1_ind])
      cens <- -log(runif(n = n)) / 0.1
      t <- pmin(cens, t_tilde)
      d <- ifelse(cens < t_tilde, 0, d_tilde)
      x[runif(n) > 0.5] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.finegray(
        originaldata = simData,
        smformula = "Surv(t, d) ~ x + z",
        method = c("", "", "norm", "")
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, coxph(Surv(newtimes, newevent) ~ x + z))
      abs(summary(MIcombine(models))[1, 1] - betas[1]) < 0.1
    },
    TRUE
  )
})

test_that("Fine-Gray imputation works with no censoring", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(4321)
      n <- 10000
      z <- rnorm(n = n)
      x <- rnorm(n = n, mean = z)
      p <- 0.15
      betas <- c(0.75, 0.5)
      xb1 <- betas[1] * x - betas[2] * z # Fine-Gray linear predictor
      xb2 <- -x + 0.5 * z # Linear predictor for hazard conditional on cause 2
      d_tilde <- 1 + rbinom(n = n, size = 1, prob = (1 - p)^exp(xb1))
      cause1_ind <- d_tilde == 1
      U <- runif(n = n)
      t_tilde <- numeric(length = n)
      num <- 1 - (1 - U[cause1_ind] * (1 - (1 - p)^xb1[cause1_ind]))^(1 / xb1[cause1_ind])
      t_tilde[cause1_ind] <- -log(1 - num / p)
      t_tilde[!cause1_ind] <- -log(U[!cause1_ind]) / exp(xb2[!cause1_ind])
      x[runif(n) > 0.5] <- NA

      simData <- data.frame(t_tilde, d_tilde, x, z)

      imps <- smcfcs.finegray(
        originaldata = simData,
        smformula = "Surv(t_tilde, d_tilde) ~ x + z",
        method = c("", "", "norm", "")
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, coxph(Surv(newtimes, newevent) ~ x + z))
      abs(summary(MIcombine(models))[1, 1] - betas[1]) < 0.1
    },
    TRUE
  )
})

#' Simulated example data with continuous outcome and quadratic covariate effects
#'
#' A dataset containing simulated data where the outcome depends quadratically
#' on a partially observed covariate.
#'
#' @format A data frame with 1000 rows and 5 variables:
#' \describe{
#'   \item{y}{Continuous outcome}
#'   \item{z}{Fully observed covariate, with linear effect on outcome}
#'   \item{x}{Partially observed normally distributed covariate, with quadratic effect on outcome}
#'   \item{xsq}{The square of x, which thus has missing values also}
#'   \item{v}{An auxiliary variable (i.e. not contained in the substantive model)}
#' }
#'
"ex_linquad"

#' Simulated example data with continuous outcome and interaction between two partially observed covariates
#'
#' A dataset containing simulated data where the outcome depends on both main
#' effects and interaction of two partially observed covariates.
#'
#' @format A data frame with 1000 rows and 4 variables:
#' \describe{
#'   \item{y}{Continuous outcome}
#'   \item{x1}{Partially observed normally distributed covariate}
#'   \item{x2}{Partially observed binary covariate}
#' }
#'
"ex_lininter"

#' Simulated example data with binary outcome and quadratic covariate effects
#'
#' A dataset containing simulated data where the binary outcome depends quadratically
#' on a partially observed covariate.
#'
#' @format A data frame with 1000 rows and 5 variables:
#' \describe{
#'   \item{y}{Binary outcome}
#'   \item{z}{Fully observed covariate, with linear effect on outcome (on log odds scale)}
#'   \item{x}{Partially observed normally distributed covariate, with quadratic effect on outcome (on log odds scale)}
#'   \item{xsq}{The square of x, which thus has missing values also}
#'   \item{v}{An auxiliary variable (i.e. not contained in the substantive model)}
#' }
#'
"ex_logisticquad"

#' Simulated example data with count outcome, modelled using Poisson regression
#'
#' A dataset containing simulated data where the count outcome depends on two
#' covariates, x and z, with missing values in x. The substantive model is
#' Poisson regression.
#'
#' @format A data frame with 1000 rows and 3 variables:
#' \describe{
#'   \item{y}{Count outcome}
#'   \item{z}{Fully observed covariate, with linear effect on outcome}
#'   \item{x}{Partially observed normally distributed covariate, with linear effect on outcome}
#' }
#'
"ex_poisson"

#' Simulated example data with time to event outcome and quadratic covariate effects
#'
#' A dataset containing simulated data where a time to event outcome depends quadratically
#' on a partially observed covariate.
#'
#' @format A data frame with 1000 rows and 6 variables:
#' \describe{
#'   \item{t}{Time to event or censoring}
#'   \item{d}{Binary indicator of whether event occurred or individual was censored}
#'   \item{z}{Fully observed covariate, with linear effect on outcome (on log hazard scale)}
#'   \item{x}{Partially observed normally distributed covariate, with quadratic effect on outcome (on log hazard scale)}
#'   \item{xsq}{The square of x, which thus has missing values also}
#'   \item{v}{An auxiliary variable (i.e. not contained in the substantive model)}
#' }
#'
"ex_coxquad"

#' Simulated example data with competing risks outcome and partially observed covariates
#'
#' A dataset containing simulated competing risks data. There are two competing risks, and
#' some times are also censored.
#'
#' @format A data frame with 1000 rows and 4 variables:
#' \describe{
#'   \item{t}{Time to event or censoring}
#'   \item{d}{Indicator of whether event 1 occurred (d=1), event 2 occurred (d=2) or individual was censored (d=0)}
#'   \item{x1}{Partially observed binary covariate, with linear effects on log competing risk hazards}
#'   \item{x2}{Partially observed normally distributed (conditional on x1) covariate, with linear effects
#'   on log competing risk hazards}
#' }
#'
"ex_compet"

#' Simulated example data with competing risks outcome and partially observed covariates
#'
#' A dataset containing simulated competing risks data. There are two competing risks, and
#' some times are also censored. Proportionality holds on the subdistribution hazard scale for
#' cause 2, where for dataset `ex_compet` it instead holds on the cause-specific
#' hazard scale.
#'
#' @format A data frame with 1000 rows and 4 variables:
#' \describe{
#'   \item{times}{Time to event or censoring}
#'   \item{d}{Indicator of whether event 1 occurred (d=1), event 2 occurred (d=2) or individual was censored (d=0)}
#'   \item{x1}{Partially observed binary covariate, with linear effects on log subdistribution hazard of cause 1}
#'   \item{x2}{Partially observed normally distributed (conditional on x1) covariate, with linear effects
#'   on log subdistribution hazard of cause 1}
#' }
#'
"ex_finegray"

#' Simulated case cohort data
#'
#' A dataset containing simulated case cohort data, where the sub-cohort was a 10\% random sample of the full cohort.
#'
#' @format A data frame with 1571 rows and 7 variables:
#' \describe{
#'   \item{t}{Time to event or censoring}
#'   \item{d}{Indicator of whether event 1 occurred (d=1), or not (d=0)}
#'   \item{x}{Partially observed continuous covariate}
#'   \item{z}{Fully observed covariate}
#'   \item{in.subco}{A binary indicator of whether the subject is in the sub-cohort}
#'   \item{id}{An id variable}
#'   \item{entertime}{The entry time variable to be used in the analysis}
#' }
#'
"ex_cc"

#' Simulated nested case-control data
#'
#' A dataset containing simulated nested case-control data.
#'
#' @format A data frame with 728 rows and 8 variables:
#' \describe{
#'   \item{t}{Time to event or censoring}
#'   \item{d}{Indicator of whether event 1 occurred (d=1), or not (d=0)}
#'   \item{x}{Partially observed binary covariate}
#'   \item{z}{Fully observed covariate}
#'   \item{id}{An id variable}
#'   \item{numrisk}{Number of patients at risk at time of case's event}
#'   \item{setno}{The case-control set number}
#'   \item{case}{Binary indicator of case (=1) or control (=0)}
#' }
#'
"ex_ncc"

#' Simulated discrete time survival data set
#'
#' A dataset containing simulated discrete time survival data.
#'
#' @format A data frame with 1000 rows and 8 variables:
#' \describe{
#'   \item{x1}{A binary variable with missing values}
#'   \item{x2}{A fully observed continuous variable}
#'   \item{failtime}{The discrete failure/censoring time}
#'   \item{d}{Indicator of failure (=1) or censoring (=0)}
#' }
#'
"ex_dtsam"

#' Simulated example data with time-to-event Weibull outcome and two covariates
#'
#' A dataset containing simulated data where the time-to-event outcome is Weibull
#' distributed with two covariates, one of which is partially observed.
#'
#' @format A data frame with 1000 rows and 4 variables:
#' \describe{
#'   \item{t}{Time to event (d=1) or censoring (d=0)}
#'   \item{d}{Event indicator}
#'   \item{x}{Partially observed binary covariate}
#'   \item{z}{Fully observed continuous covariate}
#' }
#'
"ex_flexsurv"

#' Substantive model compatible fully conditional specification imputation of covariates and
#' event times using flexible parametric survival models
#'
#' Multiply imputes missing covariate values and event times using substantive model compatible
#' fully conditional specification with a Royston-Parmar flexible parametric survival model.
#'
#' This version of \code{smcfcs} is for time-to-event outcomes which are modelled
#' using a flexible parametric proportional hazards survival model, as proposed
#' by Royston and Parmar (2002). The model is
#' fitted using the \code{\link[flexsurv]{flexsurvspline}} function in the
#' \pkg{flexsurv} package. Specifically it fits models using the hazard scale. The
#' flexibility of the model can be changed by modifying the k argument, which
#' specifies the number of knots.
#'
#' If desired, \code{smcfcs.flexsurv} can be used to impute event times for individuals
#' who are originally censored, by specifying \code{imputeTimes=TRUE}. In the resulting
#' imputed datasets every individual will have an event time and the event indicator will
#' be one for all. Alternatively, you can impute censored times, but setting a larger
#' potential censoring time, which is either a common value used for all or a vector of times,
#' by using the \code{censtime} argument. If some individuals have their time-to-event
#' outcome completely missing and you want to impute this, they should have a time of zero
#' and the event indicator set to zero.
#'
#' \code{smcfcs.flexsurv} will not let you impute using norm, latnorm or poisson methods
#' for variables that are allowed to have time-varying effects, because the usual
#' rejection sampling bound used by smcfcs is not valid in this setting.
#'
#' \code{\link[flexsurv]{flexsurvspline}} sometimes fails during model fitting.
#' If/when this occurs, \code{smcfcs.flexsurv} takes a posterior draw based
#' on the model fit from the preceding iteration, and a warning is printed at
#' the end of the \code{smcfcs.flexsurv} run detailing how many times it occurred.
#'
#' @author Jonathan Bartlett \email{jonathan.bartlett1@@lshtm.ac.uk}
#'
#' @param smformula A formula of the form "Surv(t,d)~x+z"
#' @param k Number of knots to use in the flexible parametric survival model
#' @param imputeTimes If set to TRUE, \code{smcfcs.flexsurv} will impute
#' censored survival times, as well as any missing covariates
#' @param censtime Value(s) to use for censoring of imputed event times. If
#' a vector, it should be of length equal to the number of original censored
#' individuals
#' @param originalKnots If imputing censored event times, setting
#' \code{originalKnots=TRUE} means the automatically chosen knot locations
#' from the model fitted to the observed times are used throughout. If \code{FALSE},
#' knots are chosen automatically at each iteration by \code{flexsurvspline}
#' based on the current observed+imputed event times, according to the chosen
#' value of \code{k}.
#' @param ... Additional arguments to pass on to \link[smcfcs]{smcfcs}
#'
#' @inheritParams smcfcs
#' @example data-raw/flexsurv_example.r
#'
#' @references Royston P, Parmar MKB. Flexible parametric proportional-hazards
#' and proportional-odds models for censored survival data, with application
#' to prognostic modelling and estimation of treatment effects.
#' Statistics in Medicine 2002; 21(15): 2175-2197. \doi{doi:10.1002/sim.1203}
#'
#' @export
smcfcs.flexsurv <- function(originaldata, smformula, method, k=2, imputeTimes=FALSE,
                    censtime = NULL,
                    originalKnots = TRUE,
                    ...) {

  if (!requireNamespace("flexsurv", quietly = TRUE)) {
    stop("The package 'flexsurv' is required for this function. Please install it.", call. = FALSE)
  }

  smcfcs.core(originaldata=originaldata,
              smtype = "flexsurv",
              smformula = smformula,
              method = method,
              k=k,
              imputeTimes=imputeTimes,
              censtime=censtime,
              originalKnots=originalKnots,
              ...)
}

library(smcfcs)
library(survival)
context("Flexible parametric proportional hazards model testing")

test_that("Flexsurv imputation of missing normal covariate is approximately unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- z + rnorm(n)
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
        smformula = "Surv(t, d)~x+z",
        method = c("", "", "norm", ""),
        k=2
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, flexsurv::flexsurvspline(Surv(t, d) ~ x + z, k=2))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[5] < 1) & (MIcombineRes$`upper)`[5]>1)
    },
    TRUE
  )

})

test_that("Flexsurv imputation of missing binary covariate is approximately unbiased", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, flexsurv::flexsurvspline(Surv(t, d) ~ x + z, k=2))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[5] < 1) & (MIcombineRes$`upper)`[5]>1)
    },
    TRUE
  )

})

test_that("Flexsurv imputation of censored times checks", {
  # check that event indicator equals 1 for all
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE,
                              m=1
      )
      identical(mean(imps$impDatasets[[1]]$d),1)
    },
    TRUE
  )

})


test_that("Flexsurv imputation of missing binary covariate is approximately unbiased,
          imputing covariate and censored times", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      # note that in the following fits flexsurvspline will choose knots based
      # on the obs+imputed event times in the imputed datasets, rather than in the
      # original obs times (which is what smcfcs.flexsurv has done above)
      models <- with(impobj, flexsurv::flexsurvspline(Surv(t, d) ~ x + z, k=2))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[5] < 1) & (MIcombineRes$`upper)`[5]>1)
    },
    TRUE
  )

})



test_that("Flexsurv imputation, imputing covariate and censored times,
          passing new common censoring time. Check censoring is as specified", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      c <- rexp(n)
      d <- 1 * (t < c)
      t[d==0] <- c[d==0]
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE,
                              censtime=10,
                              m=1
      )
      identical(mean(imps$impDatasets[[1]]$t[imps$impDatasets[[1]]$d==0]),10)
    },
    TRUE
  )
})

test_that("Flexsurv imputation, imputing covariate and censored times,
          passing new vector of censorings. Check censoring is as specified", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      c <- rexp(n)
      d <- 1 * (t < c)
      t[d==0] <- c[d==0]
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE,
                              censtime=t[d==0]+0.1,
                              m=1
      )
      identical(mean(imps$impDatasets[[1]]$t[imps$impDatasets[[1]]$d==0]),
                mean(t[imps$impDatasets[[1]]$d==0]+0.1))
    },
    TRUE
  )
})

test_that("Flexsurv imputation, imputing censored times only. Check it runs.", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      c <- rexp(n)
      d <- 1 * (t < c)
      t[d==0] <- c[d==0]

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "", ""),
                              k=2,
                              imputeTimes=TRUE,
                              censtime=t[d==0]+0.1,
                              m=1
      )
      identical(mean(imps$impDatasets[[1]]$t[imps$impDatasets[[1]]$d==0]),
                mean(t[imps$impDatasets[[1]]$d==0]+0.1))
    },
    TRUE
  )
})

test_that("Flexsurv imputation of missing binary covariate is approximately unbiased,
          imputing covariate and censored times, originalKnots=FALSE", {
  skip_on_cran()
  expect_equal(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE,
                              originalKnots=FALSE
      )
      library(mitools)
      impobj <- imputationList(imps$impDatasets)
      models <- with(impobj, flexsurv::flexsurvspline(Surv(t, d) ~ x + z, k=2))
      MIcombineRes <- summary(MIcombine(models))
      # 95% CI includes true value
      (MIcombineRes$`(lower`[5] < 1) & (MIcombineRes$`upper)`[5]>1)
    },
    TRUE
  )

})

test_that("Flexsurv imputation errors if NA in event time variable", {
            skip_on_cran()
            expect_error(
              {
                set.seed(1234)
                n <- 1000
                z <- rnorm(n)
                x <- 1*(runif(n)<plogis(z))
                t <- -log(runif(n)) / (1 * exp(x + z))
                d <- 1 * (t < 10)
                t[d == 0] <- 10
                x[(runif(n) < 0.5)] <- NA
                t[1] <- NA

                simData <- data.frame(t, d, x, z)

                imps <- smcfcs.flexsurv(simData,
                                        smformula = "Surv(t, d)~x+z",
                                        method = c("", "", "logreg", ""),
                                        k=2,
                                        imputeTimes=TRUE,
                                        originalKnots=FALSE
                )
              }
            )

})

test_that("Flexsurv imputation errors if NA in event indicator variable", {
  skip_on_cran()
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA
      d[1] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z",
                              method = c("", "", "logreg", ""),
                              k=2,
                              imputeTimes=TRUE,
                              originalKnots=FALSE
      )
    }
  )

})


test_that("Flexsurv imputation errors if time-varying effect used for fully obs cts cov", {
  skip_on_cran()
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      z[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z+gamma1(z)",
                              method = c("", "", "", "norm"),
                              k=2
      )
    }
    )

})

test_that("Flexsurv imputation doesn't error if time-varying effect used for fully obs cts cov", {
  skip_on_cran()
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z+gamma1(z)",
                              method = c("", "", "logreg", ""),
                              k=2
      )
    }
  , NA)

})


test_that("Flexsurv imputation doesn't error if time-varying effect used for partial binary cov", {
  skip_on_cran()
  expect_error(
    {
      set.seed(1234)
      n <- 1000
      z <- rnorm(n)
      x <- 1*(runif(n)<plogis(z))
      t <- -log(runif(n)) / (1 * exp(x + z))
      d <- 1 * (t < 10)
      t[d == 0] <- 10
      x[(runif(n) < 0.5)] <- NA

      simData <- data.frame(t, d, x, z)

      imps <- smcfcs.flexsurv(simData,
                              smformula = "Surv(t, d)~x+z+gamma1(x)",
                              method = c("", "", "logreg", ""),
                              k=2
      )
    }
    , NA)

})

#' Assess convergence of a smcfcs object
#'
#' Visualises the contents of smCoefIter. Specifically, it plots the parameter
#' estimates of the substantive model against the number of iterations from
#' the imputation procedure. This is done for each regression coefficient,
#' and each line corresponds to an imputed dataset.
#'
#' Requires loading of ggplot2 plotting library.
#'
#' @author Edouard F. Bonneville \email{e.f.bonneville@@lumc.nl}
#'
#' @param x An object of class 'smcfcs'
#' @param include Character vector of coefficient names for which to return the
#' convergence plot. Default is "all" and returns plots for all coefficients in
#' a facetted manner.
#'
#' Recommendation is to plot first with include = "all", and then select
#' coefficient names to zoom in to.
#'
#' For competing risks, the coefficients are indexed by their cause. E.g. for
#' coefficient of a variable x1 in a model for cause 2, will be labelled
#' "x1-cause2".
#' @param ... Additional parameters to pass on to ggplot2::facet_wrap(),
#' eg. nrow = 2
#'
#' @return A ggplot2 object, containing the convergence plots, facetted per
#' covariate in the substantive model
#' @export
#'
#' @examples
#' \dontrun{
#' # Use simulated competing risks example in package
#' imps <- smcfcs(
#'   originaldata = ex_compet,
#'   smtype = "compet",
#'   smformula = list(
#'     "Surv(t, d == 1) ~ x1 + x2",
#'     "Surv(t, d == 2) ~ x1 + x2"
#'   ),
#'   method = c("", "", "norm", "norm")
#' )
#'
#' plot(imps)
#' plot(imps, include = c("x1-cause1", "x2-cause2"))
#' }
#'
#' @importFrom rlang .data
plot.smcfcs <- function(x,
                        include = "all",
                        ...) {
  if (!inherits(x, "smcfcs")) {
    stop("'x' must be a 'smcfcs' object")
  }

  if (!requireNamespace("ggplot2", quietly = TRUE)) {
    stop("Package ggplot2 needed for this function to work. Please install it.",
      call. = FALSE
    )
  }

  # Prepare data
  df_plot <- prep_iters(x)

  # Choose plots to include
  if (length(include) >= 1 & include[1] != "all") {
    coef_names <- unique(df_plot$covar)
    if (any(!(include %in% coef_names))) {
      mssg <- paste0(
        "include should be character vector containing any combination of: '",
        paste0(coef_names, collapse = "','"),
        "'; or simply 'all'"
      )
      stop(mssg)
    } else {
      df_plot <- df_plot[df_plot$covar %in% include, ]
    }
  }

  # Make plot
  p <- ggplot2::ggplot(
    data = df_plot,
    ggplot2::aes(
      x = .data$iters,
      y = .data$value,
      col = factor(.data$imp)
    )
  ) +
    ggplot2::geom_line() +
    ggplot2::theme(legend.position = "none") +
    ggplot2::labs(x = "Iterations", y = "Coefficient") +
    ggplot2::facet_wrap(~covar, ...)

  return(p)
}


# Prepare data for plotting
prep_iters <- function(x) {
  # Extract meta data
  M <- dim(x$smCoefIter)[1]
  smtype <- x$smInfo$smtype
  smformula <- x$smInfo$smformula
  dat <- x$impDatasets[[1]] # for names in model matrix
  numit <- dim(x$smCoefIter)[3]

  if (numit < 2) {
    stop("Re-run smcfcs() with numit >= 2 in order to assess convergence")
  }

  # Check if competing risks
  coef_names <- if (smtype == "compet") {
    K <- length(smformula)
    cause_coef_names <- lapply(X = seq_len(K), FUN = function(k) {
      names_mod <- get_coef_names(smformula[k], dat, intercept = FALSE)
      paste0(names_mod, "-cause", as.character(k))
    })

    unlist(cause_coef_names)
  } else if (smtype %in% c("weibull", "coxph", "casecohort", "nestedcc")) {
    get_coef_names(smformula, dat, intercept = FALSE)
  } else if (smtype == "dtsam") {
    get_dtsam_names(smformula, dat, x$extraArgs$timeEffects)
  } else {
    get_coef_names(smformula, dat, intercept = TRUE)
  }

  # Prepare df for plotting
  ests_list <- lapply(X = seq_len(M), function(m) {
    coef_dat <- as.data.frame(t(x$smCoefIter[m, , ]))
    coef_dat$iters <- seq_len(numit)
    coef_dat$imp <- m

    return(coef_dat)
  })

  ests_df <- do.call(rbind.data.frame, ests_list)
  colnames(ests_df) <- c(coef_names, "iters", "imp")

  # Make long format
  ests_long <- stats::reshape(
    data = ests_df,
    varying = coef_names,
    timevar = "covar",
    v.names = "value",
    idvar = c("imp", "iters"),
    direction = "long",
    times = coef_names
  )

  return(ests_long)
}

get_dtsam_names <- function(smformula,
                            dat,
                            timeEffects) {
  # Get sides of formula
  rhs <- gsub(x = smformula, pattern = ".*~", replacement = "")
  lhs <- gsub(x = smformula, pattern = "~.*", replacement = "")
  surv_obj <- with(dat, eval(parse(text = lhs)))
  time_var <- as.matrix(surv_obj)[, "time"]

  # Make longdata
  cutPoints <- seq_len(max(time_var)) # Probably change to unique timepoints later
  longData <- survival::survSplit(as.formula(smformula), data = dat, cut = cutPoints)

  # Get model matrix formula
  smformula_matrix <- if (timeEffects == "factor") {
    paste0("~ -1 + factor(tstart) + ", rhs)
  } else if (timeEffects == "linear") {
    paste0("~ tstart + ", rhs)
  } else {
    paste0("~ tstart + I(tstart^2) + ", rhs)
  }

  # Make matrix
  model_mat <- stats::model.matrix(
    object = as.formula(smformula_matrix),
    data = longData
  )

  coef_names <- colnames(model_mat)

  return(coef_names)
}

get_coef_names <- function(smformula,
                           dat,
                           intercept) {
  rhs <- gsub(x = smformula, pattern = ".*~", replacement = "")
  smformula_matrix <- as.formula(paste0("~ +", rhs))

  # Check if there is stratification - if so remove from model matrix
  # (stratification means different baseline hazards, coefficients still same)
  if (grepl(x = rhs, pattern = "strata")) {
    strata_var <- gsub(x = rhs, pattern = ".*\\(|\\).*", replacement = "")
    rm_strata <- as.formula(paste0("~ . - strata(", strata_var, ")"))
    smformula_matrix <- update(smformula_matrix, rm_strata)
  }

  model_mat <- stats::model.matrix(
    object = smformula_matrix,
    data = dat
  )

  # For survival models
  if (intercept == FALSE) {
    model_mat <- model_mat[, !(colnames(model_mat) %in% "(Intercept)")]
  }

  coef_names <- colnames(model_mat)

  return(coef_names)
}

---
title: "smcfcs for covariate measurement error correction"
author: "Jonathan Bartlett"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{smcfcs_measerror}
  %\VignetteEngine{knitr::rmarkdown}
  %\usepackage[utf8]{inputenc}
---

This short vignette introduces the capabilities of `smcfcs` to accommodate classical covariate measurement error. We consider the cases where internal validation data and then internal replication data are available.

# Validation data
We will simulate a dataset with internal validation data where the true covariate (x) is observed for 10\% of the sample, while every subject has an error-prone measurement (w) observed:

```{r}
set.seed(1234)
n <- 1000
x <- rnorm(n)
w <- x + rnorm(n)
y <- x + rnorm(n)
x[(n * 0.1):n] <- NA
simData <- data.frame(x, w, y)
```

We have generated code where the error-prone measurement w is equal to the true covariate x plus some independent normally distributed measurement error. Since x is observed for some of the subjects in the case of interval validation data, this is a regular missing data problem. The error-prone measurement w serves as an auxiliary variable for the purposes of imputation of x. In particular, we will impute using `smcfcs' such that w is not in the substantive model. This encodes the so called non-differential error assumption, that says that conditional on x, the error-prone measurement w provides no independent information about the outcome y. An initial attempt to do this is:

```{r}
library(smcfcs)
imps <- smcfcs(simData,
  smtype = "lm", smformula = "y~x",
  method = c("norm", "", ""), m = 5
)
```

We see from the output that `smcfcs` has not mentioned that it is using w anywhere. This is because w is fully observed and is not involved in the substantive model. To force w to be conditioned on when imputing x, we must pass an appropriate `predictorMatrix` to `smcfcs`:
```{r}
predMat <- array(0, dim = c(3, 3))
predMat[1, 2] <- 1
```
We have specified that the first variable, x, be imputed using w. Note that we do not need to tell `smcfcs` to impute x using y, as this will occur automatically by virtue of y being the outcome variable in the substantive model. We can now impute again, passing `predMat` as the `predictorMatrix`:
```{r}
imps <- smcfcs(simData,
  smtype = "lm", smformula = "y~x",
  method = c("norm", "", ""), m = 5,
  predictorMatrix = predMat
)
```

Now we can fit the substantive model to each imputed dataset and use the `mitools` package to pool the estimates and standard errors using Rubin's rules:
```{r}
library(mitools)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ x))
summary(MIcombine(models))
```

We note from the results that the fraction of missing information for the coefficient of x is high. This should not surprise us, given that x was missing for 90\% of the sample and the error-prone measurement w is quite a noisy measure of x.

# Replication data
We will now demonstrate how `smcfcs` can be used to impute a covariate x which is not observed for any subjects, but we have for at least a subset of the sample two or more error-prone replicate measurements. We first simulate the dataset:
```{r}
x <- rnorm(n)
w1 <- x + rnorm(n)
w2 <- x + rnorm(n)
w2[(n * 0.1):n] <- NA
y <- x + rnorm(n)
x <- rep(NA, n)
simData <- data.frame(x, w1, w2, y)
```
Note that now x is missing for every subject. Every subject has an error-prone measurement w1 of x, and 10\% of the sample have a replicated measurement w2.

We will now impute x using `smcfcs`. To do this we specify that x be imputed using the `latnorm` method. In addition, we pass a matrix to the `errorProneMatrix` argument of `smcfcs`, whose role is to specify, for each latent normal variable to be imputed, which variables in the data frame are error-prone measurements. `smcfcs` then imputes the missing values in x, assuming a normal classical error model for the error-prone replicates.

```{r}
errMat <- array(0, dim = c(4, 4))
errMat[1, c(2, 3)] <- 1
imps <- smcfcs(simData,
  smtype = "lm", smformula = "y~x",
  method = c("latnorm", "", "", ""), m = 5,
  errorProneMatrix = errMat
)
```

Analysing the imputed datasets, we obtain:
```{r}
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ x))
summary(MIcombine(models))
```

If we summarise one of the imputed datasets (below), we will see that `smcfcs` has not only imputed the missing values in x, but also the 'missing' values in w2. We hyphenate missing here because typically a study with replicate error-prone measurements will have intentionally planned to only take a second error-prone measurement on a random subset, so the values were never intended to be measured.
```{r}
summary(imps$impDatasets[[1]])
```

One thing to be wary of when imputing covariates measured with error, particularly with replication data, is that convergence may take longer than in the regular missing data setting. To examine this, we re-impute one dataset using 100 iterations, and then plot the estimates against iteration number:
```{r, fig.width=6}
imps <- smcfcs(simData,
  smtype = "lm", smformula = "y~x",
  method = c("latnorm", "", "", ""), m = 1, numit = 100,
  errorProneMatrix = errMat
)
plot(imps)
```

This plot suggests it would probably be safer to impute using slightly more than 10 iterations per imputation.

## Multiple covariates measured with error
`smcfcs` can impute multiple covariates measured with error when internal replication data are available. It allows for a separate error variance for each such covariate. The following code adds a second covariate which is itself measured by two error-prone measurements, but this time with a smaller error variance. It then defines the `errorProneMatrix`, imputes and analyses the imputed datasets:
```{r}
x <- rnorm(n)
x1 <- x + rnorm(n)
x2 <- x + rnorm(n)
w2[(n * 0.1):n] <- NA
z <- x + rnorm(n)
z1 <- z + 0.1 * rnorm(n)
z2 <- z + 0.1 * rnorm(n)
y <- x - z + rnorm(n)
x <- rep(NA, n)
z <- rep(NA, n)
simData <- data.frame(x, x1, x2, z, z1, z2, y)

errMat <- array(0, dim = c(7, 7))
errMat[1, c(2, 3)] <- 1
errMat[4, c(5, 6)] <- 1
imps <- smcfcs(simData,
  smtype = "lm", smformula = "y~x+z",
  method = c("latnorm", "", "", "latnorm", "", "", ""), m = 5,
  errorProneMatrix = errMat
)
```

We now analyse the imputed datasets, remembering to add z into the substantive model:
```{r}
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ x + z))
summary(MIcombine(models))
```
We see that the fraction of missing information is lower for z than for x. This is a consequence of the fact that we generated the error-prone measurements of z to have smaller error variance than for the corresponding error-prone measurements of x.

---
title: "smcfcs for coarsened factor covariates"
author: "Jonathan Bartlett and Lars van der Burg"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{coarsening}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

`smcfcs` was originally created to create multiple imputations of missing values of covariates in regression models. As of 2025, it has functionality to impute unobserved values of factor variables which are 'coarsened', based on the developments in [van der Burg et al (2025)](https://doi.org/10.1002/sim.70032). By coarsened, we mean that for some of the missing values, some partial information about the value is known - we know that the value belongs to some subset of the possible values. In this vignette we demonstrate the functionality of `smcfcs` for imputing such variables.

To demonstrate how to do this, we illustrate using the dataset `ex_coarsening` that is in the `smcfcs` package:

```{r setup}
library(smcfcs)
summary(ex_coarsening)
head(ex_coarsening)
```

The variable `x` is a factor variable which has `r sum(is.na(ex_coarsening$x))` missing values. The variable `xobs` gives the known information about (some of) the missing values:

```{r}
table(ex_coarsening$x,ex_coarsening$xobs,useNA = "ifany")
```
From this we can see that among the `r sum(is.na(ex_coarsening$x))` missing values in `x`, for `r sum(ex_coarsening$xobs=="a/c")` individuals we know that their value for `x` was either a or c, as indicated by the string 'a/c', `r sum(ex_coarsening$xobs=="b/c")` individuals we know that their value for `x` was either b or c, as indicated by the string 'b/c', while for the remainder we have no further information, indicated by the character string "NA".

*Note: the variable `xobs` is a character variable, and for rows where `x` is (plain) missing, `xobs` takes the character value "NA", rather than R's missing value indicator NA. This is important, since if we used the missing value indicator NA, `smcfcs` would refused to run as we have not told it how to impute the missing values in `xobs`.*

In order to impute the missing values in `x` using `smcfcs` we have to define a value for the `restrictions` argument. For this we must pass a list of length equal to the number of variables in the data frame. For the element in this list corresponding to `x` we must give a vector of formula typ expressions to specify the possible values for `x` when `xobs` equals a/c or b/c. To achieve this we use:
```{r}
restrictionsX = c("xobs = a/c ~ a + c",
                  "xobs = b/c ~ b + c")
restrictions = append(list(restrictionsX), as.list(c("", "", "")))
```

We can then impute the missing values accounting for the partial information with:
```{r, results=FALSE, warning=FALSE}
set.seed(68204812)
imps <- smcfcs(originaldata=ex_coarsening,
               smtype="lm",
               smformula = "y~z+x",
               method = c("mlogit","", "", ""),
               restrictions = restrictions
)
```

To check that `smcfcs` has correctly used the partial information about the missing values in `x`, first we check the first few rows in the first imputed dataset:
```{r}
head(imps$impDatasets[[1]])
```
This looks fine - when `xobs=a/c` we have imputed values either of a or c, whereas when `xobs=b/c` we have imputed values of b or c. To check properly, we can repeat the earlier cross-tabulation:
```{r}
table(imps$impDatasets[[1]]$x,imps$impDatasets[[1]]$xobs,useNA = "ifany")
```
This shows that (at least in the first imputed dataset) the imputed values respect the partial information contained in `xobs`, as desired.

The `restrictions` argument can also be used for ordered factor variables in the same way.

---
title: "smcfcs"
author: "Jonathan Bartlett"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{smcfcs}
  %\VignetteEngine{knitr::rmarkdown}
  %\usepackage[utf8]{inputenc}
---

# Introduction

Missing data are a common issue in many fields of empirical research. An popular approach to handling missing data is the method of multiple imputation (MI). Multiple imputation involves replacing missing values by a number of imputations, creating multiple imputed datasets. Each completed dataset is then analysed as usual, and estimates and standard errors are combined across imputations using rules developed by Rubin.

The most popular approach to imputation uses parametric models for the missing variables given the observed. Multiple imputation gives valid inferences provided that the missing data satisfy the so called missing at random (MAR) assumption and that the imputation models used are correctly specified.

## Joint model and FCS multiple imputation

When multiple variables are affected by missingness, the traditional approach to imputation is to specify a joint (or multivariate model) for the partially observed variables. One of the earliest examples of this was MI  using the multivariate normal model. Rather than specifying a joint model directly, a popular alternative is the fully conditional specification (FCS), or chained equations approach. In FCS MI separate conditional models are specified for each partially observed variable. In each of these conditional models, by default all of the variables serve as predictors. For an overview of the FCS MI and an implementation of it in R, see <a href="https://doi.org/10.18637/jss.v045.i03">van Buuren and Groothuis-Oudshoorn</a>.

## Imputation model compatibility

When missing values are imputed from a misspecified model, in general invalid inferences will result. One way in which misspecification can occur is when the imputation and substantive (analysis) model of interest are incompatible. Loosely speaking, this means there exists no joint model which contains the imputation model and the substantive model as the corresponding conditionals. In this case, as described by <a href="https://doi.org/10.1177/0962280214521348">Bartlett *et al* (2015)</a>, assuming that the substantive model is correctly specified, unless the imputation and substantive models can be made compatible by imposing a restriction on the imputation model, incompatibility implies the imputation model is misspecified.

Such incompatibility between the imputation model used to impute a partially observed covariate and the substantive/outcome model can arise for example when the latter includes interactions or non-linear effects of variables. A further example is when the substantive model is a Cox proportional hazards model for a censored time to event outcome. In these cases, it may be difficult or impossible to specify an  imputation model for a covariate which is compatible with the model for the outcome (the substantive model) using standard imputation models as available in existing packages.

# Substantive Model Compatible Fully Conditional Specification multiple imputation

The substantive model compatible modification of FCS MI (SMC-FCS), proposed by <a href="https://doi.org/10.1177/0962280214521348">Bartlett *et al* (2015)</a>, ensures that each partially observed variable is imputed from an imputation model which is compatible with a user specified model for the outcome (which is typically the substantive model of interest, although see below regarding auxiliary variables). As described in further detail in the linked paper, for each partially observed variable, e.g. `x1`, in SMC-FCS a model is specified for the conditional distribution of `x1` given the other partially observed variables `x2,x3,..,xp` and fully observed covariates `z`. This, together with the specified substantive model (a model for the outcome `y`) defines an imputation model for `x1` which is guaranteed to be compatible with this specified substantive model.

## Sampling from the imputation distribution
Unfortunately, the resulting imputation model for each partially observed variable generally does not belong to a standard parametric family, complicating the imputation of missing values. To overcome this, `smcfcs` uses the method of rejection sampling, which is more computationally intensive than direct sampling methods.

## Statistical properties
SMC-FCS ensures compatibility between each partially observed covariate's imputation model with the substantive model. However, when there is more than one partially observed variable, it does not guarantee that the corresponding different imputation models are mutually compatible. Consequently, as described further by <a href="https://doi.org/10.1177/0962280214521348">Bartlett *et al* (2015)</a>, only in special cases does SMC-FCS generate imputations from a well defined Bayesian joint model. Nonetheless, by ensuring compatibility between each partially observed variable's imputation model and the substantive model, it arguably overcomes (compared to standard FCS MI) the type of model incompatibility which is most likely to adversely affect inferences.

## When SMC-FCS may be preferable to FCS/MICE

In certain situations it may be advantageous to use SMC-FCS rather than traditional FCS MI. Important examples, as mentioned previously, include situations where the substantive (outcome) model includes interactions or non-linear effects of some of the covariates, or where the outcome model is itself non-linear, such as a Cox proportional hazards model. See <a href="https://doi.org/10.1177/0962280214521348">Bartlett *et al* (2015)</a> for simulation results comparing the two approaches in these situations.

# The `smcfcs` package
The `smcfcs` function in the `smcfcs` package implements the SMC-FCS procedure. Currently linear, logistic and Cox proportional hazards substantive models. Competing risks outcome data can also be accommodated, with a Cox proportional hazards model used to model each cause specific hazard function. Partially observed variables can be imputed using normal linear regression, logistic regression (for binary variables), proportional odds regression (sometimes known as ordinal logistic regression, suitable for ordered categorical variables), multinomial logistic regression (for unordered categorical variables), and Poisson regression (for count variables). In the following we describe some of the important aspects of using `smcfcs` by way of an example data frame.

## Example - linear regression substantive model with quadratic covariate effects
To illustrate the package, we use the simple example data frame `ex_linquad`, which is included with the package. This data frame was simulated for `n=1000` independent rows. For each row, variables `y,x,z,v` were intended to be collected, but there are missing values in `x`. The values have been made artificially missing, with the probability of missingness dependent on (the fully observed) `y` variable. Below the first 10 rows of the data frame are shown:

```{r}
library(smcfcs)
ex_linquad[1:10, ]
```

As shown, the `xsq` variable is equal to the square of the `x` variable. Since the latter has missing values, so does the former. We now impute the missing values in `x` and `xsq`, compatibly with a substantive model for the outcome `y` which is specified as a linear regression, with `z`, `x` and `xsq` as covariates:

```{r}
set.seed(123)
# impute missing values in x, compatibly with quadratic substantive model
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""))
```

As demonstrated here, the minimal arguments to pass to `smcfcs` are the data frame to be used, the substantive model type, the substantive model formula, and a method vector. The substantive model type specifies whether the model for the outcome is linear, logistic or Cox regression, or a competing risks analysis (see documentation). The `smformula` specifies the linear predictor of the substantive/outcome model. Here we specified that the outcome `y` is assumed to follow a linear regression model, with `z`, `x` and `xsq` as predictors.

Lastly, we passed a vector of strings as the `method` argument. This specifies, for each column in the data frame, the method to use for imputation. As in the example, empty strings should be passed for those columns which are fully observed and thus are not to be imputed. For `x` we specify `norm`, in order to impute using a normal linear regression model. See the help for `smcfcs` for the syntax for other imputation model types. For `xsq` we specify `"x^2"` as the imputation method. This instructs `smcfcs` to impute `xsq` by simply squaring the imputed values of `x`. Such a specification could also be used with the `mice` package, which implements standard FCS MI. Note however that here, through specifying the substantive model as including an effect of `xsq`, `smcfcs` is imputing the missing values in `x` which allows for a quadratic effect on `y`.

Having generated the imputed datasets, we can now fit our substantive model of interest. Here we make use of the `mitools` package to fit our substantive model to each imputed dataset, collect the results, and combine them using Rubin's rules:
```{r}
# fit substantive model
library(mitools)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))
```

Here the data were simulated such that the coefficients of `z`, `x` and `xsq` are all 1. The estimates we have obtained are (reassuringly) close to these true parameter values. To illustrate the dangers of imputing a covariate using an imputation model which is not compatible with the substantive model, we now re-impute `x`, but this time imputing compatibly with a model for `y` which does not allow for the quadratic effect:

```{r}
# impute missing values in x, compatibly with model for y which omits the quadratic effect
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x", method = c("", "", "norm", "x^2", ""))
```

`smcfcs` has issued some warnings about rejection sampling. We discuss this later in this vignette, but here we will continue and proceed to fit a model for `y` which includes both `x` and `xsq` (plus `z`) as covariates:

```{r}
# fit substantive model
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))
```

Now we have an estimate of the coefficient of `xsq` of 0.60, which is considerably smaller than the true value 1 used to simulate the data. This bias is due to the imputation model we have just used for `x` being misspecified. In particular, it was misspecified due to the fact it wrongly assumed a linear dependence of `y` on `x`, rather than allowing a quadratic dependence.

## Imputing using auxiliary variables with `smcfcs`
One of the strengths of multiple imputation in general is the possibility to use variables in imputation models which are subsequently not involved in the substantive model. This may be useful in order to condition or adjust for variables which are predictive of missingness, but which are not used in the substantive model of interest. Moreover, adjusting for auxiliary variables which are strongly correlated with one or more variables which are being imputed improves efficiency.

When using `smcfcs` to impute missing covariates, auxiliary variables `v` can be included by adding them as an additional covariate in the substantive model, as passed using the `smformula` argument. Here we are imputing `x` compatibly with a certain specification of model for the outcome. Our substantive model of interest is then a simpler model which omits `v`. For example, in the quadratic example dataset, we can add the auxiliary variable `v` using:
```{r}
# impute, including v as a covariate in the substantive/outcome model
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq+v", method = c("", "", "norm", "x^2", ""))
# fit substantive model, which omits v
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))
```
For outcome models other than linear regression, this approach is not entirely justifiable due to the lack of collapsibility of non-linear models. For example, if a Cox model is assumed for a failure time given variables `x` and `v`, the hazard function given only `x` (i.e. omitting `v` from the model) is no longer a Cox model. Further research is warranted to explore how this might affect the resulting inferences.

It is also possible to include the auxiliary variable `v` without adding it to the outcome model (as given in the `smformula` argument), through specification of the `predictorMatrix` argument. Doing so conditions on `v`, but assumes that the outcome is independent of `v`, conditional on whatever covariates are specified in `smformula`. This should thus only be used when the latter assumption is justified. When it is, inferences will in general be more efficient. To make this assumption when imputing `x` in the `ex_linquad` data, we define a `predictorMatrix` which will specify that `x` be imputed using both `z` and `v`, but we omit `v` from the `smformula` argument: 
```{r}
predMatrix <- array(0, dim = c(ncol(ex_linquad), ncol(ex_linquad)))
predMatrix[3, ] <- c(0, 1, 0, 0, 1)
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""), predictorMatrix = predMatrix)
impobj <- imputationList(imps$impDatasets)
models <- with(impobj, lm(y ~ z + x + xsq))
summary(MIcombine(models))
```

## Rejection sampling warnings

Sometimes when running `smcfcs` you may receive warnings that the rejection sampling that `smcfcs` uses has failed to draw from the required distribution on a couple of occasions. Upon receiving this warning, it is generally good idea to re-run `smcfcs`, specifying a value for `rjlimit` which is larger than the default, until the warning is no longer issued. Having said that, when only a small number of warnings are issued, it may be fine to ignore the warnings, especially when the dataset is large.

## Assessing convergence
Like standard chained equations or FCS imputation, the SMC-FCS algorithm must be run for a sufficient number of iterations for the process to converge to its stationary distribution. The default number of iterations used is 10, but this may not be sufficient in any given dataset and model specification To assess convergence, the object returned by `smcfcs` includes an object called `smCoefIter`. This matrix contains the parameter estimates of the substantive model, and is indexed by imputation number, parameter number, and iteration number. To assess convergence, one can call smcfcs with `m=1` and `numit` suitably chosen (e.g. `numit=100`). The values in the resulting smCoefIter matrix can then be plotted to assess convergence. To illustrate, we re-run the imputation model used previously with the example data, but asking for only `m=1` imputation to be generated, and with 100 iterations.

```{r, fig.width = 6, fig.height = 4}
# impute once with a larger number of iterations than the default 10
imps <- smcfcs(ex_linquad, smtype = "lm", smformula = "y~z+x+xsq", method = c("", "", "norm", "x^2", ""), predictorMatrix = predMatrix, m = 1, numit = 100)
# plot estimates of the parameters of the substantive model against iteration number
plot(imps)
```

The plot shows that the process appears to converge rapidly, such that the default choice of `numit=10` is probably fine here.

<h2 id="references">References</h3>

Bartlett JW, Seaman SR, White IR, Carpenter JR. Multiple imputation of covariates
by fully conditional specification: accommodating the substantive model. 
<a href="https://doi.org/10.1177/0962280214521348">Statistical Methods in Medical Research</a>, 2015; 24(4):462-487

van Buuren S, Groothuis-Oudshoorn K. mice:  Multivariate Imputation by Chained
Equations in R. <a href="https://doi.org/10.18637/jss.v045.i03">Journal of Statistical Software</a>, 2011; 45(3)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_coarsening}
\alias{ex_coarsening}
\title{Simulated example data with a coarsened factor covariate}
\format{
A data frame with 100 rows and 4 variables:
\describe{
  \item{x}{Factor variable with 3 levels (a,b,c) and some missing values}
  \item{xobs}{Variable which contains coarsening information for x}
  \item{z}{Fully observed continuous covariate}
  \item{y}{Fully observed continuous outcome}
}
}
\usage{
ex_coarsening
}
\description{
A simulated dataset with a factor level covariate subject to coarsening.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.r
\name{smcfcs.casecohort}
\alias{smcfcs.casecohort}
\title{Substantive model compatible fully conditional specification imputation of covariates for case cohort studies}
\usage{
smcfcs.casecohort(originaldata, smformula, method, sampfrac, in.subco, ...)
}
\arguments{
\item{originaldata}{The case-cohort data set (NOT a full cohort data set with a case-cohort substudy within it)}

\item{smformula}{A formula of the form "Surv(entertime,t,d)~x", where d is the event (d=1) or censoring (d=0) indicator, t is the event or censoring time and entertime is equal to the time origin (typically 0) for individuals in the subcohort and is equal to (t-0.001) for cases outside the subcohort [this sets cases outside the subcohort to enter follow-up just before their event time. The value 0.001 may need to be modified depending on the time scale.]}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{sampfrac}{The proportion of individuals from the underlying full cohort who are in the subcohort}

\item{in.subco}{The name of a column in the dataset with 0/1s that indicates whether the subject is in the subcohort}

\item{...}{Additional arguments to pass on to \link[smcfcs]{smcfcs}}
}
\description{
Multiply imputes missing covariate values using substantive model compatible
fully conditional specification for case cohort studies.
}
\details{
This version of \code{smcfcs} is designed for use with case cohort studies but where the analyst does not wish to,
or cannot (due to not having the necessary data) impute the full cohort. The function's arguments are the same
as for the main smcfcs function, except for \code{smformula}, \code{in.subco}, and \code{sampfrac} - see above
for details on how these should be specified.
}
\examples{
#the following example is not run when the package is compiled on CRAN
#(to keep computation time down), but it can be run by package users
\dontrun{
  #as per the documentation for ex_cc, the sampling fraction is 10\%
  imps <- smcfcs.casecohort(ex_cc, smformula="Surv(entertime, t, d)~x+z", sampfrac=0.1,
                            in.subco="in.subco", method=c("", "", "norm", "", "", "", ""))
  library(mitools)
  impobj <- imputationList(imps$impDatasets)
  models <- with(impobj, coxph(Surv(entertime,t,d)~x+z+cluster(id)))
  summary(MIcombine(models))
}
}
\author{
Ruth Keogh \email{ruth.keogh@lshtm.ac.uk}

Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.flexsurv.r
\name{smcfcs.flexsurv}
\alias{smcfcs.flexsurv}
\title{Substantive model compatible fully conditional specification imputation of covariates and
event times using flexible parametric survival models}
\usage{
smcfcs.flexsurv(
  originaldata,
  smformula,
  method,
  k = 2,
  imputeTimes = FALSE,
  censtime = NULL,
  originalKnots = TRUE,
  ...
)
}
\arguments{
\item{originaldata}{The original data frame with missing values.}

\item{smformula}{A formula of the form "Surv(t,d)~x+z"}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{k}{Number of knots to use in the flexible parametric survival model}

\item{imputeTimes}{If set to TRUE, \code{smcfcs.flexsurv} will impute
censored survival times, as well as any missing covariates}

\item{censtime}{Value(s) to use for censoring of imputed event times. If
a vector, it should be of length equal to the number of original censored
individuals}

\item{originalKnots}{If imputing censored event times, setting
\code{originalKnots=TRUE} means the automatically chosen knot locations
from the model fitted to the observed times are used throughout. If \code{FALSE},
knots are chosen automatically at each iteration by \code{flexsurvspline}
based on the current observed+imputed event times, according to the chosen
value of \code{k}.}

\item{...}{Additional arguments to pass on to \link[smcfcs]{smcfcs}}
}
\description{
Multiply imputes missing covariate values and event times using substantive model compatible
fully conditional specification with a Royston-Parmar flexible parametric survival model.
}
\details{
This version of \code{smcfcs} is for time-to-event outcomes which are modelled
using a flexible parametric proportional hazards survival model, as proposed
by Royston and Parmar (2002). The model is
fitted using the \code{\link[flexsurv]{flexsurvspline}} function in the
\pkg{flexsurv} package. Specifically it fits models using the hazard scale. The
flexibility of the model can be changed by modifying the k argument, which
specifies the number of knots.

If desired, \code{smcfcs.flexsurv} can be used to impute event times for individuals
who are originally censored, by specifying \code{imputeTimes=TRUE}. In the resulting
imputed datasets every individual will have an event time and the event indicator will
be one for all. Alternatively, you can impute censored times, but setting a larger
potential censoring time, which is either a common value used for all or a vector of times,
by using the \code{censtime} argument. If some individuals have their time-to-event
outcome completely missing and you want to impute this, they should have a time of zero
and the event indicator set to zero.

\code{smcfcs.flexsurv} will not let you impute using norm, latnorm or poisson methods
for variables that are allowed to have time-varying effects, because the usual
rejection sampling bound used by smcfcs is not valid in this setting.

\code{\link[flexsurv]{flexsurvspline}} sometimes fails during model fitting.
If/when this occurs, \code{smcfcs.flexsurv} takes a posterior draw based
on the model fit from the preceding iteration, and a warning is printed at
the end of the \code{smcfcs.flexsurv} run detailing how many times it occurred.
}
\examples{
#the following example is not run when the package is compiled on CRAN
#(to keep computation time down), but it can be run by package users
\dontrun{

  set.seed(63213)
  imps <- smcfcs.flexsurv(ex_flexsurv,
                          k=2,
                          smformula="Surv(t,d)~x+z",
                          method=c("","","logreg",""))
  library(mitools)
  impobj <- imputationList(imps$impDatasets)
  models <- with(impobj, flexsurvspline(Surv(t,d)~x+z, k=2))
  summary(MIcombine(models))

  # now impute event times as well as missing covariates
  imps <- smcfcs.flexsurv(ex_flexsurv,
                          k=2,
                          smformula="Surv(t,d)~x+z",
                          method=c("","","logreg",""),
                          imputeTimes=TRUE)

  # now impute event times as well as missing covariates,
  # but setting max observed event time to 2
  imps <- smcfcs.flexsurv(ex_flexsurv,
                          k=2,
                          smformula="Surv(t,d)~x+z",
                          method=c("","","logreg",""),
                          imputeTimes=TRUE,
                          censtime=2)

}
}
\references{
Royston P, Parmar MKB. Flexible parametric proportional-hazards
and proportional-odds models for censored survival data, with application
to prognostic modelling and estimation of treatment effects.
Statistics in Medicine 2002; 21(15): 2175-2197. \doi{doi:10.1002/sim.1203}
}
\author{
Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.parallel.r
\name{smcfcs.parallel}
\alias{smcfcs.parallel}
\title{Parallel substantive model compatible imputation}
\usage{
smcfcs.parallel(
  smcfcs_func = "smcfcs",
  seed = NULL,
  m = 5,
  n_cores = parallel::detectCores() - 1,
  cl_type = "PSOCK",
  outfile = "",
  ...
)
}
\arguments{
\item{smcfcs_func}{Specifies which base smcfcs function to call. Possible values
are `smcfcs`, `smcfcs.casecohort`, `smcfcs.dtasam`, `smcfcs.nestedcc`. Defaults
to `smcfcs`.}

\item{seed}{Optional seed, set as `set.seed` when `n_cores = 1`,
or as `parallel::clusterSetRNGStream` when `n_cores > 1`.}

\item{m}{Number of imputed datasets to generate.}

\item{n_cores}{Number of cores over which to split the `m` imputations. If
`n_cores` is not divisible exactly by `m`, one of the cores will perform
more/less imputations that the rest such that the final result still contains
`m` imputed datasets.}

\item{cl_type}{Either "PSOCK" or "FORK". If running on a Windows system
"PSOCK" is recommended, otherwise for Linux/Mac machines "FORK" tends to
offer faster computation - see \link[mice]{parlmice}.}

\item{outfile}{Optional character path to location for
output from the workers. Useful to diagnose rejection sampling warnings.
File path must be formulated as "path/to/filename.txt".}

\item{...}{Additional arguments to pass on to \link[smcfcs]{smcfcs},
\link[smcfcs]{smcfcs.casecohort},
\link[smcfcs]{smcfcs.dtsam}, or
\link[smcfcs]{smcfcs.nestedcc}.}
}
\value{
An object of type "smcfcs", as would usually be returned from
\link[smcfcs]{smcfcs}.
}
\description{
Runs substantive model compatible imputation using parallel cores
}
\details{
This function can be used to call one of the substantive model compatible imputation
methods using parallel cores, to reduce computation time. You must specify
the arguments required for the standard smcfcs call, and then specify your
the arguments for how to use parallel cores.
}
\examples{
\dontrun{
# Detect number of cores
parallel::detectCores()

imps <- smcfcs.parallel(
  smcfcs_func = "smcfcs",
  seed = 2021,
  n_cores = 2,
  originaldata = smcfcs::ex_compet,
  m = 10,
  smtype = "compet",
  smformula = list(
    "Surv(t, d == 1) ~ x1 + x2",
    "Surv(t, d == 2) ~ x1 + x2"
  ),
  method = c("", "", "norm", "norm")
)
}
}
\author{
Edouard F. Bonneville \email{e.f.bonneville@lumc.nl}

Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_logisticquad}
\alias{ex_logisticquad}
\title{Simulated example data with binary outcome and quadratic covariate effects}
\format{
A data frame with 1000 rows and 4 variables:
\describe{
  \item{y}{Binary outcome}
  \item{z}{Fully observed covariate, with linear effect on outcome (on log odds scale)}
  \item{x}{Partially observed normally distributed covariate, with quadratic effect on outcome (on log odds scale)}
  \item{v}{An auxiliary variable (i.e. not contained in the substantive model)}
}
}
\usage{
ex_logisticquad
}
\description{
A dataset containing simulated data where the binary outcome depends quadratically
on a partially observed covariate.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_cc}
\alias{ex_cc}
\title{Simulated case cohort data}
\format{
A data frame with 1571 rows and 7 variables:
\describe{
  \item{t}{Time to event or censoring}
  \item{d}{Indicator of whether event 1 occurred (d=1), or not (d=0)}
  \item{x}{Partially observed continuous covariate}
  \item{z}{Fully observed covariate}
  \item{in.subco}{A binary indicator of whether the subject is in the sub-cohort}
  \item{id}{An id variable}
  \item{entertime}{The entry time variable to be used in the analysis}
}
}
\usage{
ex_cc
}
\description{
A dataset containing simulated case cohort data, where the sub-cohort was a 10\% random sample of the full cohort.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_lininter}
\alias{ex_lininter}
\title{Simulated example data with continuous outcome and interaction between two partially observed covariates}
\format{
A data frame with 1000 rows and 4 variables:
\describe{
  \item{y}{Continuous outcome}
  \item{x1}{Partially observed normally distributed covariate}
  \item{x2}{Partially observed binary covariate}
}
}
\usage{
ex_lininter
}
\description{
A dataset containing simulated data where the outcome depends on both main
effects and interaction of two partially observed covariates.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_coxquad}
\alias{ex_coxquad}
\title{Simulated example data with time to event outcome and quadratic covariate effects}
\format{
A data frame with 1000 rows and 6 variables:
\describe{
  \item{t}{Time to event or censoring}
  \item{d}{Binary indicator of whether event occurred or individual was censored}
  \item{z}{Fully observed covariate, with linear effect on outcome (on log hazard scale)}
  \item{x}{Partially observed normally distributed covariate, with quadratic effect on outcome (on log hazard scale)}
  \item{xsq}{The square of x, which thus has missing values also}
  \item{v}{An auxiliary variable (i.e. not contained in the substantive model)}
}
}
\usage{
ex_coxquad
}
\description{
A dataset containing simulated data where a time to event outcome depends quadratically
on a partially observed covariate.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.r
\name{smcfcs.dtsam}
\alias{smcfcs.dtsam}
\title{Substantive model compatible fully conditional specification imputation of covariates for
discrete time survival analysis}
\usage{
smcfcs.dtsam(
  originaldata,
  smformula,
  timeEffects = "factor",
  method,
  predictorMatrix = NULL,
  m = 5,
  numit = 10,
  rjlimit = 1000,
  noisy = FALSE,
  errorProneMatrix = NULL
)
}
\arguments{
\item{originaldata}{The data in wide form (i.e. one row per subject)}

\item{smformula}{A formula of the form "Surv(t,d)~x1+x2+x3", where t is the discrete time variable, d is the binary event
indicator, and the covariates should not include time. The time variable should be
an integer coded numeric variable taking values from 1 up to the final time period.}

\item{timeEffects}{Specifies how the effect of time is modelled. \code{timeEffects="factor"} (the default) models time as a
factor variable. \code{timeEffects="linear"} and \code{timeEffects="quad"} specify that time be modelled as a continuous
linear or quadratic effect on the log odds scale respectively.}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{predictorMatrix}{An optional predictor matrix. If specified, the matrix defines which
covariates will be used as predictors in the imputation models
(the outcome must not be included). The i'th row of the matrix should consist of
0s and 1s, with a 1 in the j'th column indicating the j'th variable be used
as a covariate when imputing the i'th variable. If not specified, when
imputing a given variable, the imputation model covariates are the other
covariates of the substantive model which are partially observed
(but which are not passively imputed) and any fully observed covariates (if present)
in the substantive model. Note that the outcome variable is implicitly conditioned
on by the rejection sampling scheme used by smcfcs, and should not be specified as a predictor
in the predictor matrix.}

\item{m}{The number of imputed datasets to generate. The default is 5.}

\item{numit}{The number of iterations to run when generating each imputation.
In a (limited) range of simulations good performance was obtained with the
default of 10 iterations. However, particularly when the proportion of missingness
is large, more iterations may be required for convergence to stationarity.}

\item{rjlimit}{Specifies the maximum number of attempts which should be made
when using rejection sampling to draw from imputation models. If the limit is reached
when running a warning will be issued. In this case it is probably advisable to
increase the \code{rjlimit} until the warning does not appear.}

\item{noisy}{logical value (default FALSE) indicating whether output should be noisy, which can
be useful for debugging or checking that models being used are as desired.}

\item{errorProneMatrix}{An optional matrix which if specified indicates that some variables
are measured with classical measurement error. If the i'th variable is measured with error
by variables j and k, then the (i,j) and (i,k) entries of this matrix should be 1, with the
remainder of entries 0. The i'th element of the method argument should then be specified
as \code{"latnorm"}. See the measurement error vignette for more details.}
}
\description{
Multiply imputes missing covariate values using substantive model compatible
fully conditional specification for discrete time survival analysis.
}
\details{
For this substantive model type, like for the other substantive model types, \code{smcfcs} expects the \code{originaldata} to have
one row per subject. Variables indicating the discrete time of failure/censoring
and the event indicator should be passed in \code{smformula}, as described.

The default is to model the effect of time as a factor. This will not work in datasets where
there is not at least one observed event in each time period. In such cases you must specify
a simpler parametric model for the effect of time. At the moment you can specify either a linear or quadratic
effect of time (on the log odds scale).
}
\examples{
#the following example is not run when the package is compiled on CRAN
#(to keep computation time down), but it can be run by package users
\dontrun{
  #discrete time survival analysis example
  M <- 5
  imps <- smcfcs.dtsam(ex_dtsam, "Surv(failtime,d)~x1+x2",
                 method=c("logreg","", "", ""),m=M)
  #fit dtsam model to each dataset manually, since we need
  #to expand to person-period data form first
  ests <- vector(mode = "list", length = M)
  vars <- vector(mode = "list", length = M)
  for (i in 1:M) {
    longData <- survSplit(Surv(failtime,d)~x1+x2, data=imps$impDatasets[[i]],
                          cut=unique(ex_dtsam$failtime[ex_dtsam$d==1]))
    mod <- glm(d~-1+factor(tstart)+x1+x2, family="binomial", data=longData)
    ests[[i]] <- coef(mod)
    vars[[i]] <- diag(vcov(mod))
  }
  library(mitools)
  summary(MIcombine(ests,vars))
}
}
\author{
Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.r
\name{smcfcs.nestedcc}
\alias{smcfcs.nestedcc}
\title{Substantive model compatible fully conditional specification imputation of covariates for nested case control
studies}
\usage{
smcfcs.nestedcc(originaldata, smformula, method, set, event, nrisk, ...)
}
\arguments{
\item{originaldata}{The nested case-control data set (NOT a full cohort data set with a case-cohort substudy within it)}

\item{smformula}{A formula of the form "Surv(t,case)~x+strata(set)", where case is case-control indicator, t is the event or censoring time. Note that t could be set to the case's event time for the matched controls in a given set. The right hand side should include the case control set as a strata term (see example).}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{set}{variable identifying matched sets in nested case-control study}

\item{event}{variable which indicates who is a case/control in the nested case-control sample. Note that this is distinct from d.}

\item{nrisk}{variable which is the number at risk (in the underlying full cohort) at the event time for the case in each matched set (i.e. nrisk is the same for all individuals in a matched set).}

\item{...}{Additional arguments to pass on to \link[smcfcs]{smcfcs}}
}
\description{
Multiply imputes missing covariate values using substantive model compatible
fully conditional specification for nested case control studies.
}
\details{
This version of \code{smcfcs} is designed for use with nested case control studies. The function's arguments are the same
as for the main smcfcs function, except for \code{smformula}, \code{set}, \code{event} and \code{nrisk} - see above
for details on how these should be specified.
}
\examples{
#the following example is not run when the package is compiled on CRAN
#(to keep computation time down), but it can be run by package users
\dontrun{
  predictorMatrix <- matrix(0,nrow=dim(ex_ncc)[2],ncol=dim(ex_ncc)[2])
  predictorMatrix[which(colnames(ex_ncc)=="x"),c(which(colnames(ex_ncc)=="z"))] <- 1

  imps <- smcfcs.nestedcc(originaldata=ex_ncc,set="setno",nrisk="numrisk",event="d",
                          smformula="Surv(t,case)~x+z+strata(setno)",
                          method=c("", "", "logreg", "", "", "", "", ""),
                          predictorMatrix=predictorMatrix)
  library(mitools)
  impobj <- imputationList(imps$impDatasets)
  models <- with(impobj, clogit(case~x+z+strata(setno)))
  summary(MIcombine(models))
}
}
\author{
Ruth Keogh \email{ruth.keogh@lshtm.ac.uk}

Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_flexsurv}
\alias{ex_flexsurv}
\title{Simulated example data with time-to-event Weibull outcome and two covariates}
\format{
A data frame with 1000 rows and 4 variables:
\describe{
  \item{t}{Time to event (d=1) or censoring (d=0)}
  \item{d}{Event indicator}
  \item{x}{Partially observed binary covariate}
  \item{z}{Fully observed continuous covariate}
}
}
\usage{
ex_flexsurv
}
\description{
A dataset containing simulated data where the time-to-event outcome is Weibull
distributed with two covariates, one of which is partially observed.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_finegray}
\alias{ex_finegray}
\title{Simulated example data with competing risks outcome and partially observed covariates}
\format{
A data frame with 1000 rows and 4 variables:
\describe{
  \item{times}{Time to event or censoring}
  \item{d}{Indicator of whether event 1 occurred (d=1), event 2 occurred (d=2) or individual was censored (d=0)}
  \item{x1}{Partially observed binary covariate, with linear effects on log subdistribution hazard of cause 1}
  \item{x2}{Partially observed normally distributed (conditional on x1) covariate, with linear effects
  on log subdistribution hazard of cause 1}
}
}
\usage{
ex_finegray
}
\description{
A dataset containing simulated competing risks data. There are two competing risks, and
some times are also censored. Proportionality holds on the subdistribution hazard scale for
cause 2, where for dataset `ex_compet` it instead holds on the cause-specific
hazard scale.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.smcfcs.R
\name{plot.smcfcs}
\alias{plot.smcfcs}
\title{Assess convergence of a smcfcs object}
\usage{
\method{plot}{smcfcs}(x, include = "all", ...)
}
\arguments{
\item{x}{An object of class 'smcfcs'}

\item{include}{Character vector of coefficient names for which to return the
convergence plot. Default is "all" and returns plots for all coefficients in
a facetted manner.

Recommendation is to plot first with include = "all", and then select
coefficient names to zoom in to.

For competing risks, the coefficients are indexed by their cause. E.g. for
coefficient of a variable x1 in a model for cause 2, will be labelled
"x1-cause2".}

\item{...}{Additional parameters to pass on to ggplot2::facet_wrap(),
eg. nrow = 2}
}
\value{
A ggplot2 object, containing the convergence plots, facetted per
covariate in the substantive model
}
\description{
Visualises the contents of smCoefIter. Specifically, it plots the parameter
estimates of the substantive model against the number of iterations from
the imputation procedure. This is done for each regression coefficient,
and each line corresponds to an imputed dataset.
}
\details{
Requires loading of ggplot2 plotting library.
}
\examples{
\dontrun{
# Use simulated competing risks example in package
imps <- smcfcs(
  originaldata = ex_compet,
  smtype = "compet",
  smformula = list(
    "Surv(t, d == 1) ~ x1 + x2",
    "Surv(t, d == 2) ~ x1 + x2"
  ),
  method = c("", "", "norm", "norm")
)

plot(imps)
plot(imps, include = c("x1-cause1", "x2-cause2"))
}

}
\author{
Edouard F. Bonneville \email{e.f.bonneville@lumc.nl}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.r
\name{smcfcs}
\alias{smcfcs}
\title{Substantive model compatible fully conditional specification imputation of covariates.}
\usage{
smcfcs(
  originaldata,
  smtype,
  smformula,
  method,
  predictorMatrix = NULL,
  m = 5,
  numit = 10,
  rjlimit = 1000,
  noisy = FALSE,
  errorProneMatrix = NULL
)
}
\arguments{
\item{originaldata}{The original data frame with missing values.}

\item{smtype}{A string specifying the type of substantive model. Possible
values are \code{"lm"}, \code{"logistic"}, \code{"brlogistic"}, \code{"poisson"}, \code{"weibull"},
\code{"coxph"}, \code{"compet"}.}

\item{smformula}{The formula of the substantive model. For \code{"weibull"} and \code{"coxph"}
substantive models the left hand side should be of the form \code{"Surv(t,d)"}. For \code{"compet"}
substantive models, a list should be passed consisting of the Cox models
for each cause of failure (see example).}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{predictorMatrix}{An optional predictor matrix. If specified, the matrix defines which
covariates will be used as predictors in the imputation models
(the outcome must not be included). The i'th row of the matrix should consist of
0s and 1s, with a 1 in the j'th column indicating the j'th variable be used
as a covariate when imputing the i'th variable. If not specified, when
imputing a given variable, the imputation model covariates are the other
covariates of the substantive model which are partially observed
(but which are not passively imputed) and any fully observed covariates (if present)
in the substantive model. Note that the outcome variable is implicitly conditioned
on by the rejection sampling scheme used by smcfcs, and should not be specified as a predictor
in the predictor matrix.}

\item{m}{The number of imputed datasets to generate. The default is 5.}

\item{numit}{The number of iterations to run when generating each imputation.
In a (limited) range of simulations good performance was obtained with the
default of 10 iterations. However, particularly when the proportion of missingness
is large, more iterations may be required for convergence to stationarity.}

\item{rjlimit}{Specifies the maximum number of attempts which should be made
when using rejection sampling to draw from imputation models. If the limit is reached
when running a warning will be issued. In this case it is probably advisable to
increase the \code{rjlimit} until the warning does not appear.}

\item{noisy}{logical value (default FALSE) indicating whether output should be noisy, which can
be useful for debugging or checking that models being used are as desired.}

\item{errorProneMatrix}{An optional matrix which if specified indicates that some variables
are measured with classical measurement error. If the i'th variable is measured with error
by variables j and k, then the (i,j) and (i,k) entries of this matrix should be 1, with the
remainder of entries 0. The i'th element of the method argument should then be specified
as \code{"latnorm"}. See the measurement error vignette for more details.}
}
\value{
A list containing:

\code{impDatasets} a list containing the imputed datasets

\code{smCoefIter} a three dimension matrix containing the substantive model parameter
values. The matrix is indexed by [imputation,parameter number,iteration]
}
\description{
Multiply imputes missing covariate values using substantive model compatible
fully conditional specification.
}
\details{
smcfcs imputes missing values of covariates using the Substantive Model Compatible
Fully Conditional Specification multiple imputation approach proposed by
Bartlett \emph{et al} 2015 (see references).

Imputation is supported for linear regression (\code{"lm"}),
logistic regression (\code{"logistic"}), bias reduced logistic regression (\code{"brlogistic"}),
Poisson regression (\code{"poisson"}), Weibull (\code{"weibull"}) and Cox regression
for time to event data (\code{"coxph"}),
and Cox models for competing risks data (\code{"compet"}). For \code{"coxph"},
the event indicator should be integer coded with 0 for censoring and 1 for event.
For \code{"compet"}, a Cox model is assumed for each cause specific hazard function,
and the event indicator
should be integer coded with 0 corresponding to censoring, 1 corresponding to
failure from the first cause etc.

The function returns a list. The first element \code{impDataset} of the list is a list of the imputed
datasets. Models (e.g. the substantive model) can be fitted to each and results
combined using Rubin's rules using the mitools package, as illustrated in the examples.

The second element \code{smCoefIter} is a three dimensional array containing the values
of the substantive model parameters obtained at the end of each iteration of the algorithm.
The array is indexed by: imputation number, parameter number, iteration.

If the substantive model is linear, logistic or Poisson regression,
\code{smcfcs} will automatically impute missing outcomes, if present, using
the specified substantive model. However, even in this case, the user should
specify "" in the element of method corresponding to the outcome variable.

The bias reduced methods make use of the \code{\link[brglm2]{brglm2}} package to fit the corresponding glms
using Firth's bias reduced approach. These may be particularly useful to use in case
of perfect prediction, since the resulting model estimates are always guaranteed to be
finite, even in the case of perfect prediction.

The development of this package was supported by the UK Medical Research Council
(Fellowship MR/K02180X/1 and grant MR/T023953/1). Part of its development took place while Bartlett was
kindly hosted by the University of Michigan's Department of Biostatistics & Institute for
Social Research.

The structure of many of the arguments to \code{smcfcs} are based on those of
the excellent \code{mice} package.
}
\examples{
#set random number seed to make results reproducible
set.seed(123)

#linear substantive model with quadratic covariate effect
imps <- smcfcs(ex_linquad, smtype="lm", smformula="y~z+x+xsq",
               method=c("","","norm","x^2",""))

#if mitools is installed, fit substantive model to imputed datasets
#and combine results using Rubin's rules
if (requireNamespace("mitools", quietly = TRUE)) {
  library(mitools)
  impobj <- imputationList(imps$impDatasets)
  models <- with(impobj, lm(y~z+x+xsq))
  summary(MIcombine(models))
}

#the following examples are not run when the package is compiled on CRAN
#(to keep computation time down), but they can be run by package users
\dontrun{
  #examining convergence, using 100 iterations, setting m=1
  imps <- smcfcs(ex_linquad, smtype="lm", smformula="y~z+x+xsq",
                 method=c("","","norm","x^2",""),m=1,numit=100)
  #convergence plot from first imputation for third coefficient of substantive model
  plot(imps$smCoefIter[1,3,])

  #include auxiliary variable assuming it is conditionally independent of Y (which it is here)
  predMatrix <- array(0, dim=c(ncol(ex_linquad),ncol(ex_linquad)))
  predMatrix[3,] <- c(0,1,0,0,1)
  imps <- smcfcs(ex_linquad, smtype="lm", smformula="y~z+x+xsq",
                 method=c("","","norm","x^2",""),predictorMatrix=predMatrix)

  #impute missing x1 and x2, where they interact in substantive model
  imps <- smcfcs(ex_lininter, smtype="lm", smformula="y~x1+x2+x1*x2",
                 method=c("","norm","logreg"))

  #logistic regression substantive model, with quadratic covariate effects
  imps <- smcfcs(ex_logisticquad, smtype="logistic", smformula="y~z+x+xsq",
                 method=c("","","norm","x^2",""))

  #Poisson regression substantive model
  imps <- smcfcs(ex_poisson, smtype="poisson", smformula="y~x+z",
                 method=c("","norm",""))
  if (requireNamespace("mitools", quietly = TRUE)) {
    library(mitools)
    impobj <- imputationList(imps$impDatasets)
    models <- with(impobj, glm(y~x+z,family=poisson))
    summary(MIcombine(models))
  }

  #Cox regression substantive model, with only main covariate effects
  if (requireNamespace("survival", quietly = TRUE)) {
    imps <- smcfcs(ex_coxquad, smtype="coxph", smformula="Surv(t,d)~z+x+xsq",
                   method=c("","","","norm","x^2",""))

    #competing risks substantive model, with only main covariate effects
    imps <- smcfcs(ex_compet, smtype="compet",
                   smformula=c("Surv(t,d==1)~x1+x2", "Surv(t,d==2)~x1+x2"),
                   method=c("","","logreg","norm"))
  }

  #if mitools is installed, fit model for first competing risk
  if (requireNamespace("mitools", quietly = TRUE)) {
    library(mitools)
    impobj <- imputationList(imps$impDatasets)
    models <- with(impobj, coxph(Surv(t,d==1)~x1+x2))
    summary(MIcombine(models))
  }

  #discrete time survival analysis example
  M <- 5
  imps <- smcfcs(ex_dtsam, "dtsam", "Surv(failtime,d)~x1+x2",
                 method=c("logreg","", "", ""),m=M)
  #fit dtsam model to each dataset manually, since we need
  #to expand to person-period data form first
  ests <- vector(mode = "list", length = M)
  vars <- vector(mode = "list", length = M)
  for (i in 1:M) {
    longData <- survSplit(Surv(failtime,d)~x1+x2, data=imps$impDatasets[[i]],
                          cut=unique(ex_dtsam$failtime[ex_dtsam$d==1]))
    mod <- glm(d~-1+factor(tstart)+x1+x2, family="binomial", data=longData)
    ests[[i]] <- coef(mod)
    vars[[i]] <- diag(vcov(mod))
  }
  summary(MIcombine(ests,vars))

}
}
\references{
Bartlett JW, Seaman SR, White IR, Carpenter JR. Multiple imputation of covariates
by fully conditional specification: accommodating the substantive model. Statistical Methods
in Medical Research 2015; 24(4): 462-487. \doi{10.1177/0962280214521348}
}
\author{
Jonathan Bartlett \email{jonathan.bartlett1@lshtm.ac.uk}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_ncc}
\alias{ex_ncc}
\title{Simulated nested case-control data}
\format{
A data frame with 728 rows and 8 variables:
\describe{
  \item{t}{Time to event or censoring}
  \item{d}{Indicator of whether event 1 occurred (d=1), or not (d=0)}
  \item{x}{Partially observed binary covariate}
  \item{z}{Fully observed covariate}
  \item{id}{An id variable}
  \item{numrisk}{Number of patients at risk at time of case's event}
  \item{setno}{The case-control set number}
  \item{case}{Binary indicator of case (=1) or control (=0)}
}
}
\usage{
ex_ncc
}
\description{
A dataset containing simulated nested case-control data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_dtsam}
\alias{ex_dtsam}
\title{Simulated discrete time survival data set}
\format{
A data frame with 1000 rows and 8 variables:
\describe{
  \item{x1}{A binary variable with missing values}
  \item{x2}{A fully observed continuous variable}
  \item{failtime}{The discrete failure/censoring time}
  \item{d}{Indicator of failure (=1) or censoring (=0)}
}
}
\usage{
ex_dtsam
}
\description{
A dataset containing simulated discrete time survival data.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.r
\docType{data}
\name{ex_poisson}
\alias{ex_poisson}
\title{Simulated example data with count outcome, modelled using Poisson regression}
\format{
A data frame with 1000 rows and 3 variables:
\describe{
  \item{y}{Count outcome}
  \item{z}{Fully observed covariate, with linear effect on outcome}
  \item{x}{Partially observed normally distributed covariate, with linear effect on outcome}
}
}
\usage{
ex_poisson
}
\description{
A dataset containing simulated data where the count outcome depends on two
covariates, x and z, with missing values in x. The substantive model is
Poisson regression.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/smcfcs.finegray.R
\name{smcfcs.finegray}
\alias{smcfcs.finegray}
\title{Substantive model compatible fully conditional specification imputation of
covariates for a Fine-Gray model}
\usage{
smcfcs.finegray(
  originaldata,
  smformula,
  method,
  cause = 1,
  m = 5,
  numit = 10,
  rjlimit = 5000,
  kmi_args = list(formula = ~1, bootstrap = FALSE, nboot = 10),
  ...
)
}
\arguments{
\item{originaldata}{The original data frame with missing values.}

\item{smformula}{The formula of the substantive model, given as a string. Needs to
be of the form "Surv(t, d) ~ x1 + x2", where t is a vector of competing event
times, and d is a (numeric) competing event indicator, where 0 must designate
a censored observation.}

\item{method}{A required vector of strings specifying for each variable either
that it does not need to be imputed (""), the type of regression model to be
be used to impute. Possible values are \code{"norm"} (normal linear regression),
\code{"logreg"} (logistic regression), \code{"brlogreg"} (bias reduced logistic regression),
\code{"poisson"} (Poisson regression),
\code{"podds"} (proportional odds regression for ordered categorical variables),
\code{"mlogit"} (multinomial logistic regression for unordered categorical variables),
or a custom expression which defines a passively imputed variable, e.g.
\code{"x^2"} or \code{"x1*x2"}. \code{"latnorm"} indicates the variable is a latent
normal variable which is measured with error. If this is specified for a variable,
the \code{"errorProneMatrix"} argument should also be used.}

\item{cause}{Numeric, designating the competing event of interest (default is
`cause = 1`).}

\item{m}{The number of imputed datasets to generate. The default is 5.}

\item{numit}{The number of iterations to run when generating each imputation.
In a (limited) range of simulations good performance was obtained with the
default of 10 iterations. However, particularly when the proportion of missingness
is large, more iterations may be required for convergence to stationarity.}

\item{rjlimit}{Specifies the maximum number of attempts which should be made
when using rejection sampling to draw from imputation models. If the limit is reached
when running a warning will be issued. In this case it is probably advisable to
increase the \code{rjlimit} until the warning does not appear.}

\item{kmi_args}{List, containing arguments to be passed on to \link[kmi]{kmi}.
The "formula" element is a formula where the right-hand side specifies the
covariates used for multiply imputing the potential censoring times for
individual's failing from competing events. The default is `formula = ~ 1`,
which uses marginal Kaplan-Meier estimator of the censoring distribution.}

\item{...}{Additional arguments to pass on to \link[smcfcs]{smcfcs}}
}
\value{
An object of type "smcfcs", as would usually be returned from
\link[smcfcs]{smcfcs}.
}
\description{
Multiply imputes missing covariate values using substantive model compatible
fully conditional specification for competing risks outcomes, when the
substantive model is a Fine-Gray model for the subdistribution hazard of one event.
}
\details{
In the presence of random right censoring, the function first multiply imputes
the potential censoring times for those failing from competing events using
\link[kmi]{kmi}, and thereafter uses \link[smcfcs]{smcfcs} to impute the missing
covariates. See Bonneville \emph{et al.} 2024 for further details on the methodology.

The function does not (yet) support parallel computation.
}
\examples{
\dontrun{
library(survival)
library(kmi)

imps <- smcfcs.finegray(
  originaldata = ex_finegray,
  smformula = "Surv(times, d) ~ x1 + x2",
  method = c("", "", "logreg", "norm"),
  cause = 1,
  kmi_args = list("formula" = ~ 1)
)

if (requireNamespace("mitools", quietly = TRUE)) {
  library(mitools)
  impobj <- imputationList(imps$impDatasets)
  # Important: use Surv(newtimes, newevent) ~ ... when pooling
  # (respectively: subdistribution time and indicator for cause of interest)
  models <- with(impobj, coxph(Surv(newtimes, newevent) ~ x1 + x2))
  summary(MIcombine(models))
}
}

}
\references{
Bonneville EF, Beyersmann J, Keogh RH, Bartlett JW, Morris TP,
Polverelli N, de Wreede LC, Putter H. Multiple imputation of missing covariates
when using the Fine--Gray model. 2024. Submitted.
}
\author{
Edouard F. Bonneville \email{e.f.bonneville@lumc.nl}
}
